.TH mdlc l "" "" "v4.5.3"
.SH Name
.PP
mdlc - Module Description Language Compiler

.B mdlc
[-v][-version][-B
.I baseName
][-debug]
.br
[-mdf | -mdfB
.I mdfBaseName
][-force]
.br
[-simtext][-simX][-simP][-type
.I typelist
][-cxxinst
.I cxxinstfile
]
.br
[-proto][-inc
.I libSyncInterface
][ -Idir][-access][-info]
.br
[
.I mdlFile
]+
.SH Description
.PP
Mdlc compiles several mdl (Module Description Language) files into a C++ class
for use with libSync. Each file describes or defines only one module, and exactly
one file must define a new module unless the -proto option is given (else, mdlc 
has nothing or too much to do).
.br
The mdl files are automatically produced by occ++ when it compiles oc code if
the -mdf option is given. Such files are Module Description Files, so their
extension is normally .mdf.
.br
To build a new module, its definition is given in mdl in terms of other modules.
Then, the mdf file and the related mdf files are given to mdlc to produce the
corresponding C++ class. Mdlc writes two files:
.RS
.IP + 3
<moduleName>.H contains the interface of class <moduleName>
.IP + 3
<moduleName>.C contains the code of class <moduleName>
.RE
.PP
where <moduleName> is the name of the new module.
.br
The class <moduleName> produced by mdlc is a subclass of Composite. Class
Composite is defined in libSync.H and libSync.a (see libSync(l)). <moduleName> 
is also a subclass of another synchronous class if the new module is defined 
by inheritance.
.br
.PP
The following options are supported:
.RS
.TP 3
.B \-v
verbose output. With this option, mdlc will show what it sees in the mdl
source files (descriptions, definitions, signals, connections etc...)
.TP 3
.B \-version
prints the version number of mdlc
.TP 3
.B \-B baseName
set base name for output files. For example, \'mdlc -B bar foo.mdl tic.mdf\' will 
produce \'bar.H\' and \'bar.C\'. If the name of the module defined in \'foo.mdl\' 
is foo, the name of the class will be \'foo\' because the -B option only sets the 
base name of the files.
.TP 3
.B \-debug
insert debugging code in the C++ class. This code will be compiled if the symbol 
_SYNC_DEBUG_ is defined. The debugging code will print the name of the module
and the current operation on the standard error file.
.br
The debugged operations are:
.RS
.IP + 3
reset of output signals
.br
In this phase, the module resets its output signals to the non-present state.
.IP + 3
set of input signals
.br
In this phase, the module gets the value of its input signals. The debugging code 
will display the value of each input signal.
.IP + 3
activation
.br
The debugging code will display the name of the module, and will show the reaction 
of each of its components.
.RE
.TP 3
.B \-mdf
produce a Module Description File. Mdlc will produce a file named <moduleName>.mdf
that contains a description of the new module.
The name of the mdf file is not affected by the -B option.
.TP 3
.B \-mdfB mdfBaseName
same as -mdf, but set a prefix for the name of the mdf file which will be 
<mdfBaseName>.mdf . This prefix can include a directory path, allowing the mdf file
to be created in a specific directory. For example \'mdlc -mdfB mdfFiles/foo foo.mdl
tic.mdf\' will save \'foo.mdf\' in the \'mdfFiles\' directory.
.TP 3
.B \-force
force the reaction of submodules, even if they do not seem to contribute to the 
reaction of the module according to the dependency graph. This is useful when 
the reaction of these submodules have side effects.
.TP 3
.B \-simtext
build the source file for a text oriented simulation of each described module. With
this option, mdlc does not require that a description file describes a new module.
The output file is named <moduleName>SimulT.C, where <moduleName> is the name of the
simulated module.
.TP 3
.B \-simX
build the source file for a graphic simulation of each described module. With
this option, mdlc does not require that a description file describes a new module.
The resulting application uses XInterface and must be linked with Motif. The output
file is named <moduleName>SimulX.C, where <moduleName> is the name of the simulated module.
.TP 3
.B \-simP
build the source file for a protocol simulation of each described module. With
this option, mdlc does not require that a description file describes a new module.
The resulting application uses the ProtoSim class which translates 
between events in the module and the standard input and standard diagnostic character 
streams. See the ProtoSim documentation for details on the simulation protocol. 
The output file is named <moduleName>SimulX.C, where <moduleName> is the name of the 
simulated module. The resulting program may be used with the simshell 
tool to be driven by a specific simulation interface.
.TP 3
.B \-type typelist
With -simtext or -simX, if the module is a template, this option allows you to specify
the type for which you want to build the simulation. For example, if you want to build
the source file for a text oriented simulation of a Delay<int>, you should type:
.br
mdlc -simtext -type int Delay.mdf
.br
If the template requires several arguments, you must provide them, separated by commas.
For example, the template Example<T,U> requires two arguments, so the command line should
be: \'mdlc -simtext -type "int,float" Example.mdf\' for instance. The types may be
instanciations of templates themself, as in:
.br
mdlc -simX -type "int,List<double>" Example.mdf
.TP 3
.B \-cxxinst cxxinstfile
With -simtext or -simX, if you use DEC cxx, the instantiation of the templates of libSync
is done with a template instantiation file. The default file is <Esterel/CxxInstanciation.H>.
You can choose another instantiation file with this option.
.TP 3
.B \-proto
build the prototype of a Synchronous Class for each module description. With this 
option, mdlc does not require that a description file describes a new module.
.br
For each description of a module <mod>, mdlc will produce <mod>.H.proto and 
<mod>.C.proto. The interface file is almost ready to use as is. In the C++ code 
file, you should write the code for the reaction of the module, and the code for 
the constructor and the destructor if there is anything special to do there.
.TP 3
.B \-inc libSyncInterface
Use libSyncInterface as the interface file of the Synchronous Library. The 
default value is <Esterel/libSync.H>.
.br
\'mdlc -inc libSync.H ...\' or \'mdlc -inc '"libSync.H"' ...\' will use "libSync.H".
.br
\'mdlc -inc "<xxx.H>" ...\' will use <xxx.H>. Do not forget to quote '<xxx.H>' to
prevent the shell from interpreting \'<\' and \'>\' as input/output redirection.
.TP 3
.B \-Idir
Add dir to the list of pathes to search for description files. When mdlc finds
a #use xxxx directive, it looks for xxxx.mdf in each directory specified by
-I, in the order in which they were specified on the command line.
.br
If no directory was specified by -I, or if none contains xxxx.mdf, mdlc looks 
in the default directory which is /usr/local/include/Esterel. This default 
directory can be changed by defining the MDLC_INC_PATH environment variable.
.TP 3
.B \-access
Display access rights to mdlc.
.TP 3
.B \-info
Display information about the compilation of mdlc (when, where, by whom).
.PP
.SH "Interface of the C++ classes produced by mdlc"
.PP
The C++ classes produced by mdlc are subclasses of Composite, which is itself a
subclass of Synchronous. Those classes are defined in libSync.H and libSync.a.
For example, the following mdl definition:
.PP
.RS
compo : {
.RS
input integer a;
.br
output integer s;
.PP
filterinput i;
.br
filteroutput o;
.br
base b;
.PP
i.a = a;
.br
s = o.s;
.PP
b.a << i.s;
.br
o.a << b.s;
.RE
}
.RE
.PP
becomes the following class through mdlc:
.PP
.RS
class compo : public Composite {
.RS
private:
.RS
filterinput i;
.br
filteroutput o;
.br
base b;
.RE
public:
.RS
void resetOutputs();
.br
void resetInputs();
.br
void setInputs();
.br
void activate();
.br
virtual InSignal<integer>& a();
.br
virtual OutSignal<integer>& s();
.br
virtual const InputSignal** inSigList() const; 
.br
        // returns a null terminated list of input signals
.br
virtual const OutputSignal** outSigList() const; 
.br
        // returns a null terminated list of output signals
.br
compo(const char* name = "compo",
.br
      const boolean schedule = true);     // Constructor
.br
.RE
private:
.RS
compo(const compo &);   // Private copy constructor
.RE
.RE
};
.RE
.PP
The private part holds the modules that compo uses to react.
.br
In the public part, the resetOutputs() method is called to reset the output 
signals of the module in their non-present state.
.br
The resetInputs() method resets the input signals of the module to the 
non-present state.
.br
The setInputs() method fetches the value of the input signals of the module.
.br
The activate() method makes the module react to its input event.
.PP
The methods a() and s() give access to the signals of the module.
.br
The inSigList() methods yields a null terminated list of these input signals.
In this list, signals are considered as instances of InputSignal, because they
are only used to walk through the interconnection graph of the modules.
.br
The outSigList() methods yields a null terminated list of the output signals.
In this list, signals are considered as instances of OutputSignal, because they
are only used to check if InputSignals are connected to them.
.br
The constructor compo() is used to build new instances of compo. The first argument
is the name of the object. This name is mainly used for debugging purposes and is
optional. The default value is the name of the module.
The second argument is optional and tells if we want the new object to register
itself on the current clock and to be automaticaly scheduled by this clock.
.br
The copy constructor is declared private to prohibit its use. Copying 
synchronous objects has no meaning.

.SH The Module Description Language
.PP
Mdl can be used to describe modules, to define composite modules and to define
modules through inheritance. Comments begin at a \'//\' token and end at the end
of the line.
.IP +
Module Description
.br
The description of a module in mdl gives its name and its interface. It also gives
information about the source files for this module (oc, C++, Esterel ou mdl).
.br
The description of a module has the following structure:
.PP
.RS
\#source <sourceFile>       // Esterel or mdl source file
.br
\#oc <ocSourceFile>         // If any
.br
\#c++ <classInterfaceFile>  // C++ file that gives the 
.br
                           // interface of the class
.br
\#import <userDefsFile>     // definition file included
.br
                           // by the interface file
.br
\#param {                   // Parameters of the module
.br
  <typeOfParam1> <nameOfParam1>
.br
  <typeOfParam2> <nameOfParam2>
.br
     ...
.br
}
.PP
<newType> is <oldType>;    // Definition of a new type
.PP
<newType> as <oldType>;    // Informs mdlc that newType
.br
                           // is same as oldType
.PP
<moduleName> {
.RS
\#nodep <sigName>, <sigName> ...;
.br
input: <pureSigName>, <sigType> <typedSigName> ...;  
.br
                          // input signals
.br
output: <pureSigName>, <sigType> <typedSigName> ...;
.br
                          // output signals
.RE
}
.PP
The \#source keyword introduces the name of the source file, if it is known.
.PP
The \#oc keyword introduces the name of the oc source file if the 
declaration was generated by occ++.
.PP
The \#c++ keyword introduces the name of the C++ file where the 
declaration of the class can be found.
.PP
The \#import keyword gives the name of a file containing user definitions 
used by the interface file. If this keyword is present, the class 
interface file \#includes the user definition file <def>.H, and the code 
associated with those definitions should be in <def>.C.
.PP
The \#param keyword, if present, declares the parameters of the module. 
These parameters are gathered in a structure of type <moduleName>Consts, 
so only one argument is needed to parameterize a module.
.PP
The \'is\' keyword is used to define a new type as an equivalent for another 
type. It is very similar to the \'typedef\' of C or C++. The statement \'<newType> 
is <oldType>;\' is translated into \'typedef <oldType> <newType>;\' in C++.
.PP
The \'as\' keyword is used to inform mdlc that two types are equivalent.
The statement \'<newType> as <oldType>;\' has no counterpart in the C++ code,
it is only used by mdlc. For example, integer is defined as int in the synchronous
library (because Lustre and Esterel use integer for the C type "int"), but mdlc
does not know about this. So, if you want mdlc to consider integer and int as the 
same type, you must write "integer as int;". You should not write "integer is int;" 
because mdlc should not define integer as int since it is already defined by libSync.
.PP
The \'input:\' , \'output:\' and \'#nodep\' keywords can appear several 
times. \'input:\' declares input signals, along with their type. If the 
type is missing, the signal is assumed to be pure. \'output:\' does the 
same for output signals.
.PP
The \'#nodep\' keyword introduces a list of input or output signals. The reaction 
of the module does not depend on the current value of the input signals of this 
list, and the value of the output signals of this list does not depend on the 
current reaction of the module. For example, in a delay, the output depends only 
on the previous reaction of the module. This can be stated with a #nodep statement.
.PP
The name of the module can be qualified by a list of type names between angle 
brackets. 
The module is then generic and must be qualified by a list of real types when 
it is used. 
.PP
The declaration of the module may specify an argument to the constructor. 
This argument is declared between parentheses between the name of the 
module and the opening brace. The special keyword \#param between 
parentheses indicates that the constructor of the module takes a constant 
reference on a <moduleName>Const structure as its first argument. The 
name of this first argument to the constructor is always \'initInstance\'.
.br
For example, to describe a generic Delay module taking a pointer on its 
initial value as argument to its constructor, one can write:
.PP
.RS
Delay<T>(const T*) {
.RS
input: T input;
.br
#nodep output;
.br
output: T output;
.RE
}
.RE
.PP
And then use a Delay for a signal of type \'int\' in:
.PP
.RS
X(const int) : {
.RS
input: int i;
.br
output: int o;
.PP
Delay<int> d(&initInstance);
.br
Y y;
.PP
d.input = i;
.br
y.inp << d.output;
.br
o = y.out;
.RE
}
.RE
.RE
.IP +
Composite Module Definition
.br
The definition of a composite module in mdl gives its structure (which other modules
are used to build the composite module) and the connections between the components.
.br
Signals are connected by the \'<<\' operator which connects an input signal to
an output signal of the same type.
.br
A signal can be identified to another signal by the \'=\' operator. This operator
specifies that a signal of the composite module \'is in fact\' a signal of one of
its components. To be identified, two signals must be of the same type and of the
same nature (input or output). Moreover, the left side operand must be the signal 
that logically receives the information from the right side operand. So, to
identify two input signals, the signal of the component is the left side operand
and the signal of the composite module is the right side operand. For two output
signals, the signal of the composite module is the left side operand and the signal
of the component is the left side operand.
.br
The description of a composite module has the following structure:
.PP
.RS
#Hinsert                   // Introduces a section of code to
.br
                           // insert in the .H file
.br
Anything you type here is not interpreted by mdlc
.br
and is merely copied to the .H file
.br
#endInsert                 // Ends the section of literal code
.PP
#Cinsert                   // Introduces a section of code to
.br
                           // insert in the .C file
.br
Anything you type here is not interpreted by mdlc
.br
and is merely copied to the .C file
.br
#endInsert                 // Ends the section of literal code
.PP
#use file                  // Tells mdlc that this definition uses
.br
                           // declarations found in file.mdf.
.br
                           // file.mdf is searched in the list
.br
                           // of directories specified by the -I
.br
                           // options and eventually in the 
.br
                           // default directory.
.PP
<moduleName> : {
.RS
<signal declarations>        // Just like in a module description
.PP
<moduleType> <subModName>, <subModName> ...;
.br
<moduleType> <subModName>, <subModName> ...;
.br
  ...                       // Declaration of the components
.PP
<subModName>.<InSigName> = <ModuleInSigName>;
.br
<ModuleOutSigName> = <subModName>.<OutSigName>;
.br
  ...                      // Signal identities
.PP
<subModName>.<InSigName> << <subModName>.<OutSigName>;
.br
  ...                      // Signal connections
.RE
}
.PP
For example, to define an ABS_Brake from a Brake and an ABS_Controler, we write:
.PP
.RS
ABS_Brake : {
.RS
input: integer pedal, skid;
.br
output: integer pressure;
.PP
Brake regular_brake;
.br
ABS_Controler abs_ctrl;
.PP
regular_brake.pedal = pedal;
.br
abs_ctrl.skid = skid;
.br
pressure = abs_ctrl.filtered_pressure;
.PP
abs_ctrl.raw_pressure << regular_brake.pressure;
.RE
}
.RE
.PP
assuming that the description of Brake and ABS_Controler is:
.PP
.RS
Brake {
.RS
input: integer pedal;
.br
output: integer pressure;
.RE
}
.PP
ABS_Controler {
.RS
input: skid, integer raw_pressure;
.br
output: integer filtered_pressure;
.RE
}
.RE
.PP
Composite Modules can be defined as templates. The name of the module is 
then qualified by a list of type names between angle brackets. These type 
names can be used as real types inside the definition of the module. The C++ 
Synchronous Class created by mdlc for such a generic module will be a template 
with the same parameters as the module. For example, the following module:
.PP
.RS
X<T,U> : {
.RS
input: T i;
.br
output: U o;
.PP
Z<T> z;
.br
H<int, U, float> h;
.PP
// Connections and equivalences
.RE
}
.RE
.PP
gives the class: template<class T, class U> class X { ... }; through mdlc.
.PP
Composite modules can take an additional first argument to their 
constructor. The type of this argument is declared, between parentheses, 
after the name of the module (including type parameters) and before the 
\':\'. This argument is named \'initInstance\' in the C++ code.
.RE
.IP +
Derived Module Description
.br
Mdl allows the description of a module that inherits its behaviour from 
another module.  The derived module inherits the signals and the behaviour 
of its super-module.  The description gives the additional signals and the 
components used to filter the inherited behaviour.
.br
The description of a derived module in mdl has the following structure:
.RS
.PP
.RS
\#Hinsert, \#Cinsert and \#use can be placed here, as for a Composite module.
.PP
<moduleName> : <superModuleName> {
.RS
<additional signal declaration>
.br
<signal identities>
.br
<signal connections>
.RE
}
.RE
.PP
The signals of the inherited module can be accessed through the \'super\' 
pseudo-module name. This \'super\' can always be omitted because the context always
allows mdlc to determine if a signal name refers to a signal of the inherited
module or to a signal of the derived module.
.PP
For example, the ABS_Brake module can be defined as deriving from Brake:
.PP
.RS
ABS_Brake : Brake {
.RS
input: skid;           // Just add skid to Brake
.PP
ABS_Controler abs_ctrl;// Add the filtering behaviour
.PP
abs_ctrl.skid = skid;
.br
pressure = abs_ctrl.filtered_pressure;
.PP
abs_ctrl.raw_pressure << pressure;
.RE
}
.RE
.PP
In the \'pressure = abs_ctrl.filtered_pressure;\' statement, pressure is the 
output signal of the derived module (because an output signal on the left side
of the \'=\' operator is a signal of the new module).
.br
In the \'abs_ctrl.raw_pressure << pressure;\' statement, pressure is 
\'super.pressure\' because a signal can only be connected to a signal of a
component.
.br
The behaviour of ABS_Brake defined as a derived module is the same as the behaviour
of ABS_Brake defined as a composite module. The main difference is that the ABS_Brake
C++ class is a subclass of Brake when ABS_Brake is defined as derived. This allows
instances of Brake and ABS_Brake to be referenced through Brake pointers.
.PP
Derived modules can be defined as templates just as composite modules, 
and they can take an additional first argument to their constructor too.
.RE
.RE
.IP+
Parameters to constructors
.PP
A Synchronous Class must have a name (char*) and a scheduling attribute (boolean)
as the two last arguments of its constructor. But there may be other arguments 
before those two. In such a case, you must give these arguments when you declare
the submodule in a module definition.
.PP
For example, if you use a Delay<int>, that takes an int* (initial value) 
as its first constructor argument, you should declare it as follows:
.PP
Delay<int> d(0);
.PP
to give a null pointer (no initial value). If you want to give an initial value 
of 5, you must give a pointer on a variable whose value is five. For example:
.PP
Delay<int> d(&initValue);
.PP
The \'initValue\' variable must be declared and defined, so you should have a
\#Cinsert section looking like:
.PP
\#Cinsert
.br
static const int initValue = 5;
.br
\#endInsert
.PP
The \'static\' avoids pollution of the global name space. If you do not want 
to declare a variable to store the initial value, you can build a new int just 
to give it as the initial value:
.PP
Delay<int> d(new int(5));
.PP
If the initial value of the delay should be a parameter of your new 
module X, you should declare that X takes an argument to its constructor, 
and use this argument to initialize the delay: 
.PP
.RS
X(const int) : {
.RS
Delay<int> d(&initValue);
   ...
.RE
}
.RE
.PP
If you need several arguments to the constructor of X, for instance if 
you have several delays to initialize with distinct values, you should 
use a \#param structure:
.PP
.RS
#param {
.RS
int init1;
.br
float init2;
.RE
}
.br
X(#param) : {
.RS
Delay<int> d1(&(initValue.init1));
.br
Delay<float> d2(&(initValue.init2));
.RE
}
.RE
.PP
Mdlc does not check the type of the argument to the constructor. It just 
takes literally what you put between parentheses.
.RE
.SH Author
.IP
Frederic Boulanger - Frederic.Boulanger@supelec.fr
.br
Supelec - Service Informatique
.br
Plateau de Moulon
.br
91192 Gif-sur-Yvette Cedex
.br
FRANCE
.SH See also
.B occ++(l)
, 
.B libSync(l)
