                                                                      mdlc(l)
                                                                       v4.5.3



   Name

     mdlc - Module Description Language Compiler

     mdlc [-v][-version][-B baseName ][-debug]
     [-mdf | -mdfB mdfBaseName ][-force]
     [-simtext][-simX][-simP][-type typelist ][-cxxinst cxxinstfile ]
     [-proto][-inc libSyncInterface ][ -Idir][-access][-info]
     [ mdlFile ]+

   Description

     Mdlc compiles several mdl (Module Description Language) files into a C++
     class for use with libSync. Each file describes or defines only one
     module, and exactly one file must define a new module unless the -proto
     option is given (else, mdlc has nothing or too much to do).
     The mdl files are automatically produced by occ++ when it compiles oc
     code if the -mdf option is given. Such files are Module Description
     Files, so their extension is normally .mdf.
     To build a new module, its definition is given in mdl in terms of other
     modules.  Then, the mdf file and the related mdf files are given to mdlc
     to produce the corresponding C++ class. Mdlc writes two files:

          +  <moduleName>.H contains the interface of class <moduleName>

          +  <moduleName>.C contains the code of class <moduleName>

     where <moduleName> is the name of the new module.
     The class <moduleName> produced by mdlc is a subclass of Composite.
     Class Composite is defined in libSync.H and libSync.a (see libSync(l)).
     <moduleName> is also a subclass of another synchronous class if the new
     module is defined by inheritance.

     The following options are supported:

          -v verbose output. With this option, mdlc will show what it sees in
             the mdl source files (descriptions, definitions, signals, con-
             nections etc...)

          -version
             prints the version number of mdlc

          -B baseName
             set base name for output files. For example, 'mdlc -B bar
             foo.mdl tic.mdf' will produce 'bar.H' and 'bar.C'. If the name
             of the module defined in 'foo.mdl' is foo, the name of the class
             will be 'foo' because the -B option only sets the base name of
             the files.

          -debug
             insert debugging code in the C++ class. This code will be com-
             piled if the symbol _SYNC_DEBUG_ is defined. The debugging code
             will print the name of the module and the current operation on
             the standard error file.
             The debugged operations are:

             +  reset of output signals
                In this phase, the module resets its output signals to the
                non-present state.

             +  set of input signals
                In this phase, the module gets the value of its input sig-
                nals. The debugging code will display the value of each input
                signal.

             +  activation
                The debugging code will display the name of the module, and
                will show the reaction of each of its components.

          -mdf
             produce a Module Description File. Mdlc will produce a file
             named <moduleName>.mdf that contains a description of the new
             module.  The name of the mdf file is not affected by the -B
             option.

          -mdfB mdfBaseName
             same as -mdf, but set a prefix for the name of the mdf file
             which will be <mdfBaseName>.mdf . This prefix can include a
             directory path, allowing the mdf file to be created in a
             specific directory. For example 'mdlc -mdfB mdfFiles/foo foo.mdl
             tic.mdf' will save 'foo.mdf' in the 'mdfFiles' directory.

          -force
             force the reaction of submodules, even if they do not seem to
             contribute to the reaction of the module according to the depen-
             dency graph. This is useful when the reaction of these submo-
             dules have side effects.

          -simtext
             build the source file for a text oriented simulation of each
             described module. With this option, mdlc does not require that a
             description file describes a new module.  The output file is
             named <moduleName>SimulT.C, where <moduleName> is the name of
             the simulated module.

          -simX
             build the source file for a graphic simulation of each described
             module. With this option, mdlc does not require that a descrip-
             tion file describes a new module.  The resulting application
             uses XInterface and must be linked with Motif. The output file
             is named <moduleName>SimulX.C, where <moduleName> is the name of
             the simulated module.

          -simP
             build the source file for a protocol simulation of each
             described module. With this option, mdlc does not require that a
             description file describes a new module.  The resulting applica-
             tion uses the ProtoSim class which translates between events in
             the module and the standard input and standard diagnostic char-
             acter streams. See the ProtoSim documentation for details on the
             simulation protocol. The output file is named
             <moduleName>SimulX.C, where <moduleName> is the name of the
             simulated module. The resulting program may be used with the
             simshell tool to be driven by a specific simulation interface.

          -type typelist
             With -simtext or -simX, if the module is a template, this option
             allows you to specify the type for which you want to build the
             simulation. For example, if you want to build the source file
             for a text oriented simulation of a Delay<int>, you should type:
             mdlc -simtext -type int Delay.mdf
             If the template requires several arguments, you must provide
             them, separated by commas.  For example, the template
             Example<T,U> requires two arguments, so the command line should
             be: 'mdlc -simtext -type "int,float" Example.mdf' for instance.
             The types may be instanciations of templates themself, as in:
             mdlc -simX -type "int,List<double>" Example.mdf

          -cxxinst cxxinstfile
             With -simtext or -simX, if you use DEC cxx, the instantiation of
             the templates of libSync is done with a template instantiation
             file. The default file is <Esterel/CxxInstanciation.H>.  You can
             choose another instantiation file with this option.

          -proto
             build the prototype of a Synchronous Class for each module
             description. With this option, mdlc does not require that a
             description file describes a new module.
             For each description of a module <mod>, mdlc will produce
             <mod>.H.proto and <mod>.C.proto. The interface file is almost
             ready to use as is. In the C++ code file, you should write the
             code for the reaction of the module, and the code for the con-
             structor and the destructor if there is anything special to do
             there.

          -inc libSyncInterface
             Use libSyncInterface as the interface file of the Synchronous
             Library. The default value is <Esterel/libSync.H>.
             'mdlc -inc libSync.H ...' or 'mdlc -inc '"libSync.H"' ...' will
             use "libSync.H".
             'mdlc -inc "<xxx.H>" ...' will use <xxx.H>. Do not forget to
             quote '<xxx.H>' to prevent the shell from interpreting '<' and
             '>' as input/output redirection.

          -Idir
             Add dir to the list of pathes to search for description files.
             When mdlc finds a #use xxxx directive, it looks for xxxx.mdf in
             each directory specified by -I, in the order in which they were
             specified on the command line.
             If no directory was specified by -I, or if none contains
             xxxx.mdf, mdlc looks in the default directory which is
             /usr/local/include/Esterel. This default directory can be
             changed by defining the MDLC_INC_PATH environment variable.

          -access
             Display access rights to mdlc.

          -info
             Display information about the compilation of mdlc (when, where,
             by whom).

   Interface of the C++ classes produced by mdlc

     The C++ classes produced by mdlc are subclasses of Composite, which is
     itself a subclass of Synchronous. Those classes are defined in libSync.H
     and libSync.a.  For example, the following mdl definition:

          compo : {
            input integer a;
            output integer s;

            filterinput i;
            filteroutput o;
            base b;

            i.a = a;
            s = o.s;

            b.a << i.s;
            o.a << b.s;
          }

     becomes the following class through mdlc:

          class compo : public Composite {
            private:
              filterinput i;
              filteroutput o;
              base b;
            public:
              void resetOutputs();
              void resetInputs();
              void setInputs();
              void activate();
              virtual InSignal<integer>& a();
              virtual OutSignal<integer>& s();
              virtual const InputSignal** inSigList() const;
                      // returns a null terminated list of input signals
              virtual const OutputSignal** outSigList() const;
                      // returns a null terminated list of output signals
              compo(const char* name = "compo",
                    const boolean schedule = true);     // Constructor
            private:
              compo(const compo &);   // Private copy constructor
          };

     The private part holds the modules that compo uses to react.
     In the public part, the resetOutputs() method is called to reset the
     output signals of the module in their non-present state.
     The resetInputs() method resets the input signals of the module to the
     non-present state.
     The setInputs() method fetches the value of the input signals of the
     module.
     The activate() method makes the module react to its input event.

     The methods a() and s() give access to the signals of the module.
     The inSigList() methods yields a null terminated list of these input
     signals.  In this list, signals are considered as instances of InputSig-
     nal, because they are only used to walk through the interconnection
     graph of the modules.
     The outSigList() methods yields a null terminated list of the output
     signals.  In this list, signals are considered as instances of Output-
     Signal, because they are only used to check if InputSignals are con-
     nected to them.
     The constructor compo() is used to build new instances of compo. The
     first argument is the name of the object. This name is mainly used for
     debugging purposes and is optional. The default value is the name of the
     module.  The second argument is optional and tells if we want the new
     object to register itself on the current clock and to be automaticaly
     scheduled by this clock.
     The copy constructor is declared private to prohibit its use. Copying
     synchronous objects has no meaning.


   The Module Description Language

     Mdl can be used to describe modules, to define composite modules and to
     define modules through inheritance. Comments begin at a '//' token and
     end at the end of the line.

     +    Module Description
          The description of a module in mdl gives its name and its inter-
          face. It also gives information about the source files for this
          module (oc, C++, Esterel ou mdl).
          The description of a module has the following structure:

          #source <sourceFile>       // Esterel or mdl source file
          #oc <ocSourceFile>         // If any
          #c++ <classInterfaceFile>  // C++ file that gives the
                                     // interface of the class
          #import <userDefsFile>     // definition file included
                                     // by the interface file
          #param {                   // Parameters of the module
            <typeOfParam1> <nameOfParam1>
            <typeOfParam2> <nameOfParam2>
               ...
          }

          <newType> is <oldType>;    // Definition of a new type

          <newType> as <oldType>;    // Informs mdlc that newType
                                     // is same as oldType

          <moduleName> {
               #nodep <sigName>, <sigName> ...;
               input: <pureSigName>, <sigType> <typedSigName> ...;
                                         // input signals
               output: <pureSigName>, <sigType> <typedSigName> ...;
                                         // output signals
          }

          The #source keyword introduces the name of the source file, if it
          is known.

          The #oc keyword introduces the name of the oc source file if the
          declaration was generated by occ++.

          The #c++ keyword introduces the name of the C++ file where the
          declaration of the class can be found.

          The #import keyword gives the name of a file containing user defin-
          itions used by the interface file. If this keyword is present, the
          class interface file #includes the user definition file <def>.H,
          and the code associated with those definitions should be in
          <def>.C.

          The #param keyword, if present, declares the parameters of the
          module. These parameters are gathered in a structure of type
          <moduleName>Consts, so only one argument is needed to parameterize
          a module.

          The 'is' keyword is used to define a new type as an equivalent for
          another type. It is very similar to the 'typedef' of C or C++. The
          statement '<newType> is <oldType>;' is translated into 'typedef
          <oldType> <newType>;' in C++.

          The 'as' keyword is used to inform mdlc that two types are
          equivalent.  The statement '<newType> as <oldType>;' has no coun-
          terpart in the C++ code, it is only used by mdlc. For example,
          integer is defined as int in the synchronous library (because Lus-
          tre and Esterel use integer for the C type "int"), but mdlc does
          not know about this. So, if you want mdlc to consider integer and
          int as the same type, you must write "integer as int;". You should
          not write "integer is int;" because mdlc should not define integer
          as int since it is already defined by libSync.

          The 'input:' , 'output:' and '#nodep' keywords can appear several
          times. 'input:' declares input signals, along with their type. If
          the type is missing, the signal is assumed to be pure. 'output:'
          does the same for output signals.

          The '#nodep' keyword introduces a list of input or output signals.
          The reaction of the module does not depend on the current value of
          the input signals of this list, and the value of the output signals
          of this list does not depend on the current reaction of the module.
          For example, in a delay, the output depends only on the previous
          reaction of the module. This can be stated with a #nodep statement.

          The name of the module can be qualified by a list of type names
          between angle brackets. The module is then generic and must be
          qualified by a list of real types when it is used.

          The declaration of the module may specify an argument to the con-
          structor. This argument is declared between parentheses between the
          name of the module and the opening brace. The special keyword
          #param between parentheses indicates that the constructor of the
          module takes a constant reference on a <moduleName>Const structure
          as its first argument. The name of this first argument to the con-
          structor is always 'initInstance'.
          For example, to describe a generic Delay module taking a pointer on
          its initial value as argument to its constructor, one can write:

               Delay<T>(const T*) {
                 input: T input;
                 #nodep output;
                 output: T output;
               }

          And then use a Delay for a signal of type 'int' in:

               X(const int) : {
                 input: int i;
                 output: int o;

                 Delay<int> d(&initInstance);
                 Y y;

                 d.input = i;
                 y.inp << d.output;
                 o = y.out;
               }

     +    Composite Module Definition
          The definition of a composite module in mdl gives its structure
          (which other modules are used to build the composite module) and
          the connections between the components.
          Signals are connected by the '<<' operator which connects an input
          signal to an output signal of the same type.
          A signal can be identified to another signal by the '=' operator.

          This operator specifies that a signal of the composite module 'is
          in fact' a signal of one of its components. To be identified, two
          signals must be of the same type and of the same nature (input or
          output). Moreover, the left side operand must be the signal that
          logically receives the information from the right side operand. So,
          to identify two input signals, the signal of the component is the
          left side operand and the signal of the composite module is the
          right side operand. For two output signals, the signal of the com-
          posite module is the left side operand and the signal of the com-
          ponent is the left side operand.
          The description of a composite module has the following structure:

          #Hinsert                   // Introduces a section of code to
                                     // insert in the .H file
          Anything you type here is not interpreted by mdlc
          and is merely copied to the .H file
          #endInsert                 // Ends the section of literal code

          #Cinsert                   // Introduces a section of code to
                                     // insert in the .C file
          Anything you type here is not interpreted by mdlc
          and is merely copied to the .C file
          #endInsert                 // Ends the section of literal code

          #use file                  // Tells mdlc that this definition uses
                                     // declarations found in file.mdf.
                                     // file.mdf is searched in the list
                                     // of directories specified by the -I
                                     // options and eventually in the
                                     // default directory.

          <moduleName> : {
               <signal declarations>        // Just like in a module descrip-
               tion

               <moduleType> <subModName>, <subModName> ...;
               <moduleType> <subModName>, <subModName> ...;
                 ...                       // Declaration of the components

               <subModName>.<InSigName> = <ModuleInSigName>;
               <ModuleOutSigName> = <subModName>.<OutSigName>;
                 ...                      // Signal identities

               <subModName>.<InSigName> << <subModName>.<OutSigName>;
                 ...                      // Signal connections
          }

          For example, to define an ABS_Brake from a Brake and an
          ABS_Controler, we write:

               ABS_Brake : {
                 input: integer pedal, skid;
                 output: integer pressure;

                 Brake regular_brake;
                 ABS_Controler abs_ctrl;

                 regular_brake.pedal = pedal;
                 abs_ctrl.skid = skid;
                 pressure = abs_ctrl.filtered_pressure;

                 abs_ctrl.raw_pressure << regular_brake.pressure;
               }

          assuming that the description of Brake and ABS_Controler is:

               Brake {
                 input: integer pedal;
                 output: integer pressure;
               }

               ABS_Controler {
                    input: skid, integer raw_pressure;
                    output: integer filtered_pressure;
               }

          Composite Modules can be defined as templates. The name of the
          module is then qualified by a list of type names between angle
          brackets. These type names can be used as real types inside the
          definition of the module. The C++ Synchronous Class created by mdlc
          for such a generic module will be a template with the same parame-
          ters as the module. For example, the following module:

               X<T,U> : {
                 input: T i;
                 output: U o;

                 Z<T> z;
                 H<int, U, float> h;

                 // Connections and equivalences
               }

          gives the class: template<class T, class U> class X { ... };
          through mdlc.

          Composite modules can take an additional first argument to their
          constructor. The type of this argument is declared, between
          parentheses, after the name of the module (including type parame-
          ters) and before the ':'. This argument is named 'initInstance' in
          the C++ code.

     +    Derived Module Description
          Mdl allows the description of a module that inherits its behaviour
          from another module.  The derived module inherits the signals and
          the behaviour of its super-module.  The description gives the addi-
          tional signals and the components used to filter the inherited
          behaviour.
          The description of a derived module in mdl has the following struc-
          ture:

               #Hinsert, #Cinsert and #use can be placed here, as for a Com-
               posite module.

               <moduleName> : <superModuleName> {
                    <additional signal declaration>
                    <signal identities>
                    <signal connections>
               }

          The signals of the inherited module can be accessed through the
          'super' pseudo-module name. This 'super' can always be omitted
          because the context always allows mdlc to determine if a signal
          name refers to a signal of the inherited module or to a signal of
          the derived module.

          For example, the ABS_Brake module can be defined as deriving from
          Brake:

               ABS_Brake : Brake {
                 input: skid;           // Just add skid to Brake

                 ABS_Controler abs_ctrl;// Add the filtering behaviour

                 abs_ctrl.skid = skid;
                 pressure = abs_ctrl.filtered_pressure;

                 abs_ctrl.raw_pressure << pressure;
               }

          In the 'pressure = abs_ctrl.filtered_pressure;' statement, pressure
          is the output signal of the derived module (because an output sig-
          nal on the left side of the '=' operator is a signal of the new
          module).
          In the 'abs_ctrl.raw_pressure << pressure;' statement, pressure is
          'super.pressure' because a signal can only be connected to a signal
          of a component.
          The behaviour of ABS_Brake defined as a derived module is the same
          as the behaviour of ABS_Brake defined as a composite module. The
          main difference is that the ABS_Brake C++ class is a subclass of
          Brake when ABS_Brake is defined as derived. This allows instances
          of Brake and ABS_Brake to be referenced through Brake pointers.

          Derived modules can be defined as templates just as composite
          modules, and they can take an additional first argument to their
          constructor too.

     +    Parameters to constructors

     A Synchronous Class must have a name (char*) and a scheduling attribute
     (boolean) as the two last arguments of its constructor. But there may be
     other arguments before those two. In such a case, you must give these
     arguments when you declare the submodule in a module definition.

     For example, if you use a Delay<int>, that takes an int* (initial value)
     as its first constructor argument, you should declare it as follows:

     Delay<int> d(0);

     to give a null pointer (no initial value). If you want to give an ini-
     tial value of 5, you must give a pointer on a variable whose value is
     five. For example:

     Delay<int> d(&initValue);

     The 'initValue' variable must be declared and defined, so you should
     have a #Cinsert section looking like:

     #Cinsert
     static const int initValue = 5;
     #endInsert

     The 'static' avoids pollution of the global name space. If you do not
     want to declare a variable to store the initial value, you can build a
     new int just to give it as the initial value:

     Delay<int> d(new int(5));

     If the initial value of the delay should be a parameter of your new
     module X, you should declare that X takes an argument to its construc-
     tor, and use this argument to initialize the delay:

          X(const int) : {
            Delay<int> d(&initValue);
               ...
          }

     If you need several arguments to the constructor of X, for instance if
     you have several delays to initialize with distinct values, you should
     use a #param structure:

          #param {
            int init1;
            float init2;
          }
          X(#param) : {
            Delay<int> d1(&(initValue.init1));
            Delay<float> d2(&(initValue.init2));
          }

     Mdlc does not check the type of the argument to the constructor. It just
     takes literally what you put between parentheses.

   Author

          Frederic Boulanger - Frederic.Boulanger@supelec.fr
          Supelec - Service Informatique
          Plateau de Moulon
          91192 Gif-sur-Yvette Cedex
          FRANCE

   See also
     occ++(l) , libSync(l)
