/*
//  File mdl.c
//
//  September 3 1992 - September 1995
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1995.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

/* This file contains the main entry point of mdlc.              */
/*****************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef THINK_C
#include <console.h>
#endif

#include "env.h"
#include "mdlErrors.h"
#include "mdlUtils.h"
#include "mdlTables.h"
#include "mdlCxx.h"
#include "mdlMdf.h"
#include "mdlParse.h"
#include "mdlProto.h"
#include "mdlSimul.h"

#define SYMBOLTOSTRING(s) #s
#define SYMTOSTR(s) SYMBOLTOSTRING(s)

char hFileName[256];           /* Name of output interface file */
char cFileName[256];           /* Name of output C++ file */
char mdfFileName[256];         /* Name of output mdf file */
char baseName[256] = "";       /* Base name of output files */
char mdfBaseName[256] = "";    /* Base name of mdf file */
char *sourceFile = NULL;       /* Name of source file */
char *optIncludeEsterel = 0;   /* Esterel Include file given by -inc */
char *optCxxInstanciate = 0;   /* cxx instanciation file given by -cxxinst */

int verbose = 0;               /* Verbose flag. Default is not verbose */
int allowDebug = 0;            /* Stuff C++ code with debug prompts on cerr */
int createMdf = 0;             /* Create mdf file for built module */
int forceReaction = 0;         /* Force reaction of unused modules for their side effects */
int buildBackbone = 0;         /* Build the backbone of a Synchronous Class */
int buildSimulText = 0;        /* Build text oriented simulation program */
int buildSimulGraph = 0;       /* Build graphic simulation program */
int buildSimulProto = 0;       /* Build protocol simulation program */
nameList *incPathes = 0;       /* List of include pathes */
char *defIncPath = defaultIncPath; /* Default include path */

/***************************************************************************/
/* Function yywrap is called by the lex input function when it sees EOF.   */
/* It must return 1 if this is really the end of source file, 0 if not.    */
/* We always return 1 because our input is a single file without includes. */
/***************************************************************************/
int yywrap() {
  return 1;
}


/******************************************************************************/
/* Main entry point of occxx. Parses the argument list to find the oc source  */
/* file to compile and options. Note that only one source file is allowed.    */
/******************************************************************************/
int main(int argc, char **argv){
  FILE *mdfSrc;                          /* current mdl source file */
  char errStr[256];                      /* Buffer for error messages */
  char *fName;                           /* For #use file name */
  nameList *curUse, *curPath;
  int i, parseOk, ok, somethingToBuild;
  
/*  yydebug = 1; */                      /* Uncomment to debug the yacc parser */
  
#ifdef THINK_C
  argc = ccommand(&argv);                /* Stand alone THINK C application */
#endif
  i = 1;
  while (i < argc) {                     /* Decipher the command line */
    switch (argv[i][0]) {
      case '-':                          /* If argument is an option */
        if (strcmp(argv[i],"-v") == 0) { /* -v : Verbose option */
          verbose = 1;
          i++;
        }else if (strcmp(argv[i],"-version") == 0) {
          printf("--- mdlc: version %s, %s\n",mdlVers,copyrightStr);
          i++;                           /* -version : Give version option */
          exit(0);
        }else if (strcmp(argv[i],"-access") == 0) {
          printf("--- mdlc: free distribution, %s\n", copyrightStr);
          i++;                           /* -access : Give access rights */
          exit(0);
        }else if (strcmp(argv[i],"-info") == 0) {
          printf("--- mdlc:  version %s\n", mdlVers);
          printf("         released: September 11 1995\n");
          printf("          formats: oc2,oc3,oc5 --> C++\n");
          printf("      compiled on: %s/%s\n",SYMTOSTR(ARCHI), SYMTOSTR(SYSTEM));
          printf("             with: %s\n",SYMTOSTR(C_COMPILER));
          printf("             date: %s\n",SYMTOSTR(THE_DATE));
          printf("               by: %s\n",SYMTOSTR(THE_GUY));
          i++;                           /* -info : Give info on mdlc */
          exit(0);
        }else if (strcmp(argv[i], "-B") == 0) {
          i++;                           /* -B myBaseName : Set basename option */
          if (argc <= i)
            usageError("base name expected after -B option.");
          strcpy(baseName,argv[i]);
          i++;
        }else if (strcmp(argv[i], "-debug") == 0) {
          i++;
          allowDebug = 1;
        }else if (strcmp(argv[i], "-mdf") == 0) {
          i++;
          createMdf = 1;
        }else if (strcmp(argv[i], "-mdfB") == 0) {
          i++;
          if (argc <= i)
            usageError("base name expected after -mdfB option.");
          strcpy(mdfBaseName, argv[i]);
          i++;
          createMdf = 1;
        }else if (strcmp(argv[i], "-force") == 0) {
          i++;
          forceReaction = 1;
        }else if (strcmp(argv[i], "-proto") == 0) {
          i++;
          buildBackbone = 1;
        }else if (strcmp(argv[i], "-inc") == 0) {
          i++;
          if (argc <= i)
            usageError("include file name expected after -inc option.");
          if ((argv[i][0] != '"') && (argv[i][0] != '<')) {
            optIncludeEsterel = checkAlloc(strlen(argv[i]) + 3, sizeof(char));
            sprintf(optIncludeEsterel, "\"%s\"", argv[i]);
          }else {
            optIncludeEsterel = checkAlloc(strlen(argv[i]) + 1, sizeof(char));
            strcpy(optIncludeEsterel, argv[i]);
          }
          i++;
        }else if (strncmp(argv[i], "-I",2) == 0) {
          incPathes = addName(incPathes,argv[i]+2);
          i++;
        }else if (strcmp(argv[i], "-simtext") == 0) {
          buildSimulText = 1;
          i++;
        }else if (strcmp(argv[i], "-simX") == 0) {
          buildSimulGraph = 1;
          i++;
        }else if (strcmp(argv[i], "-simP") == 0) {
        	buildSimulProto = 1;
        	i++;
        }else if (strcmp(argv[i], "-type") == 0) {
		  i++;
		  if (argc <= i)
		  	usageError("instanciation type list expected after -type.");
          instTypesStr = argv[i];
          i++;
        }else if (strcmp(argv[i], "-cxxinst") == 0) {
		  i++;
		  if (argc <= i)
		  	usageError("cxx instanciation file name expected after -cxxinst.");
          optCxxInstanciate = argv[i];
          i++;
        }else {
          sprintf(errStr,"Unknown option %s",argv[i]);
          i++;
          usageError(errStr);
        }
        break;
      default:
        sourceFile = argv[i];
        i++;
        if ((mdfSrc = fopen(sourceFile,"r")) == NULL) {
          sprintf(errStr,"Can't open file %s for input.",sourceFile);
          fileError(errStr);
        }
        if (verbose)
          printf("  Parsing \"%s\" ...\n\n", sourceFile);
        yyin = mdfSrc;                            /* Let yylex() read from source file */
        lineNb = 1;                               /* New file starts at line 1 */
        reinitlex();                              /* Tell flex its a new file */
        parseOk = yyparse();                      /* Let Yacc do the job */
        fclose(mdfSrc);                           /* and close the source file */
        if (parseOk)
          exit(parseOk);
        if (verbose) {
          printLastModule();
          printf("\n  ... done.\n\n");
        }
        break;
    }
  }
  mdfSrc = 0;
  if (useList != 0) {
    if ((fName = getenv("MDLC_INC_PATH")) != 0)
      defIncPath = fName;
    curUse = useList;
    while (curUse != 0) {
      if (incPathes != 0) {
        curPath = incPathes;
        while (curPath != 0) {
          fName = (char*)checkAlloc(1, strlen(curUse->name)
                                       + strlen(curPath->name)
                                       + 6);
          sprintf(fName,"%s/%s.mdf", curPath->name, curUse->name);
          mdfSrc = fopen(fName,"r");
          if (mdfSrc == 0) {
            free(fName);
            curPath = curPath->next;
          }else
            break;
        }
      }
      if (mdfSrc == 0) {
        fName = (char*)checkAlloc(1, strlen(curUse->name)
                                     + strlen(defIncPath)
                                     + 6);
        sprintf(fName,"%s/%s.mdf", defIncPath, curUse->name);
        mdfSrc = fopen(fName,"r");
      }
      if (mdfSrc == 0) {
        free(fName);
        sprintf(errStr,"Can't open definition file %s.mdf for input.", curUse->name);
        fileError(errStr);
      }else {
        if (verbose)
          printf("  Parsing \"%s\" ...\n\n", fName);
        yyin = mdfSrc;
        sourceFile = fName;
        lineNb = 1;
        reinitlex();
        parseOk = yyparse();
        fclose(mdfSrc);
        mdfSrc = 0;
        if (parseOk != 0)
          exit(parseOk);
        if (verbose) {
          printLastModule();
          printf("\n ... done.\n\n");
        }
        free(fName);
      }
      curUse = curUse->next;
    }
  }
  somethingToBuild = buildModuleTable();
  ok = 0;
  if (buildBackbone == 1) {
  	if (somethingToBuild == 0)
  		mdfWarn("-proto option with a module definition - ignored");
		else
      buildProtoFiles();
		ok = 1;
  }
  if (buildSimulText == 1) {
  	if (somethingToBuild == 0)
	 	  mdfWarn("-simtext option with a module definition - ignored");
		else
	 	  buildTextSimul();
		ok = 1;
  }
  if (buildSimulGraph == 1) {
  	if (somethingToBuild == 0)
	  	mdfWarn("-simX option with a module definition - ignored");
		else
	  	buildXSimul();
		ok = 1;
  }
  if (buildSimulProto == 1) {
  	if (somethingToBuild == 0)
	  	mdfWarn("-simP option with a module definition - ignored");
		else
	  	buildPSimul();
		ok = 1;
  }
  if (somethingToBuild == 0) {
    linkModule();
    buildHModule();
    buildCModule();
    if (createMdf)
      writeMdfFile();
  } else if (ok == 0)
    fatalError("No module to build!");
  return 0;
}
