/*
//  File mdlCxx.c
//
//  September 3 1992 - September 1996
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1996.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

#include "mdlTables.h"
#include "mdlCxx.h"
#include "mdl.h"
#include "mdlUtils.h"
#include "mdlErrors.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

int nbModules;
int builtModule;
bMod* linkedModule;
pMod* moduleTab;

/* Static strings for C++ include file generation */
/**/
static char hHeader[] = "//\n//  %s\n//  Created by mdlc v%s on %s from mdl source file \"%s\"\n\
//\n\n#include %s\n\n";
static char hModHead[] = "\n//\n//  Interface of module %s\n//\n\n";
static char hStandard[] = "  public:\n    void resetOutputs();\n\
    void resetInputs();\n    void setInputs();\n    void activate();\n";
static char hEndDecl[] = "    virtual const InputSignal** inSigList() const;\n\
    virtual const OutputSignal** outSigList() const;\n\n";

/* Static strings for C++ file generation */
/**/
static char cHeader[] = "//\n//  %s\n//  Created by mdlc v%s on %s from mdl source file \"%s\"\n\
//\n\n#include \"%s\"\n\n//  Code of module %s\n//\n\n";


void buildHBody(FILE *hFile);
void buildCBody(FILE *cFile);
#ifdef DEAD_CODE
void buildAssignOperator(FILE *cFile);
#endif
void buildBackup(FILE *cFile);
void buildModTabs(pMod* newMod, module* rawMod);
void updateDerivedMods();
int findModule(atype* name);
void linkSigs(int modId, int conId, bMod* lkMod);
void equivSigs(int modId, int equId, bMod*lkMod);
void equivSuperSigs(int modId, bMod* lkMod);
void markSubModules();
int markModulesFrom(int mNum, int sNum, int firstMark, int noDep);
void checkType(int modId, int mNumF, int sNumF, int mNumT, int sNumT);
void checkConnect(int modId, int conId, int mNumF, int sNumF, int mNumT, int sNumT, bMod* lkMod);
void checkEquiv(int modId, int equId, int mNumF, int sNumF, int mNumT, int sNumT, bMod* lkMod);
void resolveSig(int* modNum, int* sigNum, int modId, sig* theSig, int isEquiv);
int findSignal(int *sigNum, int modTyp, char* name);
int findSubMod(int *modNum, int modId, char *name);


void buildHModule() {
  FILE *hFile;
  int i, j;
  char *incName;
  char *ifdefName;
  int doInc;
  nameList *curLine;
  char *nameOfType_module;

  nameOfType_module = NameOfType(moduleTab[builtModule].name);
  
  if (strlen(baseName) == 0)
    strcpy(baseName, moduleTab[builtModule].name->name);
  sprintf(hFileName,hFileTemplate,"",baseName);
  if ((hFile = fopen(hFileName,"w")) == 0) {
    sprintf(tmpStr,"Cannot open file %s for output", hFileName);
    fileError(tmpStr);
  }
  ifdefName = strrchr(hFileName, dirSep);
  if (ifdefName == 0) {
    ifdefName = checkAlloc(strlen(hFileName) + 3, sizeof(char));
    j = 0;
  }else {
    ifdefName++;
    j = ifdefName - hFileName;
    ifdefName = checkAlloc(strlen(ifdefName) + 3, sizeof(char));
  }
  ifdefName[0] = '_';
  for(i = j; i < strlen(hFileName); i++) {
    if (hFileName[i] == '.')
      ifdefName[i-j+1] = '_';
    else
      ifdefName[i-j+1] = toupper(hFileName[i]);
  }
  ifdefName[++i-j] = '_';
  ifdefName[++i-j] = 0;
  fprintf(hFile,"#ifndef %s\n", ifdefName);
  fprintf(hFile,"#define %s\n\n", ifdefName);
  free(ifdefName);
  
  if (optIncludeEsterel == 0)
    fprintf(hFile,hHeader,hFileName, mdlVers, dateAndTime(),
            moduleTab[builtModule].source, includeEsterel);
  else
    fprintf(hFile,hHeader,hFileName, mdlVers, dateAndTime(),
            moduleTab[builtModule].source, optIncludeEsterel);

  for (i = 0 ; i < moduleTab[builtModule].nbSubModules; i++) { /* Include interface of used modules */
    incName = moduleTab[moduleTab[builtModule].modules[i].typeMod].cpp;
    doInc = 1;
    for (j = 0 ; j < i; j++)               /* Test for already included file */
      if (strcmp(incName, moduleTab[moduleTab[builtModule].modules[j].typeMod].cpp) == 0) {
        doInc = 0;
        break;
      }
    if ((doInc)&&(strlen(incName)!=0))
      fprintf(hFile,"#include \"%s\"\n", incName);
  }
  fprintf(hFile,hModHead,nameOfType_module);

  if (HInsertData != 0) {
    curLine = HInsertData;
    while(curLine != 0) {
      fprintf(hFile,"%s",curLine->name);
      curLine = curLine->next;
    }
    fprintf(hFile,"\n");
  }

  buildHBody(hFile);
  
  if (strcmp(nameOfType_module, moduleTab[builtModule].name->name) != 0) {
    fprintf(hFile, "\n#ifdef __GNUG__\n#include \"");
    fprintf(hFile, cFileTemplate, "", baseName);
    fprintf(hFile, "\"\n#endif\n");
  }
  
  fprintf(hFile,"\n#endif\n");
  fclose(hFile);
  free(nameOfType_module);
}

void buildHBody(FILE *hFile) {
  int i, nb, m, s;
  typeList *curType;
  typdefList *tdef;
  paramList *curParm;
  char *nameOfType_module;
  char *nt, *ot;
  char *st, *sigt, *modt;
  
  nameOfType_module = NameOfType(moduleTab[builtModule].name);

  tdef = theTypdefs;
  while(tdef != 0) {
    if (tdef->theDef->mustDef) {
    	ot = NameOfType(tdef->theDef->oldType);
    	nt = NameOfType(tdef->theDef->newType);
      fprintf(hFile,"typedef %s %s;\n\n", ot, nt);
      free(ot);
      free(nt);
     }
    tdef = tdef->next;
  }

  if (moduleTab[builtModule].params != 0) {
    fprintf(hFile,"struct %sConsts {\n", moduleTab[builtModule].name->name);
    curParm = moduleTab[builtModule].params;
    while(curParm != 0) {
      nt = NameOfType(curParm->val->theType);
      fprintf(hFile, "  %s %s;\n",nt,curParm->val->name);
      free(nt);
      curParm = curParm->next;
    }
    fprintf(hFile,"};\n\n");
  }
  
  nb = moduleTab[builtModule].nbInputs;
  curType = moduleTab[builtModule].name->params;
  printTemplate(hFile, curType);
  if (moduleTab[builtModule].kind == mComposite)
    fprintf(hFile,"class %s : public Synchronous, public Composite {\n  private:\n",
            moduleTab[builtModule].name->name);
  else {
    st = NameOfType(moduleTab[builtModule].super);
    fprintf(hFile,"class %s : public %s, public Composite {\n  private:\n",
                  moduleTab[builtModule].name->name, st);
    free(st);
  }
  for (i = 0; i < nb; i++) {                   /* Declare unassigned input signals */
    m = linkedModule->inputs[i].modNum;
    if (m < -1)
      m = -2 - m;
    s = linkedModule->inputs[i].sigNum;
    if ((m == -1) && (s == (-1 - i))) {
      sprintf(tmpStr,"Input signal %s is not connected to any submodule signal",
                     moduleTab[builtModule].inputs[i]->name);
      mdfWarn(tmpStr);
      if (moduleTab[builtModule].inputs[i]->kind == sPure)
        fprintf(hFile,"    PureInSignal f%s;\n",moduleTab[builtModule].inputs[i]->name);
      else {
      	sigt = NameOfType(moduleTab[builtModule].inputs[i]->theType);
        fprintf(hFile,"    InSignal<%s> f%s;\n", sigt,
                moduleTab[builtModule].inputs[i]->name);
        free(sigt);
      }
    }
  }
  if (moduleTab[builtModule].kind == mDerived) {
  	i = nb;
  	nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;
  	for ( ; i < nb; i++) {
  		m = linkedModule->inputs[i].modNum;
      if (m == -1) {
        if (moduleTab[builtModule].inputs[i]->kind == sPure)
          fprintf(hFile,"    PureInSignal ");
        else
          fprintf(hFile,"    InSignal<%s> ",
                  resolveTemplate(moduleTab[builtModule].inputs[i]->theType,
                                  moduleTab[builtModule].super,
                                  moduleTab[moduleTab[builtModule].superMod].name));
        fprintf(hFile,"f%s;\n", moduleTab[builtModule].inputs[i]->name);
      }
  	}
  }
  
  nb = moduleTab[builtModule].nbOutputs;
  for (i = 0; i < nb; i++) {                 /* Declare unassigned output signals */
    m = linkedModule->outputs[i].modNum;
    s = linkedModule->outputs[i].sigNum;
    if (((m == -1) && (s == i))
        || (linkedModule->subSigs[m].outputs[s].mark != 0)) {/* Or assigned outputs used internaly */
      if (moduleTab[builtModule].outputs[i]->kind == sPure)
        fprintf(hFile,"    PureOutSignal f%s;\n",moduleTab[builtModule].outputs[i]->name);
      else {
      	sigt = NameOfType(moduleTab[builtModule].outputs[i]->theType);
        fprintf(hFile,"    OutSignal<%s> f%s;\n", sigt,
                moduleTab[builtModule].outputs[i]->name);
        free(sigt);
      }
    }
  }
  if (moduleTab[builtModule].kind == mDerived) {
  	i = nb;
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
    for ( ; i < nb; i++) {
      m = linkedModule->outputs[i].modNum;
      s = linkedModule->outputs[i].sigNum;
      if (linkedModule->subSigs[m].outputs[s].mark != 0) {
        if (moduleTab[builtModule].outputs[i]->kind == sPure)
          fprintf(hFile,"    PureOutSignal ");
        else
          fprintf(hFile,"    OutSignal<%s> ",
                  resolveTemplate(moduleTab[builtModule].outputs[i]->theType,
                                  moduleTab[builtModule].super,
                                  moduleTab[moduleTab[builtModule].superMod].name));
        fprintf(hFile,"f%s;\n",
                  moduleTab[builtModule].outputs[i]->name);
      }
    }
  }
  
  if (moduleTab[builtModule].kind == mComposite)
    i = 0;
  else
    i = 1;
  for ( ; i < moduleTab[builtModule].nbSubModules; i++) { /* Declare submodules */
    modt = NameOfType(moduleTab[builtModule].modules[i].theType);
    fprintf(hFile, "    %s %s;\n",modt, moduleTab[builtModule].modules[i].name);
    free(modt);
  }
  
  if (moduleTab[builtModule].argsToConst != 0) {
    if (strcmp(moduleTab[builtModule].argsToConst->name, "#param") == 0) {
      fprintf(hFile,"const %sConsts fInitInstance;\n",
                     moduleTab[builtModule].name->name);
    } else {
      fprintf(hFile,"%s fInitInstance;\n",
                    NameOfType(moduleTab[builtModule].argsToConst));
    }
  }
  
#ifdef DEAD_CODE
  fprintf(hFile,"    %s *pf_backup_copy_;\n",nameOfType_module);
#endif
  fprintf(hFile,hStandard);         /* Print standard method declarations */
  for (i = 0; i < moduleTab[builtModule].nbInputs; i++) {
    if (moduleTab[builtModule].inputs[i]->kind == sPure)
      fprintf(hFile, "    virtual PureInSignal& %s();\n", moduleTab[builtModule].inputs[i]->name);
    else {
    	sigt = NameOfType(moduleTab[builtModule].inputs[i]->theType);
      fprintf(hFile, "    virtual InSignal<%s>& %s();\n", sigt,
              moduleTab[builtModule].inputs[i]->name);
      free(sigt);
    }
  }
  for (i = 0; i < moduleTab[builtModule].nbOutputs; i++) {
    if (moduleTab[builtModule].outputs[i]->kind == sPure)
      fprintf(hFile, "    virtual PureOutSignal& %s();\n", moduleTab[builtModule].outputs[i]->name);
    else {
    	sigt = NameOfType(moduleTab[builtModule].outputs[i]->theType);
      fprintf(hFile, "    virtual OutSignal<%s>& %s();\n", sigt,
              moduleTab[builtModule].outputs[i]->name);
      free(sigt);
    }
  }
  if (moduleTab[builtModule].kind == mDerived) {
    nb = moduleTab[moduleTab[builtModule].superMod].nbInputs;
    for (i = moduleTab[builtModule].nbInputs; i < nb+moduleTab[builtModule].nbInputs; i++) {
      if ((linkedModule->inputs[i].modNum != 0)
          && (linkedModule->inputs[i].modNum != -2)) {  /* Inherited inputs not connected to 'super' */
        if (moduleTab[builtModule].inputs[i]->kind == sPure)
          fprintf(hFile, "    virtual PureInSignal& %s();\n", moduleTab[builtModule].inputs[i]->name);
        else
          fprintf(hFile, "    virtual InSignal<%s>& %s();\n",
                  resolveTemplate(moduleTab[builtModule].inputs[i]->theType,
                                  moduleTab[builtModule].super,
                                  moduleTab[moduleTab[builtModule].superMod].name),
                  moduleTab[builtModule].inputs[i]->name);
      }
    }
    nb = moduleTab[moduleTab[builtModule].superMod].nbOutputs;
    for (i = moduleTab[builtModule].nbOutputs; i < nb+moduleTab[builtModule].nbOutputs; i++) {
      m = linkedModule->outputs[i].modNum;
      s = linkedModule->outputs[i].sigNum;
      if ((m != 0) || (linkedModule->subSigs[m].outputs[s].mark != 0)) {
        /* Inherited outputs not connected to 'super' or used internaly */
        if (moduleTab[builtModule].outputs[i]->kind == sPure)
          fprintf(hFile, "    virtual PureOutSignal& %s();\n", moduleTab[builtModule].outputs[i]->name);
        else
          fprintf(hFile, "    virtual OutSignal<%s>& %s();\n",
                  resolveTemplate(moduleTab[builtModule].outputs[i]->theType,
                                  moduleTab[builtModule].super,
                                  moduleTab[moduleTab[builtModule].superMod].name),
                  moduleTab[builtModule].outputs[i]->name);
      }
    }
  }
  fprintf(hFile, "    virtual void backup();\n    virtual void restore();\n");
#ifdef DEAD_CODE
  fprintf(hFile, "    %s& operator=(const %s &);\n",
                      nameOfType_module,
                                          nameOfType_module);
#endif
  fprintf(hFile,hEndDecl);
  fprintf(hFile,"    %s(", moduleTab[builtModule].name->name);
  
  if (moduleTab[builtModule].argsToConst != 0) {
    if (strcmp(moduleTab[builtModule].argsToConst->name, "#param") == 0) {
      fprintf(hFile,"const %sConsts& initInstance, ",
                     moduleTab[builtModule].name->name);
    } else {
      fprintf(hFile,"%s initInstance, ",
                    NameOfType(moduleTab[builtModule].argsToConst));
    }
  }
  fprintf(hFile, "const char* name = \"%s\", const bool schedule = true);\n",
						      moduleTab[builtModule].name->name);
  fprintf(hFile,"    ~%s();\n", moduleTab[builtModule].name->name);
  fprintf(hFile, "  private:\n    %s(const %s &);\n",
                 moduleTab[builtModule].name->name,
                 nameOfType_module);
  fprintf(hFile,"};\n");

  free(nameOfType_module);
}

void buildCModule() {
  FILE *cFile;
  nameList *curLine;
  char *ifdefName;
  int i, j;

  if (strlen(baseName) == 0)
    strcpy(baseName, moduleTab[builtModule].name->name);
  sprintf(cFileName,cFileTemplate,"",baseName);
  if (strlen(hFileName) == 0)
    sprintf(hFileName,hFileTemplate,"",baseName);
  if ((cFile = fopen(cFileName,"w")) == 0) {
    sprintf(tmpStr,"Cannot open file %s for output", cFileName);
    fileError(tmpStr);
  }
  
  ifdefName = strrchr(cFileName, dirSep);
  if (ifdefName == 0) {
    ifdefName = checkAlloc(strlen(cFileName) + 3, sizeof(char));
    j = 0;
  }else {
    ifdefName++;
    j = ifdefName - cFileName;
    ifdefName = checkAlloc(strlen(ifdefName) + 3, sizeof(char));
  }
  ifdefName[0] = '_';
  for(i = j; i < strlen(cFileName); i++) {
    if (cFileName[i] == '.')
      ifdefName[i-j+1] = '_';
    else
      ifdefName[i-j+1] = toupper(cFileName[i]);
  }
  ifdefName[++i-j] = '_';
  ifdefName[++i-j] = 0;
  fprintf(cFile,"#ifndef %s\n", ifdefName);
  fprintf(cFile,"#define %s\n\n", ifdefName);
  free(ifdefName);
  
  fprintf(cFile,cHeader,cFileName, mdlVers, dateAndTime(),
          moduleTab[builtModule].source, hFileName, moduleTab[builtModule].name->name);

  if (CInsertData != 0) {
    curLine = CInsertData;
    while(curLine != 0) {
      fprintf(cFile,"%s",curLine->name);
      curLine = curLine->next;
    }
    fprintf(cFile,"\n");
  }
  buildCBody(cFile);
  fprintf(cFile,"\n#endif\n");
  
  fclose(cFile);
}

void buildCBody(FILE *cFile) {
  int i,j;
  int m,s;
  int nb;
  pMod *typMod;
  typeList *curType;

  char *nameOfType_module;
  char *nameOfType_super;
  char *sigt;
  
  nameOfType_module = NameOfType(moduleTab[builtModule].name);
  if (moduleTab[builtModule].kind == mDerived)
    nameOfType_super = NameOfType(moduleTab[builtModule].super);

  curType = moduleTab[builtModule].name->params;
  printTemplate(cFile, curType);               /* Code for resetOutputs() */
  fprintf(cFile,"void %s::resetOutputs() {\n",nameOfType_module);
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \".resetOutputs()\" << endl;\n");
    fprintf(cFile,"#endif\n");
  }

  for(i = 0; i < moduleTab[builtModule].nbSubModules; i++) {     /* Reset outputs of submodules */
    if ((moduleTab[builtModule].kind == mDerived) && (i == 0)) { /* Special case of super-module */
      if (allowDebug) {
        fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
        fprintf(cFile,"  cerr << name() << \": %s::resetOutputs()\" << endl;\n",
                nameOfType_super);
        fprintf(cFile,"#endif\n");
      }
      fprintf(cFile, "  %s::resetOutputs();\n",
              nameOfType_super);
    } else {                                                     /* Regular submodule */
      if (allowDebug) {
        fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
        fprintf(cFile,"  cerr << name() << \": %s.resetOutputs()\" << endl;\n",
                moduleTab[builtModule].modules[i].name);
        fprintf(cFile,"#endif\n");
      }
      fprintf(cFile, "  %s.resetOutputs();\n",
              moduleTab[builtModule].modules[i].name);
    }
  }

  nb = moduleTab[builtModule].nbOutputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  for (i = 0; i < nb; i++) {
    m = linkedModule->outputs[i].modNum;
    s = linkedModule->outputs[i].sigNum;
    if ((m != -1) && (linkedModule->subSigs[m].outputs[s].mark != 0)) {
      if (linkedModule->subSigs[m].outputs[s].deps == 0) {   /* Update nodep output signals */
        if ((moduleTab[builtModule].kind == mDerived) && (m == 0))
          fprintf(cFile,"  %s().value() = %s::%s().value();\n",
                  moduleTab[builtModule].outputs[i]->name,
                  nameOfType_super,
                  moduleTab[moduleTab[builtModule].superMod].outputs[s]->name);
        else
          fprintf(cFile,"  %s().value() = %s.%s().value();\n",
                  moduleTab[builtModule].outputs[i]->name,
                  moduleTab[builtModule].modules[m].name,
                  moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name);
      }else
        fprintf(cFile,"  %s().value().reset();\n",moduleTab[builtModule].outputs[i]->name);
    }else if (m == -1)
      fprintf(cFile,"  %s().value().reset();\n",moduleTab[builtModule].outputs[i]->name);
  }
  fprintf(cFile, "}\n\n");

  printTemplate(cFile, curType);             /* Code for setInputs() */
  fprintf(cFile,"void %s::setInputs() {\n", nameOfType_module);
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \".setInputs()\" << endl;\n");
    fprintf(cFile,"#endif\n");
  }
  nb = moduleTab[builtModule].nbInputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;
  for (i = 0; i < nb; i++) {                 /* Un-assigned input signals */
    if (linkedModule->inputs[i].modNum == -1) {
      fprintf(cFile,"//  f%s.update();\n", moduleTab[builtModule].inputs[i]->name);
      if (allowDebug) {
        fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
        fprintf(cFile,"#ifndef __GNUG__\n");
        fprintf(cFile,"//  cerr << f%s << \"=\" << f%s.value() << endl;\n",
                      moduleTab[builtModule].inputs[i]->name,
                      moduleTab[builtModule].inputs[i]->name);
        fprintf(cFile,"#else\n// GNU g++ 2.5.8 requires ::operator<<(...) instead of ..<<.. for templates\n");
        fprintf(cFile,"//  ::operator<<(cerr, f%s) << \"=\";\n",
                      moduleTab[builtModule].inputs[i]->name);
        fprintf(cFile,"//  ::operator<<(cerr, f%s.value()) << endl;\n",
                      moduleTab[builtModule].inputs[i]->name);
        fprintf(cFile,"#endif\n");
        fprintf(cFile,"#endif\n");
      }
    }
  }
  fprintf(cFile,"}\n\n");

  printTemplate(cFile, curType);             /* Code for resetInputs() */
  fprintf(cFile,"void %s::resetInputs() {\n", nameOfType_module);
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \".resetInputs()\" << endl;\n");
    fprintf(cFile,"#endif\n");
  }
  /* Reset inputs of submodules */
  for(i = 0; i < moduleTab[builtModule].nbSubModules; i++) {     /* Reset outputs of submodules */
    if ((moduleTab[builtModule].kind == mDerived) && (i == 0)) { /* Special case of super-module */
      if (allowDebug) {
        fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
        fprintf(cFile,"  cerr << name() << \": %s::resetInputs()\" << endl;\n",
                nameOfType_super);
        fprintf(cFile,"#endif\n");
      }
      fprintf(cFile, "  %s::resetInputs();\n",
              nameOfType_super);
    } else {                                                     /* Regular submodule */
      if (allowDebug) {
        fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
        fprintf(cFile,"  cerr << name() << \": %s.resetInputs()\" << endl;\n",
                moduleTab[builtModule].modules[i].name);
        fprintf(cFile,"#endif\n");
      }
      fprintf(cFile, "  %s.resetInputs();\n",
              moduleTab[builtModule].modules[i].name);
    }
  }

  nb = moduleTab[builtModule].nbInputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;
  for (i = 0; i < nb; i++) {                 /* Un-assigned input signals */
    if (linkedModule->inputs[i].modNum == -1) {
      fprintf(cFile,"    f%s.reset();\n", moduleTab[builtModule].inputs[i]->name);
    }
  }
  fprintf(cFile,"}\n\n");

  printTemplate(cFile, curType);
  fprintf(cFile,"void %s::activate() {\n",
          nameOfType_module);/* Activate method */
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \".activate()\" << endl;\n");
    fprintf(cFile,"#endif\n");
  }
  for (i = moduleTab[builtModule].reactSteps; i >= 0; i--) {           /* For each reaction step */
    for (j = 0; j < moduleTab[builtModule].nbSubModules; j++) {
      if (linkedModule->subSigs[j].rStep == i)                         /* Make modules of this step react */
        if ((moduleTab[builtModule].kind == mDerived)&&(j == 0)) {
          if (allowDebug) {
            fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
            fprintf(cFile,"  cerr << name() << \": %s::setInputs()\" << endl;\n",
                          nameOfType_super);
            fprintf(cFile,"#endif\n");
          }
          fprintf(cFile, "  %s::setInputs();\n", nameOfType_super);
          if (allowDebug) {
            fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
            fprintf(cFile,"  cerr << name() << \": %s::activate()\" << endl;\n",
                          nameOfType_super);
            fprintf(cFile,"#endif\n");
          }
          fprintf(cFile, "  %s::activate();\n",
                  nameOfType_super);
        }else {
          if (allowDebug) {
            fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
            fprintf(cFile,"  cerr << name() << \": %s.react()\" << endl;\n",
                          moduleTab[builtModule].modules[j].name);
            fprintf(cFile,"#endif\n");
          }
          fprintf(cFile, "  %s.react();\n", moduleTab[builtModule].modules[j].name);
        }
    }
  }
  nb = moduleTab[builtModule].nbOutputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  for (i = 0; i < nb; i++) {   /* Update assigned output signals used internaly */
    m = linkedModule->outputs[i].modNum;
    s = linkedModule->outputs[i].sigNum;
    if ((m != -1) && (linkedModule->subSigs[m].outputs[s].mark != 0)) {
      if (linkedModule->subSigs[m].outputs[s].deps == 1) {
        if ((moduleTab[builtModule].kind == mDerived) && (m == 0))
          fprintf(cFile,"  %s().value() = %s::%s().value();\n",
                  moduleTab[builtModule].outputs[i]->name,
                  nameOfType_super,
                  moduleTab[moduleTab[builtModule].superMod].outputs[s]->name);
        else
          fprintf(cFile,"  %s().value() = %s.%s().value();\n",
                  moduleTab[builtModule].outputs[i]->name,
                  moduleTab[builtModule].modules[m].name,
                  moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name);
      }
    }
  }
  fprintf(cFile,"}\n\n");
  
  nb = moduleTab[builtModule].nbInputs;
  for (i = 0; i < nb; i++) {           /* Inputs signal access methods */
    printTemplate(cFile, curType);
    if (moduleTab[builtModule].inputs[i]->kind == sPure)
      fprintf(cFile,"PureInSignal& %s::%s() {\n  return ",
              nameOfType_module,
              moduleTab[builtModule].inputs[i]->name);
    else {
    	sigt = NameOfType(moduleTab[builtModule].inputs[i]->theType);
      fprintf(cFile,"InSignal<%s>& %s::%s() {\n  return ", sigt,
              nameOfType_module,
              moduleTab[builtModule].inputs[i]->name);
      free(sigt);
    }
    if ((linkedModule->inputs[i].modNum == -1)&&(linkedModule->inputs[i].sigNum == (-1 - i)))
      fprintf(cFile,"f%s",moduleTab[builtModule].inputs[i]->name);
    else {
      m = linkedModule->inputs[i].modNum;
      if (m < -1)
        m = -2 - m;
      fprintf(cFile,"%s.%s()",moduleTab[builtModule].modules[m].name,
                            moduleTab[
                              moduleTab[builtModule].modules[m].typeMod
                            ].inputs[-1 - linkedModule->inputs[i].sigNum]->name);
    }
    fprintf(cFile,";\n}\n\n");
  }
  
  nb = moduleTab[builtModule].nbOutputs;
  for (i = 0; i < nb; i++) {            /* Output signal access methods */
    printTemplate(cFile, curType);
    if (moduleTab[builtModule].outputs[i]->kind == sPure)
      fprintf(cFile,"PureOutSignal& %s::%s() {\n  return ",
              nameOfType_module,
              moduleTab[builtModule].outputs[i]->name);
    else {
    	sigt = NameOfType(moduleTab[builtModule].outputs[i]->theType);
      fprintf(cFile,"OutSignal<%s>& %s::%s() {\n  return ", sigt,
              nameOfType_module,
              moduleTab[builtModule].outputs[i]->name);
      free(sigt);
    }
    m = linkedModule->outputs[i].modNum;
    s = linkedModule->outputs[i].sigNum;
    if (((m == -1) && (s == i))
        || (linkedModule->subSigs[m].outputs[s].mark != 0))
      fprintf(cFile,"f%s",moduleTab[builtModule].outputs[i]->name);
    else
      fprintf(cFile,"%s.%s()",moduleTab[builtModule].modules[linkedModule->outputs[i].modNum].name,
                            moduleTab[
                              moduleTab[builtModule].modules[linkedModule->outputs[i].modNum].typeMod
                            ].outputs[linkedModule->outputs[i].sigNum]->name);
    fprintf(cFile,";\n}\n\n");
  }
    
  if (moduleTab[builtModule].kind == mDerived) {     /* Access methods for redefined signals of superclass */
    nb = moduleTab[moduleTab[builtModule].superMod].nbInputs;
    for (i = moduleTab[builtModule].nbInputs; i < moduleTab[builtModule].nbInputs + nb; i++) {
      if ((linkedModule->inputs[i].modNum != 0) && (linkedModule->inputs[i].modNum != -2)) {
        printTemplate(cFile, curType);
        if (moduleTab[builtModule].inputs[i]->kind == sPure)
          fprintf(cFile,"PureInSignal& %s::%s() {\n  return ",
                  nameOfType_module,
                  moduleTab[builtModule].inputs[i]->name);
        else
          fprintf(cFile,"InSignal<%s>& %s::%s() {\n  return ",
                  resolveTemplate(moduleTab[builtModule].inputs[i]->theType,
                                  moduleTab[builtModule].super,
                                  moduleTab[moduleTab[builtModule].superMod].name),
                  nameOfType_module,
                  moduleTab[builtModule].inputs[i]->name);
        if (linkedModule->inputs[i].modNum == -1)
          fprintf(cFile,"f%s;\n}\n\n", moduleTab[builtModule].inputs[i]->name);
        else {
          m = linkedModule->inputs[i].modNum;
          if (m < -1)
            m = -2 - m;
          fprintf(cFile,"%s.%s();\n}\n\n",
                  moduleTab[builtModule].modules[m].name,
                  moduleTab[
                    moduleTab[builtModule].modules[m].typeMod
                  ].inputs[-1 - linkedModule->inputs[i].sigNum]->name);
        }
      }
    }
    nb = moduleTab[moduleTab[builtModule].superMod].nbOutputs;
    for (i = moduleTab[builtModule].nbOutputs; i < moduleTab[builtModule].nbOutputs + nb; i++) {
      m = linkedModule->outputs[i].modNum;
      s = linkedModule->outputs[i].sigNum;
      if ((m != 0) || (s != (i - moduleTab[builtModule].nbOutputs))
          || (linkedModule->subSigs[m].outputs[s].mark != 0)) {
        printTemplate(cFile, curType);
        if (moduleTab[builtModule].outputs[i]->kind == sPure)
          fprintf(cFile,"PureOutSignal& %s::%s() {\n  return ",
                  nameOfType_module,
                  moduleTab[builtModule].outputs[i]->name);
        else
          fprintf(cFile,"OutSignal<%s>& %s::%s() {\n  return ",
                  resolveTemplate(moduleTab[builtModule].outputs[i]->theType,
                                  moduleTab[builtModule].super,
                                  moduleTab[moduleTab[builtModule].superMod].name),
                  nameOfType_module,
                  moduleTab[builtModule].outputs[i]->name);
        if (linkedModule->subSigs[m].outputs[s].mark != 0)
          fprintf(cFile,"f%s;\n}\n\n",
                  moduleTab[builtModule].outputs[i]->name);
        else
          fprintf(cFile,"%s.%s();\n}\n\n",
                  moduleTab[builtModule].modules[m].name,
                  moduleTab[
                    moduleTab[builtModule].modules[m].typeMod
                  ].outputs[linkedModule->outputs[i].sigNum]->name);
      }
    }
  }

  nb = moduleTab[builtModule].nbInputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;  /* inSigList() input signal list access method */

  printTemplate(cFile, curType);
  fprintf(cFile,"const InputSignal** %s::inSigList() const {\n  const InputSignal** inList;\n",
          nameOfType_module);
  fprintf(cFile,"  inList = (const InputSignal**) new (InputSignal*[%d]);\n", nb + 1);
  for (i = 0; i < nb; i++)
    fprintf(cFile,"  inList[%d] = &((%s*)this)->%s();\n", i,
            nameOfType_module,
            moduleTab[builtModule].inputs[i]->name);
  fprintf(cFile,"  inList[%d] = 0;\n  return inList;\n}\n", nb);


  nb = moduleTab[builtModule].nbOutputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;  /* outSigList() output signal list access method */

  printTemplate(cFile, curType);
  fprintf(cFile,"const OutputSignal** %s::outSigList() const {\n  const OutputSignal** outList;\n",
          nameOfType_module);
  fprintf(cFile,"  outList = (const OutputSignal**) new (OutputSignal*[%d]);\n", nb + 1);
  for (i = 0; i < nb; i++)
    fprintf(cFile,"  outList[%d] = &((%s*)this)->%s();\n", i,
            nameOfType_module,
            moduleTab[builtModule].outputs[i]->name);
  fprintf(cFile,"  outList[%d] = 0;\n  return outList;\n}\n\n", nb);

#ifdef DEAD_CODE
  buildAssignOperator(cFile);
#endif
  buildBackup(cFile);

  printTemplate(cFile, curType);            /* Constructor */
  fprintf(cFile, "%s::%s(", nameOfType_module, moduleTab[builtModule].name->name);
  if (moduleTab[builtModule].argsToConst != 0) {
    if (strcmp(moduleTab[builtModule].argsToConst->name,"#param") == 0) {
      fprintf(cFile, "const %sConsts& initInstance, ",
                      moduleTab[builtModule].name->name);
    } else {
      fprintf(cFile,"%s initInstance, ",
                     NameOfType(moduleTab[builtModule].argsToConst));
    }
  }
  fprintf(cFile, "const char* name, const bool schedule)\n");
  fprintf(cFile,"        : ");
  if (moduleTab[builtModule].kind == mDerived) {
  	fprintf(cFile, "%s(name, schedule)",nameOfType_super);
  	fprintf(cFile, ",\n          Synchronous(name, schedule)");
  	i = 1;
  } else {
  	fprintf(cFile,"Synchronous(name,schedule)");
  	i = 0;
  }
  
  for ( ; i < moduleTab[builtModule].nbSubModules; i++)        /* Submodule initialization */
    if (moduleTab[builtModule].modules[i].litArgs == 0) {
      fprintf(cFile,",\n          %s(subname(name,\"%s\"), false)",
                    moduleTab[builtModule].modules[i].name,
                    moduleTab[builtModule].modules[i].name);
      if (moduleTab[moduleTab[builtModule].modules[i].typeMod].argsToConst != 0) {
        sprintf(tmpStr,"argument of type \'%s\' expected for module %s.\n",
                NameOfType(moduleTab[moduleTab[builtModule].modules[i].typeMod].argsToConst),
                moduleTab[builtModule].modules[i].name);
        mdfWarn(tmpStr);
      }
    } else {
      fprintf(cFile,",\n          %s(%s, subname(name,\"%s\"), false)",
                    moduleTab[builtModule].modules[i].name,
                    moduleTab[builtModule].modules[i].litArgs,
                    moduleTab[builtModule].modules[i].name);
      if (moduleTab[moduleTab[builtModule].modules[i].typeMod].argsToConst == 0) {
        sprintf(tmpStr,"unexpected argument for module %s.\n",
                       moduleTab[builtModule].modules[i].name);
      }
    }
  nb = moduleTab[builtModule].nbInputs;                       /* Unassigned input signal initialization */
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;
  for (i = 0; i < nb; i++) {
    m = linkedModule->inputs[i].modNum;
    if (m < -1)
      m = -2 - m;
    s = linkedModule->inputs[i].sigNum;
    if ((m == -1) && (s == (-1 - i)))  {
      fprintf(cFile,",\n          f%s(this, \"%s\", false)",
              moduleTab[builtModule].inputs[i]->name,
              moduleTab[builtModule].inputs[i]->name);
    }
  }
  nb = moduleTab[builtModule].nbOutputs;                       /* Unassigned output signal initialization */
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  for (i = 0; i < nb; i++) {
    m = linkedModule->outputs[i].modNum;
    s = linkedModule->outputs[i].sigNum;
    if ((m == -1) && (s == i))
      fprintf(cFile,",\n          f%s(this, \"%s\", false)",
              moduleTab[builtModule].outputs[i]->name,
              moduleTab[builtModule].outputs[i]->name);
    else if (linkedModule->subSigs[m].outputs[s].mark != 0)
      if (linkedModule->subSigs[m].outputs[s].deps == 1)
        fprintf(cFile,",\n          f%s(this, \"%s\")",
                moduleTab[builtModule].outputs[i]->name,
                moduleTab[builtModule].outputs[i]->name);
      else
        fprintf(cFile,",\n          f%s(this, \"%s\", false)",
                moduleTab[builtModule].outputs[i]->name,
                moduleTab[builtModule].outputs[i]->name);
  }

  if (moduleTab[builtModule].argsToConst != 0) {
  	fprintf(cFile,",\n          fInitInstance(initInstance)");
  }

#ifdef DEAD_CODE
  fprintf(cFile,",\n          pf_backup_copy_(0)\n {\n");
#endif
  fprintf(cFile," {\n");
  nb = moduleTab[builtModule].nbInputs;					/* Rename assigned input signals */
  for(i = 0; i < nb; i++) {
  	m = linkedModule->inputs[i].modNum;
  	if (m < -1)
  		m = -2 -m;
  	s = linkedModule->inputs[i].sigNum;
  	if ((m != -1)||(s != (-1 - i))) {
  		fprintf(cFile,"  %s.%s().rename(\"%s\");\n",moduleTab[builtModule].modules[m].name,
  									moduleTab[moduleTab[builtModule].modules[m].typeMod].inputs[-1 -s]->name,
  									moduleTab[builtModule].inputs[i]->name);
  	}
  }
  nb = moduleTab[builtModule].nbOutputs;					/* Rename assigned output signals */
  for(i = 0; i < nb; i++) {
  	m = linkedModule->outputs[i].modNum;
  	if (m < -1)
  		m = -2 -m;
  	s = linkedModule->outputs[i].sigNum;
  	if (((m != -1)||(s != i))&&(linkedModule->subSigs[m].outputs[s].mark == 0)) {
  		fprintf(cFile,"  %s.%s().rename(\"%s\");\n",moduleTab[builtModule].modules[m].name,
  									moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name,
  									moduleTab[builtModule].outputs[i]->name);
  	}
  }
  if (moduleTab[builtModule].kind == mDerived) {	/* Input and output signals of super module */
  	nb = moduleTab[moduleTab[builtModule].superMod].nbInputs;
  	for(i = moduleTab[builtModule].nbInputs; i < nb; i++) {
  		m = linkedModule->inputs[i].modNum;
  		s = linkedModule->inputs[i].sigNum;
  		if ((m != 0)&&(m != -2)&& (m != -1)) {
	  		fprintf(cFile,"  %s.%s().rename(\"%s\");\n",moduleTab[builtModule].modules[m].name,
	  									moduleTab[moduleTab[builtModule].modules[m].typeMod].inputs[-1 - s]->name,
	  									moduleTab[builtModule].inputs[i]->name);
  		}
  	}
  	nb = moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  	for(i = moduleTab[builtModule].nbOutputs; i < nb; i++) {
  		m = linkedModule->outputs[i].modNum;
  		s = linkedModule->outputs[i].sigNum;
  		if ((linkedModule->subSigs[m].outputs[s].mark == 0) &&
  				((m != 0) || (s != (i - moduleTab[builtModule].nbOutputs)))) {
	  		fprintf(cFile,"  %s.%s().rename(\"%s\");\n",moduleTab[builtModule].modules[m].name,
	  									moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name,
	  									moduleTab[builtModule].outputs[i]->name);
  		}
  	}
  }
  for (i = 0; i < moduleTab[builtModule].nbSubModules; i++) {  /* Captures and connections of signals */
    typMod = &moduleTab[moduleTab[builtModule].modules[i].typeMod];
    for (j = 0; j < typMod->nbInputs; j++) {
      m = linkedModule->subSigs[i].inputs[j].modNum;
      s = linkedModule->subSigs[i].inputs[j].sigNum;
      if ((m == -1)&&((i != 0)||(moduleTab[builtModule].kind != mDerived)))
        /* Input of submodule is input of main module, and submodule is not 'super' */
        fprintf(cFile,"  %s.%s().chown(this);\n",moduleTab[builtModule].modules[i].name,
                                                 typMod->inputs[j]->name);
      else if ((m != i)&&(m != -1))
        /* Input signal is connected, but not to the main module */
        if ((moduleTab[builtModule].kind == mDerived)&&(i == 0))
          /* Module of signal is 'super' */
          fprintf(cFile,"  %s::%s() << %s.%s();\n",
                  nameOfType_super,
                  typMod->inputs[j]->name, moduleTab[builtModule].modules[m].name,
                  moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name);
        else if ((moduleTab[builtModule].kind == mDerived)&&(m == 0))
          /* Signal is connected to 'super' */
          fprintf(cFile,"  %s.%s() << %s::%s();\n",moduleTab[builtModule].modules[i].name,
                  typMod->inputs[j]->name, nameOfType_super,
                  moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name);
        else
          /* else, signal is connected, not to 'super' and not to main -> other submodule */
          fprintf(cFile,"  %s.%s() << %s.%s();\n",moduleTab[builtModule].modules[i].name,
                  typMod->inputs[j]->name, moduleTab[builtModule].modules[m].name,
                  moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name);
    }
    for (j = 0; j < typMod->nbOutputs; j++) {        /* Capture of output signals */
      m = linkedModule->subSigs[i].outputs[j].modNum;
      if (((m == -1)&&((i != 0)||(moduleTab[builtModule].kind != mDerived)))
          && (linkedModule->subSigs[i].outputs[j].mark == 0))
        /* Signal is output of main, submodule is not 'super' and signal is not internaly used */
        fprintf(cFile,"  %s.%s().chown(this);\n",
                moduleTab[builtModule].modules[i].name, typMod->outputs[j]->name);
    }
  }
  fprintf(cFile,"}\n\n");
  printTemplate(cFile, curType);
  fprintf(cFile, "%s::~%s() {\n",           /* Destructor */
          nameOfType_module,
          moduleTab[builtModule].name->name);
#ifdef DEAD_CODE
  fprintf(cFile,"  if (pf_backup_copy_ != 0)\n");
  fprintf(cFile,"    delete pf_backup_copy_;\n");
#endif
  fprintf(cFile,"}\n");
  free(nameOfType_module);
  if (moduleTab[builtModule].kind == mDerived)
    free(nameOfType_super);
}

#ifdef DEAD_CODE
void buildAssignOperator(FILE *cFile) {
  char *t,*st = 0;
  int nb, i, m, s;
  
  t = NameOfType(moduleTab[builtModule].name);
  
  printTemplate(cFile, moduleTab[builtModule].name->params);
  fprintf(cFile,"void %s::backup() {\n", t);
  fprintf(cFile,"  if (pf_backup_copy_ == 0) {\n");
  if (moduleTab[builtModule].argsToConst != 0) {
    fprintf(cFile,"    pf_backup_copy_ = new %s(fInitInstance,\"backup\",false);\n", t);
  } else {
    fprintf(cFile,"    pf_backup_copy_ = new %s(\"backup\",false);\n", t);
  }
  fprintf(cFile,"  }\n  *pf_backup_copy_ = *this;\n}\n\n");
  
  printTemplate(cFile, moduleTab[builtModule].name->params);
  fprintf(cFile,"void %s::restore() {\n", t);
  fprintf(cFile,"  if (pf_backup_copy_ != 0)\n");
  fprintf(cFile,"    *this = *pf_backup_copy_;\n}\n\n");
  
  printTemplate(cFile, moduleTab[builtModule].name->params);
  fprintf(cFile,"%s& %s::operator=(const %s& m) {\n", t, t, t);
  if (moduleTab[builtModule].kind == mDerived) {
    i = 1;
    st = NameOfType(moduleTab[builtModule].super);
	fprintf(cFile," *((%s*)this) = *((%s*)&m);\n",st,st);
  } else
    i = 0;
  for( ; i < moduleTab[builtModule].nbSubModules; i++) {
  	fprintf(cFile,"  %s = m.%s;\n",
	                 moduleTab[builtModule].modules[i].name,
					         moduleTab[builtModule].modules[i].name);
  }
  nb = moduleTab[builtModule].nbInputs;
  if (moduleTab[builtModule].kind == mDerived)
  	nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;
  for(i = 0; i < nb; i++) {
    m = linkedModule->inputs[i].modNum;
	  if (m < -1)
	    m = -2 - m;
	  s = linkedModule->inputs[i].sigNum;
	  if ((m == -1) && (s == (-1 - i))) {
	    fprintf(cFile,"  f%s.value() = ((%s&)m).f%s.value();\n",
	                   moduleTab[builtModule].inputs[i]->name,
					           t,     moduleTab[builtModule].inputs[i]->name);
	  }
  }
  nb = moduleTab[builtModule].nbOutputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  for (i = 0; i < nb; i++) {                 /* unassigned output signals */
    m = linkedModule->outputs[i].modNum;
    s = linkedModule->outputs[i].sigNum;
    if (((m == -1) && (s == i))
        || (linkedModule->subSigs[m].outputs[s].mark != 0)) {/* Or assigned outputs used internaly */
	    fprintf(cFile,"  f%s.value() = ((%s&)m).f%s.value();\n",
	                   moduleTab[builtModule].outputs[i]->name,
					           t,     moduleTab[builtModule].outputs[i]->name);
	  }
  }
  fprintf(cFile,"  return *this;\n}\n\n");
  free(t);
  if (st != 0)
    free(st);
}
#endif

void buildBackup(FILE *cFile) {
  char *t,*st = 0;
  int nb, i, m, s;
  
  t = NameOfType(moduleTab[builtModule].name);
  
  printTemplate(cFile, moduleTab[builtModule].name->params);
  fprintf(cFile,"void %s::backup() {\n", t);
  if (moduleTab[builtModule].kind == mDerived) {
    i = 1;
    st = NameOfType(moduleTab[builtModule].super);
	  fprintf(cFile," %s::backup();\n",st);
  } else
    i = 0;
  for( ; i < moduleTab[builtModule].nbSubModules; i++) {
	  fprintf(cFile,"  %s.backup();\n",
				  	         moduleTab[builtModule].modules[i].name);
  }
  fprintf(cFile,"}\n\n");
  
  printTemplate(cFile, moduleTab[builtModule].name->params);
  fprintf(cFile,"void %s::restore() {\n", t);
  if (moduleTab[builtModule].kind == mDerived) {
    i = 1;
	  fprintf(cFile," %s::restore();\n",st);
  } else
    i = 0;
  for( ; i < moduleTab[builtModule].nbSubModules; i++) {
	  fprintf(cFile,"  %s.restore();\n",
				  	         moduleTab[builtModule].modules[i].name);
  }
  fprintf(cFile,"}\n\n");
  
  free(t);
  if (st != 0)
    free(st);
}

int buildModuleTable() {
  int i;
  parsedList *curMod, *oldMod;
 
  curMod = parsedModules;
  nbModules = 0;
  builtModule = -1;
  while(curMod != 0) {
    nbModules ++;
    curMod = curMod->next;
  }
  moduleTab = (pMod*)checkAlloc(nbModules, sizeof(pMod));
  curMod = parsedModules;
  parsedModules = 0;
  i = 0;
  while(curMod != 0) {
    moduleTab[i].kind = curMod->parsedMod->kind;
    moduleTab[i].name = curMod->parsedMod->name;
    moduleTab[i].super = curMod->parsedMod->super;
    moduleTab[i].source = curMod->parsedMod->source;
    moduleTab[i].oc = curMod->parsedMod->oc;
    moduleTab[i].cpp = curMod->parsedMod->cpp;
    moduleTab[i].params = curMod->parsedMod->params;
    moduleTab[i].argsToConst = curMod->parsedMod->argsToConst;
    moduleTab[i].superMod = -1;
    moduleTab[i].reactSteps = -1;
    if ((moduleTab[i].kind == mComposite) || (moduleTab[i].kind == mDerived)) {
      if (builtModule == -1)
        builtModule = i;
      else {
        sprintf(tmpStr,"Two new module definition: %s and %s",
                NameOfType(moduleTab[builtModule].name),
                NameOfType(moduleTab[i].name));
        fatalError(tmpStr);
      }
    }
    if (moduleTab[i].kind == mDerived)
      moduleTab[i].nbSubModules = 1;   /* One implicit submodule : inherited */
    else
      moduleTab[i].nbSubModules = 0;
    moduleTab[i].nbInputs = 0;
    moduleTab[i].nbOutputs = 0;
    moduleTab[i].nbConnect = 0;
    moduleTab[i].nbEquiv = 0;
    moduleTab[i].nbNoDep = 0;
    buildModTabs(&moduleTab[i], curMod->parsedMod);
    oldMod = curMod;
      curMod = curMod->next;
    free(oldMod);
    i++;
  }
  if (builtModule == -1)
    return 1;
  updateDerivedMods();
  return 0;
}

void buildModTabs(pMod* newMod, module* rawMod) {
  declList *curDecl, *oldDecl;
  List curArgs, oldArgs;
  int i, o, m, c, e, n;
  atype *theType;

  curDecl = rawMod->body;
  while(curDecl != 0) {            /* Count declarations in module */
    curArgs = curDecl->val->val;
    switch(curDecl->val->kind) {
      case dInput:
        while(curArgs.sig != 0) {
          newMod->nbInputs ++;
          curArgs.sig = curArgs.sig->next;
        }
        break;
      case dOutput:
        while(curArgs.sig != 0) {
          newMod->nbOutputs ++;
          curArgs.sig = curArgs.sig->next;
        }
        break;
      case dModule:
        curArgs.mod = curArgs.sub->theMods;
        while (curArgs.mod != 0) {
          newMod->nbSubModules ++;
          curArgs.mod = curArgs.mod->next;
        }
        break;
      case dConnect:
        newMod->nbConnect ++;
        break;
      case dEquiv:
        newMod->nbEquiv ++;
        break;
      case dNoDep:
        while (curArgs.name != 0) {
          newMod->nbNoDep ++;
          curArgs.name = curArgs.name->next;
        }
        break;
      default:
        sprintf(tmpStr,"Unknown declaration kind %d in module %s",
              curDecl->val->kind, rawMod->name);
      intError(tmpStr);
        break;
    }
    curDecl = curDecl->next;
  }
  /* Now, allocate needed memory to store declarations */
  newMod->inputs = (sig**)checkAlloc(newMod->nbInputs, sizeof(sig*));
  newMod->outputs = (sig**)checkAlloc(newMod->nbOutputs, sizeof(sig*));
  newMod->modules = (pSubMod*)checkAlloc(newMod->nbSubModules, sizeof(pSubMod));
  newMod->connections = (connect**)checkAlloc(newMod->nbConnect, sizeof(connect*));
  newMod->sigEquiv = (connect**)checkAlloc(newMod->nbEquiv, sizeof(connect*));
  newMod->noDeps = (char**)checkAlloc(newMod->nbNoDep, sizeof(char*));
  i = 0; o = 0; m = 0; c = 0; e = 0; n = 0;
  if (newMod->kind == mDerived) {
    newMod->modules[m].theType = newMod->super;
    newMod->modules[m].typeMod = -1;
    newMod->modules[m++].name = "super";
  }
  curDecl = rawMod->body;
  while(curDecl != 0) {    /* Fill structures with info gathered during parse */
    curArgs = curDecl->val->val;
    switch(curDecl->val->kind) {
      case dInput:
        while (curArgs.sig != 0) {
          newMod->inputs[i++] = curArgs.sig->val;
          oldArgs = curArgs;
          curArgs.sig = curArgs.sig->next;
          free(oldArgs.sig);
        }
        break;
      case dOutput:
        while (curArgs.sig != 0) {
          newMod->outputs[o++] = curArgs.sig->val;
          oldArgs = curArgs;
          curArgs.sig = curArgs.sig->next;
          free(oldArgs.sig);
        }
        break;
      case dModule:
        theType = curArgs.sub->theType;
        curArgs.mod = curArgs.sub->theMods;
        while(curArgs.mod != 0) {
          newMod->modules[m].theType = theType;
          newMod->modules[m].name = curArgs.mod->modInst->modName;
          newMod->modules[m].litArgs = curArgs.mod->modInst->litArgs;
          newMod->modules[m++].typeMod = -1;
          oldArgs = curArgs;
          curArgs.mod = curArgs.mod->next;
          free(oldArgs.mod);
        }
        free(curDecl->val->val.sub);
        break;
      case dConnect:
        newMod->connections[c++] = curArgs.con;
        break;
      case dEquiv:
        newMod->sigEquiv[e++] = curArgs.con;
        break;
      case dNoDep:
        while (curArgs.name != 0) {
          newMod->noDeps[n++] = curArgs.name->name;
          oldArgs = curArgs;
          curArgs.name = curArgs.name->next;
          free(oldArgs.name);
        }
        break;
      default:
        sprintf(tmpStr,"Unknown declaration kind %d in module %s",
              curDecl->val->kind, rawMod->name);
        intError(tmpStr);
        break;
    }
    oldDecl = curDecl;
    curDecl = curDecl->next;
    free(oldDecl->val);
    free(oldDecl);
  }
}

void updateDerivedMods() {  /* Touch up signal tables of derived modules */
  int i, j;                 /* to take inherited signals into account */
  int supMod, nbIn, nbOut, nbSupIn, nbSupOut;
  sig **newSigTab;
  for (i = 0; i < nbModules; i++) {
    if (moduleTab[i].kind == mDerived) {
      supMod = findModule(moduleTab[i].super);
      nbIn = moduleTab[i].nbInputs;
      nbSupIn = moduleTab[supMod].nbInputs;
      nbOut = moduleTab[i].nbOutputs;
      nbSupOut = moduleTab[supMod].nbOutputs;
      if (nbSupIn != 0) {
        newSigTab = (sig**)checkAlloc(nbSupIn + nbIn, sizeof(sig*));
        for (j = 0; j < nbIn; j++)
          newSigTab[j] = moduleTab[i].inputs[j];
        for (j = 0; j < nbSupIn; j++)
          newSigTab[nbIn + j] = moduleTab[supMod].inputs[j];
        if (moduleTab[i].inputs != 0)
          free(moduleTab[i].inputs);
        moduleTab[i].inputs = newSigTab;
      }
      if (nbSupOut != 0) {
        newSigTab = (sig**)checkAlloc(nbSupOut + nbOut, sizeof(sig*));
        for (j = 0; j < nbOut; j++)
          newSigTab[j] = moduleTab[i].outputs[j];
        for (j = 0; j < nbSupOut; j++)
          newSigTab[nbOut + j] = moduleTab[supMod].outputs[j];
        if (moduleTab[i].outputs != 0)
          free(moduleTab[i].outputs);
        moduleTab[i].outputs = newSigTab;
      }
    }
  }
}

void linkModule() {
  int i, j, k, nb;
  
  if (moduleTab[builtModule].kind == mDerived)       /* Find module number for each module name */
    moduleTab[builtModule].superMod = findModule(moduleTab[builtModule].super);
  for (i = 0; i < moduleTab[builtModule].nbSubModules; i++)
    moduleTab[builtModule].modules[i].typeMod = findModule(moduleTab[builtModule].modules[i].theType);
  linkedModule = (bMod*)checkAlloc(1, sizeof(bMod));
  linkedModule->rStep = -1;
  linkedModule->mark = 0;
  nb = moduleTab[builtModule].nbInputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;
  linkedModule->inputs = (pSig*)checkAlloc(nb, sizeof(pSig));
  for (i = 0; i < nb; i++) {
    linkedModule->inputs[i].modNum = -1;     /* Signals are initialy connected to themselves */
    linkedModule->inputs[i].sigNum = -1 - i; /* Input signal are coded negatively from -1 */
    linkedModule->inputs[i].deps = 1;        /* The reaction depends on input signals */
    linkedModule->inputs[i].mark = 0;
  }
  nb = moduleTab[builtModule].nbOutputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  linkedModule->outputs = (pSig*)checkAlloc(nb, sizeof(pSig));
  for (i = 0; i < nb; i++) {
    linkedModule->outputs[i].modNum = -1;    /* Main module is -1, Super module (if any) is 0 */
    linkedModule->outputs[i].sigNum = i;
    linkedModule->outputs[i].deps = 1;       /* Output signals depend on the reaction */
    linkedModule->outputs[i].mark = 0;
  }
  linkedModule->subSigs = (bMod*)checkAlloc(moduleTab[builtModule].nbSubModules, sizeof(bMod));
  for (i = 0; i < moduleTab[builtModule].nbSubModules; i++) {  /* Initialize submodules */
    linkedModule->subSigs[i].rStep = -1;
    linkedModule->subSigs[i].mark = 0;
    linkedModule->subSigs[i].subSigs = 0;
    nb = moduleTab[moduleTab[builtModule].modules[i].typeMod].nbInputs;
    linkedModule->subSigs[i].inputs = (pSig*)checkAlloc(nb, sizeof(pSig));
    for (j = 0; j < nb; j++) {
      linkedModule->subSigs[i].inputs[j].modNum = i;
      linkedModule->subSigs[i].inputs[j].sigNum = -1 - j;
      linkedModule->subSigs[i].inputs[j].deps = 1;
      linkedModule->subSigs[i].inputs[j].mark = 0;
    }
    nb = moduleTab[moduleTab[builtModule].modules[i].typeMod].nbOutputs;
    linkedModule->subSigs[i].outputs = (pSig*)checkAlloc(nb, sizeof(pSig));
    for (j = 0; j < nb; j++) {
      linkedModule->subSigs[i].outputs[j].modNum = i;
      linkedModule->subSigs[i].outputs[j].sigNum = j;
      linkedModule->subSigs[i].outputs[j].deps = 1;
      linkedModule->subSigs[i].outputs[j].mark = 0;
    }
    nb = moduleTab[moduleTab[builtModule].modules[i].typeMod].nbNoDep;
    for (j = 0; j < nb; j++) {
      if (findSignal(&k,
                     moduleTab[builtModule].modules[i].typeMod,
                     moduleTab[moduleTab[builtModule].modules[i].typeMod].noDeps[j])) {
        if (k >= 0)
          linkedModule->subSigs[i].outputs[k].deps = 0; /* Out sig does not depend on the reaction */
        else
          linkedModule->subSigs[i].inputs[-1 - k].deps = 0; /* Reaction does not depend on in sig */
      } else {
        sprintf(tmpStr,
                "Unknown signal %s of module %s declared as not depending on input signals",
                 moduleTab[moduleTab[builtModule].modules[i].typeMod].noDeps[j],
                 NameOfType(moduleTab[moduleTab[builtModule].modules[i].typeMod].name));
        mdfError(tmpStr);
      }
    }
  }
  for (i = 0; i < moduleTab[builtModule].nbConnect; i++)
    linkSigs(builtModule, i, linkedModule);  /* Store connections in signal table */
  for (i = 0; i < moduleTab[builtModule].nbEquiv; i++)
    equivSigs(builtModule, i, linkedModule); /* Store signal captures in sig table */
  if (moduleTab[builtModule].kind == mDerived)
    equivSuperSigs(builtModule, linkedModule); /* Store automatic equivalences between */
                                               /* superclass signals and subclass signals */
  markSubModules();                          /* Compute nb of reaction steps between each */
                                             /* submodule and output signals */
}

void markSubModules() {
  int i, j, m, s, nbOut, nbIn;
  char *t;
  
  nbIn = moduleTab[builtModule].nbInputs;        /* Get number of signals for module */
  nbOut = moduleTab[builtModule].nbOutputs;
  if (moduleTab[builtModule].kind == mDerived) {
    nbIn += moduleTab[moduleTab[builtModule].superMod].nbInputs;
    nbOut += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  }
  for (i = 0; i < nbOut; i++) {     /* Starting from outputs, follow up the dependency chain */
    m = linkedModule->outputs[i].modNum;
    s = linkedModule->outputs[i].sigNum;
    if (m != -1) {
      markModulesFrom(m, s, 0, 1);  /* Module m will mark itself with 0 */
                                    /* and upper modules with 1 and so on */
    } else {          /* If output signal is not connected, warn user */
      sprintf(tmpStr,"Output signal %s is not connected to any submodule signal",
                     moduleTab[builtModule].outputs[i]->name);
      mdfWarn(tmpStr);
      linkedModule->outputs[i].deps = 0; /* Signal does not depend on the reaction */
    }
  }
  for (i = 0; i < moduleTab[builtModule].nbSubModules; i++) {
    if (linkedModule->subSigs[i].rStep == -1) { /* Check for any unreached module */
      if (forceReaction != 0) {
        for (j = 0; j < moduleTab[moduleTab[builtModule].modules[i].typeMod].nbOutputs; j++)
          markModulesFrom(i, j, 0, 1);
      } else {
        sprintf(tmpStr, "Submodule %s does not contribute to the reaction of %s",
                        moduleTab[builtModule].modules[i].name,
                        moduleTab[builtModule].name->name);
        mdfWarn(tmpStr);      /* and warn user of such dead modules */
      }
    }
  }
  for (i = 0; i < nbIn; i++) {
/*    if (linkedModule->inputs[i].modNum < -1) */  /* If input was reached during mark */
/*      linkedModule->inputs[i].modNum = -2 - linkedModule->inputs[i].modNum;*/
/*    else {             */                       /* Restore original modNum, else */
    if (linkedModule->inputs[i].modNum >= -1) {   /* If input was not reached during mark */
    	t = NameOfType(moduleTab[builtModule].name);
      sprintf(tmpStr,"Outputs of %s do not depend on input signal %s",
                     t, moduleTab[builtModule].inputs[i]->name);
      mdfWarn(tmpStr);                        /* Warn of unused input */
      free(t);
    }
  }
}

int markModulesFrom(int mNum, int sNum, int firstMark, int noDep) {
  int i, m, s, loopmod, newCircuit;
  if (linkedModule->subSigs[mNum].outputs[sNum].deps == 0) { /* We come here through a dead link */
    newCircuit = 1;
    noDep ++;                                     /* begin a new circuit */
  } else
    newCircuit = 0;
  if (linkedModule->subSigs[mNum].mark == noDep) { /* If we reach an already marked module */
    mdfLog("### Error : Dependency loop in submodules.\n");
    return mNum;        /* Return module number were the loop closed itself */
  }
  linkedModule->subSigs[mNum].mark = noDep;       /* Mark this module */
  if (firstMark > linkedModule->subSigs[mNum].rStep)
    linkedModule->subSigs[mNum].rStep = firstMark;  /* Store the step number */
  if (firstMark > moduleTab[builtModule].reactSteps)
    moduleTab[builtModule].reactSteps = firstMark; /* Update max number of steps */
  for (i = 0; i < moduleTab[moduleTab[builtModule].modules[mNum].typeMod].nbInputs; i++) {
    if ((linkedModule->subSigs[mNum].inputs[i].mark > 0)
        && (linkedModule->subSigs[mNum].inputs[i].mark < noDep))
      continue;      /* If already explored, no need to redo */
    linkedModule->subSigs[mNum].inputs[i].mark = noDep;
    m = linkedModule->subSigs[mNum].inputs[i].modNum; /* Look through each input signal */
    s = linkedModule->subSigs[mNum].inputs[i].sigNum;
    if (m != -1) {
      if ((m != mNum)||(s != (-1 - i))) {  /* If signal is connected */
        if (linkedModule->subSigs[mNum].inputs[i].deps == 0) /* If input signal is dead, new circuit */
          loopmod = markModulesFrom(m, s, firstMark + 1, noDep + 1);
        else
          loopmod = markModulesFrom(m, s, firstMark + 1, noDep); /* Follow up the link */
        if (loopmod != -1) {     /* If a loop was detected at module loopmod */
          sprintf(tmpStr,"%s.%s << %s.%s\n", moduleTab[builtModule].modules[mNum].name,
                  moduleTab[moduleTab[builtModule].modules[mNum].typeMod].inputs[i]->name,
                  moduleTab[builtModule].modules[m].name,
                  moduleTab[moduleTab[builtModule].modules[m].typeMod].outputs[s]->name);
          mdfLog(tmpStr);        /* Print the connection that we followed */
          if (loopmod != mNum)
            return loopmod;      /* If we are not the node of the loop, lets go down */
          else                   /* else the entire connection path is printed, so quit */
            mdfError("Loop detected in dependency graph of submodules");
        }
      }
    } else { /* Signal is input signal of the main module. Note that we reached it */
    	if (linkedModule->inputs[-1 - s].modNum >= -1)
    		linkedModule->inputs[-1 - s].modNum = -2 -linkedModule->inputs[-1 - s].modNum;
    }
  }
  for (i = 0; i < moduleTab[moduleTab[builtModule].modules[mNum].typeMod].nbInputs; i++)
    linkedModule->subSigs[mNum].inputs[i].mark = 0; /* Unmark input signals */
  linkedModule->subSigs[mNum].mark = 0;   /* Unmark module before returning */
  return -1;                              /* Everything is Ok */
}

int findModule(atype* name) {
  int i;
  atype* tdef;
  tdef = findTypdef(name);
  if (tdef != 0)
    name = tdef;
  for (i = 0; i < nbModules; i++) {       /* Look for name in each submodule */
    if (compatTypes(name, moduleTab[i].name) == 0)
    return i;
  }
  sprintf(tmpStr,"No definition for module type %s.", NameOfType(name));
  mdfError(tmpStr);      /* If not found, error! */
  return -1;
}

void linkSigs(int modId, int conId, bMod* lkMod) {
  int modNumF, sigNumF, modNumT, sigNumT;

  resolveSig(&modNumF, &sigNumF, modId, moduleTab[modId].connections[conId]->from, 0);
  resolveSig(&modNumT, &sigNumT, modId, moduleTab[modId].connections[conId]->to, 0);
  if (sigNumF < 0) {       /* Input signal ... */
    if (modNumF >= 0) {    /* belongs to submodule or superclass */
      if (sigNumT < 0) {   /* Connected to an input signal => Error... */
        if (moduleTab[modId].connections[conId]->from->kind == sMain)
          sprintf(tmpStr,"Input signal %s cannot be connected to another input signal",
                  moduleTab[modId].connections[conId]->from->name);
        else
          sprintf(tmpStr,"Input signal %s.%s cannot be connected to another input signal",
                  moduleTab[modId].connections[conId]->from->owner,
                  moduleTab[modId].connections[conId]->from->name);
        mdfError(tmpStr);
      }else {               /* Connected to an output signal */
        if (modNumT >= 0) { /* belonging to submodule or superclass */
          checkType(modId, modNumF, sigNumF, modNumT, sigNumT);
          checkConnect(modId, conId, modNumF, sigNumF, modNumT, sigNumT, lkMod);
          lkMod->subSigs[modNumF].inputs[-1 - sigNumF].modNum = modNumT;
          lkMod->subSigs[modNumF].inputs[-1 - sigNumF].sigNum = sigNumT;
          lkMod->subSigs[modNumT].outputs[sigNumT].modNum = modNumF;
          lkMod->subSigs[modNumT].outputs[sigNumT].sigNum = sigNumF;
          lkMod->subSigs[modNumT].outputs[sigNumT].mark = 1 + modNumF; /* Note that signal is internaly used */
        } else {            /* belonging to main module => Error */
          if (moduleTab[modId].connections[conId]->from->kind == sMain)
            sprintf(tmpStr,"Input signal %s cannot be connected to output signal %s",
                    moduleTab[modId].connections[conId]->from->name,
                    moduleTab[modId].connections[conId]->to->name);
          else
            sprintf(tmpStr,"Input signal %s.%s cannot be connected to output signal %s",
                    moduleTab[modId].connections[conId]->from->owner,
                    moduleTab[modId].connections[conId]->from->name,
                    moduleTab[modId].connections[conId]->to->name);
            mdfError(tmpStr);
        }
      }
    }else {                 /* Input signal belongs to main module => Error */
      sprintf(tmpStr,"Input signal %s cannot be internally connected",
              moduleTab[modId].connections[conId]->from->name);
      mdfError(tmpStr);
    }
  }else {                 /* An output signal => Cannot be connected to any signal : Error */
    if (moduleTab[modId].connections[conId]->from->kind == sMain)
      sprintf(tmpStr,"Output signal %s cannot be connected to another signal",
              moduleTab[modId].connections[conId]->from->name);
    else
      sprintf(tmpStr,"Output signal %s.%s cannot be connected to another signal",
              moduleTab[modId].connections[conId]->from->owner,
              moduleTab[modId].connections[conId]->from->name);
  }
}

void equivSigs(int modId, int equId, bMod* lkMod) {
  int modNumF, sigNumF, modNumT, sigNumT;

  resolveSig(&modNumF, &sigNumF, modId, moduleTab[modId].sigEquiv[equId]->from, 1);
  resolveSig(&modNumT, &sigNumT, modId, moduleTab[modId].sigEquiv[equId]->to, 1);
  if (sigNumF < 0) {         /* Input signal */
    if (modNumF < 0) {       /* Of main module => Error */
      sprintf(tmpStr,"Cannot assign to input signal %s",
              moduleTab[modId].sigEquiv[equId]->from->name);
      mdfError(tmpStr);
    }else {                  /* Input signal of a subModule */
      if (sigNumT < 0) {     /* is input signal... */
        if (modNumT < 0) {   /* of main module => Ok */
          checkType(modId, modNumF, sigNumF, modNumT, sigNumT);
          checkEquiv(modId, equId, modNumF, sigNumF, modNumT, sigNumT, lkMod);
          lkMod->subSigs[modNumF].inputs[-1 - sigNumF].modNum = modNumT;
          lkMod->subSigs[modNumF].inputs[-1 - sigNumF].sigNum = sigNumT;
          lkMod->inputs[-1 - sigNumT].modNum = modNumF;
          lkMod->inputs[-1 - sigNumT].sigNum = sigNumF;
        }else {              /* of a submodule => should be a connection */
          sprintf(tmpStr,"Cannot assign input signal %s.%s to input signal %s.%s",
                  moduleTab[modId].sigEquiv[equId]->to->owner,
                  moduleTab[modId].sigEquiv[equId]->to->name,
                  moduleTab[modId].sigEquiv[equId]->from->owner,
                  moduleTab[modId].sigEquiv[equId]->from->name);
          mdfError(tmpStr);
        }
      }else {               /* Input signal of submodule is an output signal!! */
        sprintf(tmpStr,"Cannot assign an output signal to input signal %s.%s",
                moduleTab[modId].sigEquiv[equId]->from->owner,
                moduleTab[modId].sigEquiv[equId]->from->name);
        mdfError(tmpStr);
      }
    }
  }else {                   /* Output signal */
    if (modNumF < 0) {      /* of main module */
      if (sigNumT < 0) {    /* is input signal => Error */
        sprintf(tmpStr,"Cannot assign an input signal to output signal %s",
                moduleTab[modId].sigEquiv[equId]->from->name);
        mdfError(tmpStr);
      }else {               /* Output signal of Main is output signal... */
        if (modNumT < 0) {  /* of main module => Error */
          sprintf(tmpStr,"Cannot assign output signal %s to output signal %s",
                  moduleTab[modId].sigEquiv[equId]->to->name,
                  moduleTab[modId].sigEquiv[equId]->from->name);
          mdfError(tmpStr);
        }else {             /* of a submodule : Ok */
          checkType(modId, modNumF, sigNumF, modNumT, sigNumT);
          lkMod->outputs[sigNumF].modNum = modNumT;
          lkMod->outputs[sigNumF].sigNum = sigNumT;
          lkMod->subSigs[modNumT].outputs[sigNumT].modNum = modNumF;
          lkMod->subSigs[modNumT].outputs[sigNumT].sigNum = sigNumF;
        }
      }
    }else {                 /* Output signal of submodule => Error */
      sprintf(tmpStr,"Output signal %s.%s cannot be assigned to",
              moduleTab[modId].sigEquiv[equId]->from->owner,
              moduleTab[modId].sigEquiv[equId]->from->name);
      mdfError(tmpStr);
    }
  }
}

void equivSuperSigs(int modId, bMod* lkMod) {
  int i, nb;
  int nbOut, nbIn;
  nbOut = moduleTab[builtModule].nbOutputs;
  nbIn = moduleTab[builtModule].nbInputs;
  nb = moduleTab[builtModule].nbOutputs
       + moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  for (i = nbOut; i < nb; i++) {
    if ((lkMod->outputs[i].modNum == -1)&&(lkMod->outputs[i].sigNum == i)) {
      lkMod->outputs[i].modNum = 0;
      lkMod->outputs[i].sigNum = i - nbOut;
      lkMod->subSigs[0].outputs[i - nbOut].modNum = -1;
      lkMod->subSigs[0].outputs[i - nbOut].sigNum = i;
    }
  }
  nb = moduleTab[moduleTab[builtModule].superMod].nbInputs;
  for (i = 0; i < nb; i++) {
    if ((lkMod->subSigs[0].inputs[i].modNum == 0)
        &&(lkMod->subSigs[0].inputs[i].sigNum == (-1 - i))) {
      lkMod->subSigs[0].inputs[i].modNum = -1;
      lkMod->subSigs[0].inputs[i].sigNum = -1 - (i + nbIn);
      lkMod->inputs[i + nbIn].modNum = 0;
      lkMod->inputs[i + nbIn].sigNum = -1 - i;
    }
  }
}

void checkType(int modId, int mNumF, int sNumF, int mNumT, int sNumT) {
  sigType kindF, kindT;
  char *typF, *typT, *nameF, *nameT;
  char mNameF[50], mNameT[50];
  pMod *mod;
  pSubMod *subMod;
  
  if (mNumF == -1) {
    strcpy(mNameF, "");
    if (sNumF < 0) {
      kindF = moduleTab[modId].inputs[-1 - sNumF]->kind;
      if (kindF != sPure) {
        if ((-1 -sNumF) >= moduleTab[modId].nbInputs) /* If signal of super */
          typF = resolveTemplate(moduleTab[modId].inputs[-1 - sNumF]->theType,
                                 moduleTab[modId].super,
                                 moduleTab[moduleTab[modId].superMod].name);
        else
          typF = NameOfType(moduleTab[modId].inputs[-1 - sNumF]->theType);
      }
      nameF = moduleTab[modId].inputs[-1 - sNumF]->name;
    }else {
      kindF = moduleTab[modId].outputs[sNumF]->kind;
      if (kindF != sPure) {
        if (sNumF >= moduleTab[modId].nbOutputs) /* If signal of super */
          typF = resolveTemplate(moduleTab[modId].outputs[sNumF]->theType,
                                 moduleTab[modId].super,
                                 moduleTab[moduleTab[modId].superMod].name);
        else
          typF = NameOfType(moduleTab[modId].outputs[sNumF]->theType);
      }
      nameF = moduleTab[modId].outputs[sNumF]->name;
    }
  }else {
    subMod = &(moduleTab[modId].modules[mNumF]);
    mod = &moduleTab[subMod->typeMod];
    strcpy(mNameF, moduleTab[modId].modules[mNumF].name);
    strcat(mNameF, ".");
    if (sNumF < 0) {
      kindF = mod->inputs[-1 - sNumF]->kind;
      if (kindF != sPure)
        typF = resolveTemplate(mod->inputs[-1 - sNumF]->theType,
                               subMod->theType,
                               mod->name);
      nameF = mod->inputs[-1 - sNumF]->name;
    }else {
      kindF = mod->outputs[sNumF]->kind;
      if (kindF != sPure)
        typF = resolveTemplate(mod->outputs[sNumF]->theType,
                               subMod->theType,
                               mod->name);
      nameF = mod->outputs[sNumF]->name;
    }
  }
  if (mNumT == -1) {
    strcpy(mNameT, "");
    if (sNumT < 0) {
      kindT = moduleTab[modId].inputs[-1 - sNumT]->kind;
      if (kindT != sPure) {
        if ((-1 - sNumT) >= moduleTab[modId].nbInputs)  /* If signal of super */
          typT = resolveTemplate(moduleTab[modId].inputs[-1 - sNumT]->theType,
                                 moduleTab[modId].super,
                                 moduleTab[moduleTab[modId].superMod].name);
        else
          typT = NameOfType(moduleTab[modId].inputs[-1 - sNumT]->theType);
      }
      nameT = moduleTab[modId].inputs[-1 - sNumT]->name;
    }else {
      kindT = moduleTab[modId].outputs[sNumT]->kind;
      if (kindT != sPure) {
        if (sNumT >= moduleTab[modId].nbOutputs)  /* If signal of super */
          typT = resolveTemplate(moduleTab[modId].outputs[sNumT]->theType,
                                 moduleTab[modId].super,
                                 moduleTab[moduleTab[modId].superMod].name);
        else
          typT = NameOfType(moduleTab[modId].outputs[sNumT]->theType);
      }
      nameT = moduleTab[modId].outputs[sNumT]->name;
    }
  }else {
    subMod = &(moduleTab[modId].modules[mNumT]);
    mod = &moduleTab[subMod->typeMod];
    strcpy(mNameT, moduleTab[modId].modules[mNumT].name);
    strcat(mNameT, ".");
    if (sNumT < 0) {
      kindT = mod->inputs[-1 - sNumT]->kind;
      if (kindT != sPure)
        typT = resolveTemplate(mod->inputs[-1 - sNumT]->theType,
                               subMod->theType,
                               mod->name);
      nameT = mod->inputs[-1 - sNumT]->name;
    }else {
      kindT = mod->outputs[sNumT]->kind;
      if (kindT != sPure)
        typT = resolveTemplate(mod->outputs[sNumT]->theType,
                               subMod->theType,
                               mod->name);
      nameT = mod->outputs[sNumT]->name;
    }
  }
  if ((kindF == sPure)&&(kindT == sPure))
    return;
  if ((kindF == sTyped)&&(kindT == sTyped))
    if (compareTypeNames(typT, typF) == 0)
      return;
  if (kindT == sPure)
    typT = "pure";
  if (kindF == sPure)
    typF = "pure";
  sprintf(tmpStr,"Type mismatch between %s signal %s%s and %s signal %s%s",
          typF, mNameF, nameF, typT, mNameT, nameT);
  mdfError(tmpStr);
}

void checkConnect(int modId, int conId, int mNumF, int sNumF, int mNumT, int sNumT, bMod* lkMod) {
  char toBuf[256];
  int toMod, toSig;
  
  if (mNumF == -1)
     return;
  toMod = lkMod->subSigs[mNumF].inputs[-1 - sNumF].modNum;
  toSig = lkMod->subSigs[mNumF].inputs[-1 - sNumF].sigNum;
  if ((toMod != mNumF) || (toSig != sNumF)) {
      sprintf(tmpStr,"Connecting %s.%s to %s.%s",
              moduleTab[builtModule].modules[mNumF].name,
              moduleTab[modId].connections[conId]->from->name,
              moduleTab[builtModule].modules[mNumT].name,
              moduleTab[modId].connections[conId]->to->name);
      mdfLog(tmpStr);
      sprintf(tmpStr,"Input signal %s.%s is already connected to ",
              moduleTab[builtModule].modules[mNumF].name,
              moduleTab[modId].connections[conId]->from->name);
      if (toMod == -1) {
        if (toSig < 0) 
          sprintf(toBuf, "%s\n", moduleTab[builtModule].inputs[-1 - toSig]->name);
        else
          sprintf(toBuf, "%s\n", moduleTab[builtModule].outputs[toSig]->name);
      } else {
        if (toSig < 0) 
          sprintf(toBuf, "%s.%s\n", moduleTab[builtModule].modules[toMod].name,
                  moduleTab[moduleTab[builtModule].modules[toMod].typeMod].inputs[-1 - toSig]->name);
        else
          sprintf(toBuf, "%s.%s\n", moduleTab[builtModule].modules[toMod].name,
                  moduleTab[moduleTab[builtModule].modules[toMod].typeMod].outputs[toSig]->name);
      }
      strcat(tmpStr, toBuf);
      mdfError(tmpStr);
  }
}

void checkEquiv(int modId, int equId, int mNumF, int sNumF, int mNumT, int sNumT, bMod* lkMod) {
  char toBuf[256];
  int toMod, toSig;
  
  toMod = lkMod->subSigs[mNumF].inputs[-1 - sNumF].modNum;
  toSig = lkMod->subSigs[mNumF].inputs[-1 - sNumF].sigNum;
  if ((toMod != mNumF) || (toSig != sNumF)) {
    sprintf(tmpStr,"Assigning %s to %s.%s",
            moduleTab[modId].sigEquiv[equId]->to->name,
            moduleTab[builtModule].modules[mNumF].name,
            moduleTab[modId].sigEquiv[equId]->from->name);
    mdfLog(tmpStr);
    sprintf(tmpStr,"Input signal %s.%s is already connected to ",
             moduleTab[builtModule].modules[mNumF].name,
             moduleTab[modId].sigEquiv[equId]->from->name);
    if (toMod == -1) {
      if (toSig < 0)
        sprintf(toBuf, "%s\n", moduleTab[builtModule].inputs[-1 - toSig]->name);
      else
        sprintf(toBuf, "%s\n", moduleTab[builtModule].outputs[toSig]->name);
    } else {
      if (toSig < 0)
        sprintf(toBuf, "%s.%s\n", moduleTab[builtModule].modules[toMod].name,
                moduleTab[moduleTab[builtModule].modules[toMod].typeMod].inputs[-1 - toSig]->name);
      else
        sprintf(toBuf, "%s.%s\n", moduleTab[builtModule].modules[toMod].name,
                moduleTab[moduleTab[builtModule].modules[toMod].typeMod].outputs[toSig]->name);
    }
    strcat(tmpStr, toBuf);
    mdfError(tmpStr);
  }
}

void resolveSig(int* modNum, int* sigNum, int modId, sig* theSig, int isEquiv) {
  switch(theSig->kind) {
    case sMain:
      if (findSignal(sigNum, modId, theSig->name)) {
        *modNum = -1;
        return;
      }
      if (moduleTab[modId].kind == mDerived)
        if (findSignal(sigNum, moduleTab[modId].superMod, theSig->name)) {
          if (isEquiv) {
            *modNum = -1;
            if (*sigNum < 0)
              *sigNum -= moduleTab[modId].nbInputs;
            else
              *sigNum += moduleTab[modId].nbOutputs;
          }else {
            *modNum = 0;
          }
          return;
        }
      sprintf(tmpStr,"Unknown signal %s in module %s",
              theSig->name, moduleTab[modId].name->name);
      mdfError(tmpStr);
      break;
    case sSub:
      if (findSubMod(modNum, modId, theSig->owner)) {
        if (findSignal(sigNum, moduleTab[modId].modules[*modNum].typeMod, theSig->name))
          return;
        else {
          sprintf(tmpStr, "Unknown signal %s in instance %s of class %s",
                  theSig->name, theSig->owner, moduleTab[modId].modules[*modNum].theType);
          mdfError(tmpStr);
        }
      }
      sprintf(tmpStr, "Unknown module %s in %s",
              theSig->owner, moduleTab[modId].name->name);
      mdfError(tmpStr);
      break;
  default:
    sprintf(tmpStr, "Bad signal kind %d for signal %s in module %s",
            theSig->kind, theSig->name, moduleTab[modId].name->name);
    break;
  }
}

int findSignal(int *sigNum, int modTyp, char* name) {
  int i;
  for (i = 0; i < moduleTab[modTyp].nbInputs; i++)
    if (strcmp(name, moduleTab[modTyp].inputs[i]->name) == 0) {
      *sigNum = -1 - i;
      return 1;
    }
  for (i = 0; i < moduleTab[modTyp].nbOutputs; i++)
    if (strcmp(name, moduleTab[modTyp].outputs[i]->name) == 0) {
      *sigNum = i;
      return 1;
    }
  return 0;
}

int findSubMod(int *modNum, int modId, char *name) {
  int i;
  for (i = 0; i < moduleTab[modId].nbSubModules; i++)
    if (strcmp(name, moduleTab[modId].modules[i].name) == 0) {
      *modNum = i;
      return 1;
    }
  return 0;
}
