/*
//  File mdlErrors.c
//
//  September 3 1992 - September 1994
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1994.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

/* This file contains functions used for error handling.         */
/*****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "mdlParse.h"

char tmpStr[256];          /* Utility buffer for error messages */

/*************************************************************/
/* Procedure usageError handles mdlc parameter list errors.  */
/* It prints an error message followed by the correct usage  */
/* on stderr and then exits with a status of 1.              */
/*************************************************************/
void usageError(char *msg) {
  fprintf(stderr,"### mdlc - Fatal usage error\n");
  fprintf(stderr,"# %s\n",msg);
  fprintf(stderr,"# Usage : mdlc [-v][-version][-B baseName] [-debug]\n"
                 "#              [-mdf | -mdfB mdfBaseName] [-force] [-cxxinst instfile]\n"
                 "#              [-proto] [-simtext] [-simX] [-simP] [-type typelist]\n"
                 "#              [-inc libSyncInterface] [-Idir] [-access] [-info]\n"
                 "#              mdlFile1 mdlFile2 [mdlFileX]*\n");
  exit(1);
}

/*********************************************************/
/* Procedure fileError handle file related fatal errors. */
/* It displays the given error message and then exits    */
/* with an error status of 1                             */
/*********************************************************/
void fileError(char *msg) {
  fprintf(stderr,"### mdlc - Fatal I/O error : %s\n",msg);
  exit(1);
}

/*********************************************************/
/* Procedure fatalError handle unqualified fatal errors. */
/* It displays the given error message and then exits    */
/* with an error status of 1                             */
/*********************************************************/
void fatalError(char *msg) {
  fprintf(stderr,"### mdlc - Fatal error : %s\n",msg);
  exit(1);
}

/*********************************************************************/
/* Procedure mdfError handles errors due to inconsistencies in the   */
/* mdlc source code.                                                 */
/*********************************************************************/
void mdfError(char *msg) {
  fprintf(stderr,"### mdlc - Fatal : mdl source error\n");
  fprintf(stderr," # %s\n",msg);
  exit(1);
}

/****************************************************************************/
/* Procedure intError handles internal errors (table inconsistency etc...). */
/* It prints the message describing the error and exits with status 1. This */
/* procedure should never be called !!!                                     */
/****************************************************************************/
void intError(char *msg) {
  fprintf(stderr,"### mdlc - Fatal Internal Error : %s\n",msg);
  exit(1);
}

/****************************************************************************/
/* Procedure mdfWarn handles warnings.                                      */
/* It prints the message describing the warning and returns.                */
/****************************************************************************/
void mdfWarn(char* msg) {
  fprintf(stderr,"### mdlc - Warning : %s\n", msg);
}

/****************************************************************************/
/* Procedure mdfLog handles message logging.                                */
/* It prints the log message and returns.                                   */
/****************************************************************************/
void mdfLog(char* msg) {
  fprintf(stderr,"#  %s\n", msg);
}

/***********************************************************************/
/* Function yyerror is used by the yacc parser to signal a parse error.*/
/* It receives an error message and is expected to return 0 if we      */
/* cannot recover from the error, 1 if we can.                         */
/***********************************************************************/
int yyerror(char *s) {
  fprintf(stderr,"### mdlc - Fatal parse error\n");
  fprintf(stderr," File '%s'; Line %d\n",sourceFile,lineNb);
  fprintf(stderr," # %s, offending token is '%s'\n",s,yytext);
  return 0;    /* We are not enough clever to recover from a parse error.*/
}
