/*
//  File mdlGram.y
//
//  September 3 1992 - January 1995
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1995.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

/* This file contains the grammar used to parse Module           */
/* Desciption files.                                             */
/* This grammar contains 3 shift/reduce conflicts                */
/*****************************************************************/


%{
/*****************************************************************/
/* File mdlGram.tab.c                                            */
/* This file is part of the occxx oc to C++ compiler             */
/*         Frederic Boulanger, Ecole Superieure d'Electricite    */
/*            boulange@supelec.fr                                */
/*                                                               */
/* This file is produced by Yacc from mdlGram.y.                 */
/*****************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "mdlParse.h"
#include "mdlTables.h"
#include "mdlErrors.h"

#define YYDEBUG 0
%}

%union{
  char *strval;
  strpair st2val;
  module *modval;
  declList *dlsval;
  decl *decval;
  sigList *slsval;
  sig *sigval;
  paramList *prlval;
  param *parval;
  modList *mlsval;
  atype *typval;
  typeList *tlsval;
  nameList *nlsval;
  subModInst *smival;
}

%token  tComma      tSemiColon    tColon        tLeftBrace
%token  tRightBrace tLeftBracket  tRightBracket tIs
%token  tConnect    tIdent        tCompIdent    tString
%token  tTypdef     tInputs       tOutputs      tSource
%token  tOc         tCpp          tImport       tNoDep        tTypequ
%token  tLeftPar    tRightPar     tLiteral      tHinsert
%token  tCinsert    tEndInsert    tRawLine      tUse         tParam

%type <strval> tIdent tString tLiteral tRawLine
%type <st2val> tCompIdent

%type <strval> mdf.source mdf.oc mdf.cpp mdf.imp
%type <modval> mdf.prog mdf.declared mdf.composite mdf.derived
%type <dlsval> mdf.decl mdf
%type <decval> mdf.stat sig.decl module.decl sig.connect sig.equiv
%type <decval> mdf.decl.stat sig.rel
%type <slsval> typed.input.decl typed.output.decl typed.sig.list
%type <sigval> sig pure.sig typed.sig signal main.signal sub.signal
%type <mlsval> module.list
%type <typval> objtype simple.type param.type constargs
%type <tlsval> paramt.list
%type <nlsval> sig.name.list raw.data
%type <smival> mod.inst
%type <prlval> mdf.param mdf.param.elems
%type <parval> mdf.param.elem

%start mdf.fich

%%
mdf.fich : mdf.source mdf.oc mdf.cpp mdf.imp mdf.hins mdf.cins mdf.use mdf.param mdf.typdefs mdf.prog
           { $10->source = $1;
             $10->oc = $2;
             $10->cpp = $3;
             $10->imp = $4;
             $10->params = $8;
             declareModule($10);
           }
         | mdf.source mdf.oc mdf.cpp mdf.imp mdf.hins mdf.cins mdf.use mdf.param mdf.prog
           { $9->source = $1;
             $9->oc = $2;
             $9->cpp = $3;
             $9->imp = $4;
             $9->params = $8;
             declareModule($9);
           }
         ;

mdf.source : tSource tString
             { $$ = $2; }
           | /* empty */
             { $$ = ""; }
           ;

mdf.oc : tOc tString
         { $$ = $2; }
       | /* empty */
         { $$ = ""; }
       ;

mdf.cpp : tCpp tString
          { $$ = $2; }
        | /* empty */
          { $$ = ""; }
        ;

mdf.imp : tImport tString
          { $$ = $2; }
        | /* empty */
          { $$ = ""; }
        ;

mdf.hins : /* empty */
         | tHinsert raw.data tEndInsert
           { HInsert($2); }
         ;

mdf.cins : /* empty */
         | tCinsert raw.data tEndInsert
           { CInsert($2); }
         ;

raw.data : /* empty */
           { $$ = 0; }
         | tRawLine
           { $$ = addName((nameList*)0,$1); }
         | raw.data tRawLine
           { $$ = addName($1,$2); }
         ;

mdf.use : /*empty*/
        | tUse tIdent
          { declareUse($2); }
        | mdf.use tUse tIdent
          { declareUse($3); }
        ;

mdf.param : /* empty */
             { $$ = (paramList*)0; }
          | tParam tLeftBrace mdf.param.elems tRightBrace
             { $$ = $3; }
          ;

mdf.param.elems : mdf.param.elem
                  { $$ = addParam((paramList*)0, $1); }
                | mdf.param.elems mdf.param.elem
                  { $$ = addParam($1, $2); }
                ;

mdf.param.elem : objtype tIdent tSemiColon
                 { $$ = newParam($1,$2); }
               ;

mdf.typdefs : mdf.typdef
            | mdf.typequ
            | mdf.typdefs mdf.typdef
            | mdf.typdefs mdf.typequ
            ;

mdf.typdef : simple.type tTypdef objtype tSemiColon
             { declareTypdef($1, $3); }
           ;

mdf.typequ : simple.type tTypequ objtype tSemiColon
             { declareTypequ($1, $3); }
           ;

mdf.prog : mdf.declared
           { $$ = $1; }
         | mdf.composite
           { $$ = $1; }
         | mdf.derived
           { $$ = $1; }
         ;

mdf.declared : objtype constargs tLeftBrace mdf.decl tRightBrace
               { $$ = newModule(mDeclared, $1, $2, 0, $4); }
             ;

mdf.composite : objtype constargs tColon tLeftBrace mdf tRightBrace
               { $$ = newModule(mComposite, $1, $2, 0, $5); }
              ;

mdf.derived : objtype constargs tColon objtype tLeftBrace mdf tRightBrace
               { $$ = newModule(mDerived, $1, $2, $4, $6); }
            ;

constargs : /* empty */
             { $$ = (atype*)0; }
          | tLeftPar tLiteral tRightPar
             { $$ = buildTypeFromLiteral($2); }
          ;

mdf.decl : mdf.decl.stat
          { $$ = addDecl((declList*)0, $1); }
         | mdf.decl mdf.decl.stat
          { $$ = addDecl($1, $2); }
         ;

mdf.decl.stat : sig.decl
                { $$ = $1; }
              | sig.rel
                { $$ = $1; }
              ;

objtype : simple.type
          { $$ = $1; }
        | param.type
          { $$ = $1; }
        ;

simple.type : tIdent
              { $$ = newType($1, 0); }
            ;

param.type : tIdent tLeftBracket paramt.list tRightBracket
             { $$ = newType($1, $3); }
           ;

paramt.list : objtype
              { $$ = addType((typeList*)0, $1); }
            | paramt.list tComma objtype
              { $$ = addType($1, $3); }
            ;

typed.input.decl : tInputs typed.sig.list tSemiColon
                   { $$ = $2; }
                 ;

typed.output.decl : tOutputs typed.sig.list tSemiColon
                    { $$ = $2; }
                  ;

typed.sig.list : sig
                 { $$ = addSig((sigList*)0, $1); }
               | typed.sig.list tComma sig
                 { $$ = addSig($1, $3); }
               ;

sig : pure.sig
      { $$ = $1; }
    | typed.sig
      { $$ = $1; }
    ;

mdf : mdf.stat
          { $$ = addDecl((declList*)0, $1); }
        | mdf mdf.stat
          { $$ = addDecl($1, $2); }
        ;

mdf.stat : sig.decl
           { $$ = $1; }
         | module.decl
           { $$ = $1; }
         | sig.connect
           { $$ = $1; }
         | sig.equiv
           { $$ = $1; }
         | sig.rel
           { $$ = $1; }
         ;

sig.decl : typed.input.decl
           { $$ = newInpDecl($1); }
         | typed.output.decl
           { $$ = newOutpDecl($1); }
         ;

typed.sig : objtype tIdent
            { $$ = newSignal(sTyped, $1, 0, $2); }
          ;

pure.sig : tIdent
            { $$ = newSignal(sPure, 0, 0, $1); }
         ;

module.decl : objtype module.list tSemiColon
              { $$ = newSubModDecl(newSubModules($1, $2)); }
            ;

module.list : mod.inst
              { $$ = addModule((modList*)0, $1); }
            | module.list tComma mod.inst
              { $$ = addModule($1, $3); }
            ;

mod.inst : tIdent
           { $$ = newSubModInst($1,0); }
         | tIdent tLeftPar tLiteral tRightPar
           { $$ = newSubModInst($1, $3); }
         ;

sig.connect : signal tConnect signal tSemiColon
              { $$ = newConnDecl(newConnect($1, $3)); }
            ;

sig.equiv : signal tIs signal tSemiColon
            { $$ = newEquivDecl(newConnect($1, $3)); }
          ;

sig.rel : tNoDep sig.name.list tSemiColon
          { $$ = newNoDepDecl($2);}
          ;

sig.name.list : tIdent
                { $$ = addName((nameList*)0, $1); }
              | sig.name.list tComma tIdent
                { $$ = addName($1, $3); }
              ;

signal : main.signal
         { $$ = $1; }
       | sub.signal
         { $$ = $1; }
       ;

main.signal : tIdent
              { $$ = newSignal(sMain, 0, 0, $1); }
            ;

sub.signal : tCompIdent
             { $$ = newSignal(sSub, 0, $1.str1, $1.str2);}
           ;

