/*
//  File mdlMdf.c
//
//  September 3 1992 - March 1994
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1994.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

/* This file contains code to produce Module Description files.     */
/* mdf files are used to build compound or derived synchronous      */
/* classes from previously compiled oc modules.                     */
/********************************************************************/

#include "env.h"
#include "mdlCxx.h"
#include "mdlErrors.h"
#include "mdl.h"
#include "mdlUtils.h"
#include "mdlTables.h"
#include <stdio.h>
#include <string.h>

void writeSigTab(FILE *mdfFile, int nb, sig** sigTab) {
  int first = 1;
  int i;
  char *sigt;

  for (i = 0; i < nb; i++) {
    if (first)
      first = 0;
    else
      fprintf(mdfFile,",\n");
    if (sigTab[i]->kind == sPure)
      fprintf(mdfFile, "    %s", sigTab[i]->name);
    else {
	  sigt = NameOfType(sigTab[i]->theType);
      fprintf(mdfFile, "    %s %s", sigt,
              sigTab[i]->name);
	  free(sigt);
	}
  }
  fprintf(mdfFile,";\n");
}

void writeNoDepsInputs(FILE *mdfFile, int nbInp) {
  int i, m, s;
  for(i = 0; i < nbInp; i++) {
    m = linkedModule->inputs[i].modNum;
	s = linkedModule->inputs[i].sigNum;
	if (m < -1)
	  m = -2 - m;
    if ((m == -1) || (linkedModule->subSigs[m].inputs[-1 - s].deps == 0))
      fprintf(mdfFile,"  #nodep %s;\n",
              moduleTab[builtModule].inputs[i]->name);
  }
}

void writeNoDepsOutputs(FILE *mdfFile, int nbOut) {
  int i, m, s;
  for(i = 0; i < nbOut; i++) {
    m = linkedModule->outputs[i].modNum;
	s = linkedModule->outputs[i].sigNum;
    if ((m == -1) || (linkedModule->subSigs[m].outputs[s].deps == 0))
      fprintf(mdfFile,"  #nodep %s;\n",
              moduleTab[builtModule].outputs[i]->name);
  }
}

void writeMdfFile() {
  FILE *mdfFile;
  int nb;
  sig **sigTab;
  typdefList *curTypdef;
  char *ot, *nt, *mt;
  paramList *plist;
  
  if (strlen(mdfBaseName) == 0)
    strcpy(mdfBaseName, moduleTab[builtModule].name->name);
  sprintf(mdfFileName, mdfFileTemplate,"",mdfBaseName);
  if ((mdfFile = fopen(mdfFileName, "w")) == 0) {
    sprintf(tmpStr,"Cannot open file %s for output", mdfFileName);
    fileError(tmpStr);
  }
  fprintf(mdfFile, "//\n//  %s\n// Created by mdlc v%s on %s\n//\n",
                   mdfFileName, mdlVers, dateAndTime());
  fprintf(mdfFile, "#source \"%s\"\n", moduleTab[builtModule].source);
  fprintf(mdfFile, "#c++ \"%s\"\n\n", hFileName);
  
  if (theTypdefs != 0) {
    fprintf(mdfFile,"\n");
    curTypdef = theTypdefs;
    while(curTypdef != 0) {
      ot = NameOfType(curTypdef->theDef->oldType);
      nt = NameOfType(curTypdef->theDef->newType);
      if (curTypdef->theDef->mustDef)
        fprintf(mdfFile,"%s is %s;\n", nt, ot);
      else
        fprintf(mdfFile,"%s as %s;\n", nt, ot);
      curTypdef = curTypdef->next;
      free(ot);
      free(nt);
    }
    fprintf(mdfFile,"\n");
  }
  if (moduleTab[builtModule].params != 0) {
    fprintf(mdfFile,"#param {\n");
    plist = moduleTab[builtModule].params;
    while(plist != 0) {
      nt = NameOfType(plist->val->theType);
      fprintf(mdfFile,"  %s %s;\n", nt, plist->val->name);
      free(nt);
      plist = plist->next;
    }
    fprintf(mdfFile,"}\n\n");
  }
  mt = NameOfType(moduleTab[builtModule].name);
  if (moduleTab[builtModule].argsToConst != 0) {
    nt = NameOfType(moduleTab[builtModule].argsToConst);
    fprintf(mdfFile,"%s(%s) {\n", mt, nt);
    free(nt);
  } else
    fprintf(mdfFile, "%s {\n", mt);
  free(mt);
  nb = moduleTab[builtModule].nbInputs;
  sigTab = moduleTab[builtModule].inputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbInputs;
  if (nb > 0) {
    writeNoDepsInputs(mdfFile, nb);
    fprintf(mdfFile, "  input:\n");
    writeSigTab(mdfFile, nb, sigTab);
  }
  nb = moduleTab[builtModule].nbOutputs;
  sigTab = moduleTab[builtModule].outputs;
  if (moduleTab[builtModule].kind == mDerived)
    nb += moduleTab[moduleTab[builtModule].superMod].nbOutputs;
  if (nb > 0) {
    writeNoDepsOutputs(mdfFile, nb);
    fprintf(mdfFile, "  output:\n");
    writeSigTab(mdfFile, nb, sigTab);
  }
  fprintf(mdfFile,"}\n\n");
  fclose(mdfFile);
}
