%{
/*
//  File mdlParse.l
//
//  September 3 1992 - January 1995
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1995.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

#include "mdlParseTypes.h"
#include "mdlGram.tab.h"

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

/* Prototypes for lex functions */
extern int yyless(int);
extern int yylook();
#ifndef FLEX
extern int yywrap();
#endif
extern int yyback(int* p, int m);

/* StrDup duplicates a C string */
#define StrDup(str) ((char*)strcpy(malloc(strlen(str)+1),str))

/* TRASH expands to the null statement. */
/* It is used to explicit the fact that we trash some input characters */
#define TRASH ;

extern YYSTYPE yylval; /* Union used to send values to the Yacc parser. */

int lineNb = 1;        /* Parse begins at line 1 of source file */
char *dotPtr;          /* Pointer to dot '.' in composed signal names */
int parLevel = 0;      /* Parenthese level for literals */

%}

blank        [\ \t]
alpha        [a-zA-Z]
digit        [0-9]
special      [\_]
ident        (({alpha}|{special})({alpha}|{digit}|{special})*)
litinteger   ({digit}+)

%START COMMENT LITERAL RAW
%%

<INITIAL>\/\/               {BEGIN COMMENT;}
<COMMENT>.                  {TRASH}

<INITIAL>{blank}            {TRASH}
<INITIAL>\n                 {lineNb++;}
<COMMENT>\n                 {lineNb++;BEGIN INITIAL;}

<INITIAL>\,                 {return tComma;}
<INITIAL>\;                 {return tSemiColon;}
<INITIAL>\:                 {return tColon;}
<INITIAL>\{                 {return tLeftBrace;}
<INITIAL>\}                 {return tRightBrace;}
<INITIAL>\(                 {
                              BEGIN LITERAL;
                              parLevel = 0;
                              return tLeftPar;
                            }

<LITERAL>\(                 {parLevel++; yymore();}
<LITERAL>[^()]+             {yymore();}
<LITERAL>\)                 {
                              switch (parLevel) {
                                case 0:
                                  yyless(yyleng - 1);
                                  parLevel--;
                                  yylval.strval = StrDup((char*)yytext);
                                  return tLiteral;
                                  break;
                                case -1:
                                  BEGIN INITIAL;
                                  return tRightPar;
                                  break;
                                default:
                                  parLevel--;
                                  yymore();
                                  break;
                              }
                            }

<INITIAL>\<                 {return tLeftBracket;}
<INITIAL>\>                 {return tRightBracket;}
<INITIAL>\=                 {return tIs;}
<INITIAL>\<\<               {return tConnect;}

<INITIAL>is                 {return tTypdef;}
<INITIAL>as                 {return tTypequ;}
<INITIAL>input:             {return tInputs;}
<INITIAL>output:            {return tOutputs;}
<INITIAL>^#source           {return tSource;}
<INITIAL>^#oc               {return tOc;}
<INITIAL>^#c\+\+            {return tCpp;}
<INITIAL>^#import           {return tImport;}
<INITIAL>#nodep             {return tNoDep;}
<INITIAL>^#use              {return tUse;}
<INITIAL>#param             {return tParam;}
<INITIAL>^#Hinsert          {BEGIN RAW; return tHinsert;}
<INITIAL>^#Cinsert          {BEGIN RAW; return tCinsert;}
<INITIAL>^#HInsert          {BEGIN RAW; return tHinsert;}
<INITIAL>^#CInsert          {BEGIN RAW; return tCinsert;}

<RAW>\n                     {
                              lineNb++;
                              yylval.strval = StrDup((char*)yytext);
                              return tRawLine;
                            }
<RAW>.                      {yymore();}
<RAW>^#endInsert            {BEGIN INITIAL; return tEndInsert;}
<RAW>^#EndInsert            {BEGIN INITIAL; return tEndInsert;}

<INITIAL>{ident}            {
                              yylval.strval = StrDup((char*)yytext);
                              return tIdent;
                            }
<INITIAL>{ident}\.{ident}   {
                              yylval.st2val.str2 = StrDup(strchr((char*)yytext,'.')+1);
                              dotPtr = strchr((char*)yytext,'.');
                              *dotPtr = 0;
                              yylval.st2val.str1 = StrDup((char*)yytext);
                              *dotPtr = '.';
                              return tCompIdent;
                            }
<INITIAL>\"([^"]|\\\")*\"   {
                              yylval.strval = StrDup((char*)yytext+1);
                              yylval.strval[strlen(yylval.strval)-1] = 0;
                              return tString;
                            }

%%

void reinitlex() {
#ifdef FLEX
  yy_init = 1;              /* Tell flex it must start over to get the next token */
#endif
}

