/*
//  File mdlParseTypes.h
//
//  September 3 1992 - December 1994
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1994.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

#ifndef _MDL_PARSETYPES_H_
#define _MDL_PARSETYPES_H_

#include "env.h"

/* This file contains the definitions of parsed elements.        */
/*****************************************************************/

typedef struct {
  char *str1;
  char *str2;
} strpair;

typedef enum {mDeclared, mComposite, mDerived} modType;

struct atype;

typedef struct typeList {
  struct atype *theType;
  struct typeList *next;
} typeList;

typedef struct atype {
  char *name;
  struct typeList *params;
} atype;

typedef struct typdef {
  atype *newType;
  atype *oldType;
  int mustDef;
} typdef;

typedef struct typdefList {
  typdef *theDef;
  struct typdefList *next;
} typdefList;

struct declList;
struct paramList;

typedef struct {
  modType kind;
  char *source;
  char *oc;
  char *cpp;
  char *imp;
  struct paramList *params;
  atype *argsToConst;
  atype *name;
  atype *super;
  struct declList *body;
} module;

typedef enum {sTyped, sPure, sMain, sSub} sigType;

typedef struct {
  sigType kind;
  atype *theType;
  char *owner;
  char *name;
} sig;

typedef struct sigList {
  sig *val;
  struct sigList *next;
} sigList;

typedef struct  {
	atype *theType;
	char *name;
} param;

typedef struct paramList {
	param *val;
	struct paramList *next;
} paramList;

typedef struct {
  char *modName;
  char *litArgs;
} subModInst;

typedef struct modList {
  subModInst *modInst;
  struct modList *next;
} modList;

typedef struct {
  atype *theType;
  modList *theMods;
} subMod ;

typedef struct {
  sig* from;
  sig* to;
} connect;

typedef struct nameList {
  char* name;
  struct nameList *next;
} nameList;

typedef union {
  modList *mod;
  sigList *sig;
  paramList *param;
  subMod *sub;
  connect *con;
  nameList *name;
} List;

typedef enum {dInput, dOutput, dModule, dConnect, dEquiv, dNoDep} declType;

typedef struct {
  declType kind;
  List val;
} decl;

typedef struct declList {
  decl *val;
  struct declList* next;
} declList;

typedef struct parsedList {
  module *parsedMod;
  struct parsedList *next;
} parsedList;

#endif

