/*
//  File mdlProto.c
//
//  September 3 1992 - September 1995
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1994.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

/* This file contains functions for synchronous class prototypes.*/
/*****************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "env.h"
#include "mdl.h"
#include "mdlProto.h"
#include "mdlErrors.h"
#include "mdlUtils.h"
#include "mdlTables.h"

void buildHProto(pMod* m, FILE *f);
void buildCProto(pMod* m, FILE *f);

void buildProtoFiles() {
  int i, buildOne = 0;
  FILE *proto;
  char fName[256];
  
  for(i = 0; i < nbModules; i++) {
    if (moduleTab[i].kind == mDeclared) {
	  buildOne = 1;
      sprintf(fName,"%s.H.proto",moduleTab[i].name->name);
      proto = fopen(fName,"w");
      if (proto == 0) {
        sprintf(tmpStr,"Cannot open file \"%s\" for output.",fName);
        fileError(tmpStr);
      }
      buildHProto(&(moduleTab[i]), proto);
      fclose(proto);
      sprintf(fName,"%s.C.proto",moduleTab[i].name->name);
      proto = fopen(fName,"w");
      if (proto == 0) {
        sprintf(tmpStr,"Cannot open file \"%s\" for output.",fName);
        fileError(tmpStr);
      }
      buildCProto(&(moduleTab[i]), proto);
      fclose(proto);
    }
  }
  if (!buildOne)
    mdfWarn("-proto option ignored. No description to process.");
}

void buildHProto(pMod* m, FILE *f) {
  char buf[256];
  int i;
  typdefList *tdef;
  char *nameOfType_module;
  char *ot, *nt, *sigt;
  paramList *plist;
  
  nameOfType_module = NameOfType(m->name);
  
  for(i = 0; i < strlen(m->name->name); i++)
    buf[i] = toupper(m->name->name[i]);
  buf[i] = 0;
  fprintf(f, "#ifndef _%s_H_\n", buf);
  fprintf(f, "#define _%s_H_\n\n", buf);
  fprintf(f, "//\n// Prototype of interface file for module %s\n",
          nameOfType_module);
  fprintf(f, "// Created by mdlc v%s on %s\n//\n\n",
          mdlVers, dateAndTime());
  if (optIncludeEsterel == 0)
    fprintf(f, "# include %s\n\n", includeEsterel);
  else
    fprintf(f, "# include %s\n\n", optIncludeEsterel);

  tdef = theTypdefs;
  while(tdef != 0) {
    if (tdef->theDef->mustDef) {
      ot = NameOfType(tdef->theDef->oldType);
      nt = NameOfType(tdef->theDef->newType);
      fprintf(f,"typedef %s %s;\n\n", ot, nt);
      free(ot); free(nt);
    }
    tdef = tdef->next;
  }

  if (m->params != 0) {
    fprintf(f,"struct %sConsts {\n", m->name->name);
    plist = m->params;
    while(plist != 0) {
      fprintf(f,"  %s %s;\n",
                NameOfType(plist->val->theType),
                plist->val->name);
      plist = plist->next;
    }
    fprintf(f,"};\n\n");
  }

  printTemplate(f, m->name->params);
  fprintf(f, "class %s : public Synchronous {\n",m->name->name);
  fprintf(f, "  public:\n");
  fprintf(f, "    void resetOutputs();\n");
  fprintf(f, "    void resetInputs();\n");
  fprintf(f, "    void setInputs();\n");
  fprintf(f, "    void activate();\n");
  fprintf(f, "    const InputSignal** inSigList() const;\n");
  fprintf(f, "    const OutputSignal** outSigList() const;\n");
  for(i = 0; i < m->nbInputs; i++) {
    if (m->inputs[i]->kind == sPure) {
      fprintf(f, "    virtual PureInSignal& %s();\n",
              m->inputs[i]->name);
    } else {
      sigt = NameOfType(m->inputs[i]->theType);
      fprintf(f, "    virtual InSignal<%s>& %s();\n",
                 sigt, m->inputs[i]->name);
      free(sigt);
    }
  }
  for(i = 0; i < m->nbOutputs; i++) {
    if (m->outputs[i]->kind == sPure)
      fprintf(f, "    virtual PureOutSignal& %s();\n",
                 m->outputs[i]->name);
    else {
      sigt = NameOfType(m->outputs[i]->theType);
      fprintf(f, "    virtual OutSignal<%s>& %s();\n",
                 sigt, m->outputs[i]->name);
      free(sigt);
    }
  }
  fprintf(f, "\n    virtual void backup();\n");
  fprintf(f, "    virtual void restore();\n");
  fprintf(f, "    %s& operator=(const %s &);\n", nameOfType_module, nameOfType_module);
  if (m->argsToConst == 0) {
    fprintf(f, "\n    %s(const char* name = \"%s\", const bool schedule = true);\n",
               m->name->name, m->name->name);
    if (m->params != 0) {
      fprintf(f, "    virtual void setParams(const %sConsts& newParams);\n", m->name->name);
      fprintf(f, "    virtual void getParams(%sConsts& theParams) const;\n", m->name->name);
    }
  } else {
    if (strcmp(m->argsToConst->name,"#param") == 0)
      fprintf(f,"\n    %s(const %sConsts& initInstance, const char* name = \"%s\", const bool schedule = true);\n",
               m->name->name, m->name->name, m->name->name);
    else {
      fprintf(f,"\n    %s(%s initInstance, const char* name = \"%s\", const bool schedule = true);\n",
               m->name->name, NameOfType(m->argsToConst), m->name->name);
      if (m->params != 0) {
        fprintf(f, "    virtual void setParams(const %sConsts& newParams);\n", m->name->name);
        fprintf(f, "    virtual void getParams(%sConsts& theParams) const;\n", m->name->name);
      }
    }
  }
  fprintf(f, "    ~%s();\n", m->name->name);
  fprintf(f, "\n  private:\n");
  fprintf(f, "  %s(const %s &);\n",m->name->name,nameOfType_module);
  
  for(i = 0; i < m->nbInputs; i++) {
    if (m->inputs[i]->kind == sPure)
      fprintf(f, "    PureInSignal f%s;\n", m->inputs[i]->name);
    else {
      sigt = NameOfType(m->inputs[i]->theType);
      fprintf(f, "    InSignal<%s> f%s;\n", sigt, m->inputs[i]->name);
      free(sigt);
    }
  }

  for(i = 0; i < m->nbOutputs; i++) {
    if (m->outputs[i]->kind == sPure)
      fprintf(f, "    PureOutSignal f%s;\n", m->outputs[i]->name);
    else {
      sigt = NameOfType(m->outputs[i]->theType);
      fprintf(f, "    OutSignal<%s> f%s;\n", sigt,
                 m->outputs[i]->name);
      free(sigt);
    }
  }

  if (m->params != 0) {
    fprintf(f,"// Parameters\n");
    plist = m->params;
    while(plist != 0) {
      nt = NameOfType(plist->val->theType);
      fprintf(f, "    %s %s;\n",nt,plist->val->name);
      free(nt);
      plist = plist->next;
    }
    fprintf(f,"\n");
  }
  
  if (m->argsToConst != 0) {
  	if (strcmp(m->argsToConst->name,"#param") == 0) {
  		fprintf(f, "    %sConsts fInitInstance;\n",m->name->name);
  	} else {
  		fprintf(f, "    %s fInitInstance;\n",NameOfType(m->argsToConst));
  	}
  }
  
  fprintf(f, "    %s *pf_backup_copy_;\n", nameOfType_module);
  fprintf(f, "};\n");
  
  if (strcmp(nameOfType_module, m->name->name) != 0) {
    fprintf(f, "\n#ifdef __GNUG__\n#include \"");
    fprintf(f, "%s.C", m->name->name);
    fprintf(f, "\"\n#endif\n");
  }
  
  fprintf(f, "\n#endif\n");
  free(nameOfType_module);
}

void buildCProto(pMod* m, FILE *f) {
  int i, j;
  char buf[256];
  char *t, *sigt;
  paramList *plist;
  
  t = NameOfType(m->name);
  
  for(i = 0; i < strlen(m->name->name); i++)
  	buf[i] = toupper(m->name->name[i]);
  buf[i] = 0;
  fprintf(f, "#ifndef _%s_C_\n", buf);
  fprintf(f, "#define _%s_C_\n\n", buf);

  
  fprintf(f, "//\n// Prototype of code file for module %s\n",t);
  fprintf(f, "// Created by mdlc v%s on %s\n//\n\n",
          mdlVers, dateAndTime());
  fprintf(f, "# include \"%s.H\"\n\n", m->name->name);
  
  printTemplate(f, m->name->params);
  fprintf(f, "void %s::resetOutputs() {\n",t);
  for(i = 0; i < m->nbOutputs; i++) {
    for(j = 0; j < m->nbNoDep; j++) {
      if (strcmp(m->outputs[i]->name, m->noDeps[j]) == 0) {
        fprintf(f, "//");
        break;
      }
    }
    fprintf(f, "  f%s.value().reset();\n", m->outputs[i]->name);
  }
  fprintf(f, "}\n\n");
 
  printTemplate(f, m->name->params);
  fprintf(f, "void %s::resetInputs() {\n", t);
  for(i = 0; i < m->nbInputs; i++) {
    for(j = 0; j < m->nbNoDep; j++) {
      if (strcmp(m->inputs[i]->name, m->noDeps[j]) == 0) {
        fprintf(f, "//");
        break;
      }
    }
    fprintf(f, "  f%s.value().reset();\n",m->inputs[i]->name);
  }
  fprintf(f, "}\n\n");
 
  printTemplate(f, m->name->params);
  fprintf(f, "void %s::setInputs() {\n", t);
  for(i = 0; i < m->nbInputs; i++) {
    for(j = 0; j < m->nbNoDep; j++) {
      if (strcmp(m->inputs[i]->name, m->noDeps[j]) == 0) {
        fprintf(f, "//");
        break;
      }
    }
    fprintf(f, "  f%s.update();\n",m->inputs[i]->name);
  }
  fprintf(f, "}\n\n");
  
  printTemplate(f, m->name->params);
  fprintf(f, "void %s::activate() {\n", t);
  fprintf(f, "// Compute new state and outputs versus current state and inputs\n");
  fprintf(f, "}\n\n");
  
  printTemplate(f, m->name->params);
  fprintf(f, "const InputSignal** %s::inSigList() const {\n", t);
  fprintf(f, "  const InputSignal** inList;\n\n");
  fprintf(f, "  inList = (const InputSignal**) new (InputSignal*[%d]);\n", m->nbInputs + 1);
  for(i = 0; i < m->nbInputs; i++)
    fprintf(f, "  inList[%d] = &f%s;\n", i, m->inputs[i]->name);
  fprintf(f, "  inList[%d] = 0;\n", i);
  fprintf(f, "  return inList;\n}\n\n");
  
  printTemplate(f, m->name->params);
  fprintf(f, "const OutputSignal** %s::outSigList() const {\n", t);
  fprintf(f, "  const OutputSignal** outList;\n\n");
  fprintf(f, "  outList = (const OutputSignal**) new (OutputSignal*[%d]);\n", m->nbOutputs + 1);
  for(i = 0; i < m->nbOutputs; i++)
    fprintf(f, "  outList[%d] = &f%s;\n", i, m->outputs[i]->name);
  fprintf(f, "  outList[%d] = 0;\n", i);
  fprintf(f, "  return outList;\n}\n\n");
  
  for(i = 0; i < m->nbInputs; i++) {
    printTemplate(f, m->name->params);
    if (m->inputs[i]->kind == sPure)
      fprintf(f, "PureInSignal& ");
    else {
	  sigt = NameOfType(m->inputs[i]->theType);
      fprintf(f, "InSignal<%s>& ", sigt);
	  free(sigt);
	}
    fprintf(f, "%s::%s() {\n", t,m->inputs[i]->name);
    fprintf(f, "  return f%s;\n", m->inputs[i]->name);
    fprintf(f, "}\n\n");
  }
  
  for(i = 0; i < m->nbOutputs; i++) {
    printTemplate(f, m->name->params);
    if (m->outputs[i]->kind == sPure)
      fprintf(f, "PureOutSignal& ");
    else {
	  sigt = NameOfType(m->outputs[i]->theType);
      fprintf(f, "OutSignal<%s>& ", sigt);
	  free(sigt);
	}
    fprintf(f, "%s::%s() {\n", t,m->outputs[i]->name);
    fprintf(f, "  return f%s;\n", m->outputs[i]->name);
    fprintf(f, "}\n\n");
  }
  
  printTemplate(f, m->name->params);
  fprintf(f, "void %s::backup() {\n", t);
  fprintf(f, "  if (pf_backup_copy_ == 0) {\n");
  
  if (m->params != 0) {
  	fprintf(f, "    pf_backup_copy_ = new %s(fInitInstance, \"backup\", false);\n",t);
  } else {
  	fprintf(f, "    pf_backup_copy_ = new %s(\"backup\", false);\n",t);
  }
  
  fprintf(f, "  }\n");
  fprintf(f, "  *pf_backup_copy_ = *this;\n");
  fprintf(f, "}\n\n");
  
  printTemplate(f, m->name->params);
  fprintf(f, "void %s::restore() {\n", t);
  fprintf(f, "  if (pf_backup_copy_ != 0) {\n");
  fprintf(f, "    *this = *pf_backup_copy_;\n");
  fprintf(f, "  }\n");
  fprintf(f, "}\n\n");
  
  printTemplate(f, m->name->params);
  fprintf(f, "%s& %s::operator=(const %s& m) {\n", t, t, t);
  if ((m->params != 0)&&
      ((m->argsToConst == 0)||
       (strcmp(m->argsToConst->name,"#param") != 0))) {
    fprintf(f, "  %sConsts the%sConsts;\n", m->name->name, m->name->name);
    fprintf(f, "  m.getParams(the%sConsts);\n", m->name->name);
    fprintf(f, "  setParams(the%sConsts);\n", m->name->name);
  }
  for(i = 0; i < m->nbInputs; i++)
    fprintf(f, "  f%s.value() = ((%s&)m).f%s.value();\n",
                  m->inputs[i]->name, t,    m->inputs[i]->name);
  for(i = 0; i < m->nbOutputs; i++)
    fprintf(f, "  f%s.value() = ((%s&)m).f%s.value();\n",
                  m->outputs[i]->name, t,    m->outputs[i]->name);
  fprintf(f, "// You should also copy any other part of the state\n");
  fprintf(f, "// of the module: variables, automaton state, ...\n");
  fprintf(f, "  return *this;\n");
  fprintf(f, "}\n\n");
  
  printTemplate(f, m->name->params);
  if (m->argsToConst == 0) {
    fprintf(f, "%s::%s(const char* name, const bool schedule)\n",
               t, m->name->name);
  } else {
    if (strcmp(m->argsToConst->name, "#param") == 0) {
      fprintf(f, "%s::%s(const %sConsts& initInstance, const char* name, const bool schedule)\n",
                 t, m->name->name, m->name->name);
    } else {
      sigt = NameOfType(m->argsToConst);
      fprintf(f, "%s::%s(%s initInstance, const char* name, const bool schedule)\n",
                 t, m->name->name, sigt);
      free(sigt);
    }
  }
  fprintf(f, "        : Synchronous(name, schedule)");
  for(i = 0; i < m->nbInputs; i++) {
    fprintf(f, ",\n          f%s(this, \"%s\"",
            m->inputs[i]->name, m->inputs[i]->name);
    for(j = 0; j < m->nbNoDep; j++) {
      if (strcmp(m->inputs[i]->name, m->noDeps[j]) == 0) {
        fprintf(f, ", false");
        break;
      }
    }
    fprintf(f, ")");
  }
  for(i = 0; i < m->nbOutputs; i++) {
    fprintf(f, ",\n          f%s(this, \"%s\"",
            m->outputs[i]->name, m->outputs[i]->name);
    for(j = 0; j < m->nbNoDep; j++) {
      if (strcmp(m->outputs[i]->name, m->noDeps[j]) == 0) {
        fprintf(f, ", false");
        break;
      }
    }
    fprintf(f,")");
  }
  if ((m->params != 0)&&(m->argsToConst != 0)&&(strcmp(m->argsToConst->name,"#param") == 0)) {
    plist = m->params;
    while(plist != 0) {
      fprintf(f,",\n          %s(initInstance.%s)",
                 plist->val->name, plist->val->name);
      plist = plist->next;
    }
  }
  if (m->argsToConst != 0) {
  	fprintf(f, ",\n          fInitInstance(initInstance)");
  }
  fprintf(f, ",\n          pf_backup_copy_(0)");
  fprintf(f, " {\n");
  fprintf(f, "//  Do any initializations here\n");
  fprintf(f, "}\n\n");
  
  if ((m->params != 0)&&(m->argsToConst != 0)&&(strcmp(m->argsToConst->name,"#param") != 0)) {
    fprintf(f,"void %s::setParams(const %sConsts& newParams) {\n",
               t,m->name->name);
    plist = m->params;
    while(plist != 0) {
      fprintf(f,"  %s = newParams.%s;\n", plist->val->name, plist->val->name);
      plist = plist->next;
    }
    fprintf(f,"};\n\n");
    fprintf(f,"void %s::getParams(%sConsts& theParams) const {\n",
               t,m->name->name);
    plist = m->params;
    while(plist != 0) {
      fprintf(f,"  theParams.%s = %s;\n", plist->val->name, plist->val->name);
      plist = plist->next;
    }
    fprintf(f,"};\n\n");
  }
  printTemplate(f, m->name->params);
  fprintf(f, "%s::~%s() {\n", t, m->name->name);
  fprintf(f, "  if (pf_backup_copy_ != 0)\n");
  fprintf(f, "    delete pf_backup_copy_;\n");
  fprintf(f, "// Do anything you should do when an object dies\n");
  fprintf(f, "}\n\n");
  fprintf(f, "#endif\n");
}
