/*
//  File mdlSimul.c
//
//  July 1994 - December 1994
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1994.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "env.h"
#include "mdl.h"
#include "mdlSimul.h"
#include "mdlErrors.h"
#include "mdlUtils.h"
#include "mdlTables.h"

void buildT(pMod* m, FILE *f);
void buildX(pMod* m, FILE *f);
void buildP(pMod* m, FILE *f);
void GNUDeclarePlug(pMod *m, FILE *f, atype *instdef, char *macro);
void CXXInstanciate(pMod *m, FILE *f, int Graph, atype *instdef, char *sym);
int isPrimType(atype *t, atype *instdef, atype *tmpl);

char *instTypesStr = 0;

void buildTextSimul() {
  int i;
  FILE *proto;
  char fName[256];
  
  for(i = 0; i < nbModules; i++) {
    if (moduleTab[i].kind == mDeclared) {
      sprintf(fName,"%sSimulT.C",moduleTab[i].name->name);
      proto = fopen(fName,"w");
      if (proto == 0) {
        sprintf(tmpStr,"Cannot open file \"%s\" for output.",fName);
        fileError(tmpStr);
      }
      buildT(&(moduleTab[i]), proto);
      fclose(proto);
    }
  }
}

void buildXSimul() {
  int i;
  FILE *proto;
  char fName[256];
  
  for(i = 0; i < nbModules; i++) {
    if (moduleTab[i].kind == mDeclared) {
      sprintf(fName,"%sSimulX.C",moduleTab[i].name->name);
      proto = fopen(fName,"w");
      if (proto == 0) {
        sprintf(tmpStr,"Cannot open file \"%s\" for output.",fName);
        fileError(tmpStr);
      }
      buildX(&(moduleTab[i]), proto);
      fclose(proto);
    }
  }
}

void buildPSimul() {
  int i;
  FILE *proto;
  char fName[256];
  
  for(i = 0; i < nbModules; i++) {
    if (moduleTab[i].kind == mDeclared) {
      sprintf(fName,"%sSimulP.C",moduleTab[i].name->name);
      proto = fopen(fName,"w");
      if (proto == 0) {
        sprintf(tmpStr,"Cannot open file \"%s\" for output.",fName);
        fileError(tmpStr);
      }
      buildP(&(moduleTab[i]), proto);
      fclose(proto);
    }
  }
}

atype *readType(char **s, char *buf) {
	atype *t;
	char *p, *q, *n;
	
	p = *s;
	q = buf;
	while ((*p != 0) && (*p != ',') && (*p != '<') && (*p != '>'))
		*(q++) = *(p++);
	*q = 0;
	n = checkAlloc(strlen(buf)+1,1);
	strcpy(n,buf);
	t = newType(n,0);
	if (*p == '<') {
		do {
			p++;
			t->params = addType(t->params,readType(&p,buf));
		} while(*p == ',');
		if (*p != '>') {
			fatalError("Badly formatted list of types for -type.");
		}
		p++;
	}
	*s = p;
	return t;
}

atype *parseTypeOpt(char *name, char *s) {
	char buf[64];
	char *p;
	atype *t;
	
	t = newType(name,0);
	p = s-1;
	do {
		p++;
		t->params = addType(t->params, readType(&p,buf));
	} while(*p == ',');
	if (*p != 0) {
		fatalError("Badly formatted list of types for -type.");
	}
	return t;
}

void buildT(pMod* m, FILE *f) {
	int i, first = 1;
	typdefList *tdef;
	char *fullName;
	char *ot, *nt;
	atype *instdef;
	paramList *plist;
	
	if (m->name->params != 0) {	/* If template */
		if (instTypesStr == 0) {
			fatalError("Module is a template. Specify its type with -type.");
		}
		instdef = parseTypeOpt(m->name->name,instTypesStr);
		if(compatTypes(m->name,instdef) != 0) {
			sprintf(tmpStr,"Wrong type list for template %s.",
					NameOfType(m->name));
			fatalError(tmpStr);
		}
	} else {
		instdef = m->name;
	}
	
	fullName = NameOfType(instdef);
	fprintf(f,"//\n// Text oriented simulation program for %s\n", fullName);
	fprintf(f,"// Created by mdlc v%s on %s\n//\n\n",
				mdlVers, dateAndTime());

	if ((m->cpp != 0) && (m->cpp[0] != 0))
		fprintf(f,"#include \"%s\"\n\n", m->cpp);
	else {
		if (optIncludeEsterel == 0)
			fprintf(f, "# include %s\n\n", includeEsterel);
		else
			fprintf(f, "# include %s\n\n", optIncludeEsterel);
	}

	tdef = theTypdefs;
	while(tdef != 0) {
		if (tdef->theDef->mustDef) {
		  ot = NameOfType(tdef->theDef->oldType);
		  nt = NameOfType(tdef->theDef->newType);
		  fprintf(f,"typedef %s %s;\n\n", ot, nt);
		  free(ot); free(nt);
		}
		tdef = tdef->next;
	}
	CXXInstanciate(m,f,0,instdef,"");
	fprintf(f,"int main() {\n");
	fprintf(f,"  SScheduler clock(\"%s\",true);\n", m->name->name);

	if (m->params != 0) {
	  fprintf(f,"  %sConsts the%sConsts;\n",
	             m->name->name, m->name->name);
    fprintf(f,"  cout << \"Module %s requires parameters. Please enter them:\" << endl;\n",
               m->name->name);
    fprintf(f,"  Synchronous *getArg;\n");
      
    plist = m->params;
    while(plist != 0) {
      nt = NameOfType(plist->val->theType);
      fprintf(f,"  getArg = new InputModule<%s>(\"%s (%s):\",\"\",false,\"\",false);\n",
                 nt, plist->val->name, nt);
      fprintf(f, "  getArg->react();\n");
      fprintf(f, "  the%sConsts.%s = ((InputModule<%s>*)getArg)->input().value().value();\n",
                    m->name->name, plist->val->name, nt);
      free(nt);
      fprintf(f, "  delete getArg;\n");
      plist = plist->next;
    }
    fprintf(f,"\n");
	}
  if (m->argsToConst != 0) {
    if (strcmp(m->argsToConst->name, "#param") == 0) {
      fprintf(f,"  %s the%s(the%sConsts, \"%s\");\n\n",
                   fullName, m->name->name,
                   m->name->name, m->name->name);
    } else  {
      fprintf(f,"  %s the%s(/* %s */, \"%s\");\n\n",
                   fullName, m->name->name,
                   NameOfType(m->argsToConst),
                   m->name->name);
      mdfWarn("You will have to edit the simulation program to set a parameter of the module.");
      if (m->params != 0)
        fprintf(f,"  the%s.setParams(the%sConsts);\n\n",
                  m->name->name, m->name->name);
    }
  } else {
      fprintf(f,"  %s the%s(\"%s\");\n",
                fullName, m->name->name, m->name->name);
    if (m->params != 0)
      fprintf(f,"  the%s.setParams(the%sConsts);\n\n",
                m->name->name, m->name->name);
    else
      fprintf(f,"\n");
  }

	for(i = 0; i < m->nbInputs; i++) {
		if (m->inputs[i]->kind == sPure)
			fprintf(f,"  PureInModule i_%s(\"%s: \");\n",
				m->inputs[i]->name, m->inputs[i]->name);
		else
			fprintf(f, "  InputModule<%s> i_%s(\"%s: \");\n",
				resolveTemplate(m->inputs[i]->theType,
					instdef, m->name),
				m->inputs[i]->name, m->inputs[i]->name);
		fprintf(f, "  the%s.%s() << i_%s();\n",
			m->name->name,
			m->inputs[i]->name,
			m->inputs[i]->name);
	}
	for(i = 0; i < m->nbOutputs; i++) {
		if (m->outputs[i]->kind == sPure)
			fprintf(f, "  PureOutModule o_%s(\"%s: \");\n",
				m->outputs[i]->name, m->outputs[i]->name);
		else
			fprintf(f,"  OutputModule<%s> o_%s(\"%s: \");\n",
				resolveTemplate(m->outputs[i]->theType,
					instdef, m->name),
				m->outputs[i]->name, m->outputs[i]->name);
		fprintf(f,"  o_%s() << the%s.%s();\n",
			m->outputs[i]->name,
			m->name->name,
			m->outputs[i]->name);
	}
	if ((m->nbInputs + m->nbOutputs) > 1) {
		for(i = 0; i < m->nbInputs; i++) {
			if (first) {
				fprintf(f,"\n  i_%s", m->inputs[i]->name);
				first = 0;
			} else
				fprintf(f," << i_%s", m->inputs[i]->name);
		}
		for(i = 0; i < m->nbOutputs; i++) {
			if (first) {
				fprintf(f,"\n  o_%s", m->outputs[i]->name);
				first = 0;
			} else
				fprintf(f," << o_%s", m->outputs[i]->name);
		}
		fprintf(f,";\n");
	}
	fprintf(f,"\n  while(clock.tick())\n    ;\n\n  return 0;\n}\n");
	free(fullName);
}

void buildX(pMod* m, FILE *f) {
	int i;
	typdefList *tdef;
	char *fullName;
	char *ot, *nt;
	atype *instdef;
	paramList *plist;
	
	if (m->name->params != 0) {	/* If template */
		if (instTypesStr == 0) {
			fatalError("Module is a template. Specify its type with -type.");
		}
		instdef = parseTypeOpt(m->name->name,instTypesStr);
		if(compatTypes(m->name,instdef) != 0) {
			sprintf(tmpStr,"Wrong type list for template %s.",
					NameOfType(m->name));
			fatalError(tmpStr);
		}
	} else {
		instdef = m->name;
	}
	
	fullName = NameOfType(instdef);
	fprintf(f,"//\n// Graphic simulation program for %s\n", fullName);
	fprintf(f,"// Created by mdlc v%s on %s\n//\n\n",
				mdlVers, dateAndTime());

	if ((m->cpp != 0) && (m->cpp[0] != 0))
		fprintf(f,"#include \"%s\"\n\n", m->cpp);
	else {
		if (optIncludeEsterel == 0)
			fprintf(f, "# include %s\n\n", includeEsterel);
		else
			fprintf(f, "# include %s\n\n", optIncludeEsterel);
	}

	tdef = theTypdefs;
	while(tdef != 0) {
		if (tdef->theDef->mustDef) {
		  ot = NameOfType(tdef->theDef->oldType);
		  nt = NameOfType(tdef->theDef->newType);
		  fprintf(f,"typedef %s %s;\n\n", ot, nt);
		  free(ot); free(nt);
		}
		tdef = tdef->next;
	}
	GNUDeclarePlug(m,f,instdef,"_GNUG_DECLARE_PLUG_");
	CXXInstanciate(m,f,1,instdef, "_XINTERFACE_INSTANCIATE_");
	fprintf(f,"int main(int argc, char **argv) {\n");
	fprintf(f,"  XInterface x(argc, argv, \"%s\");\n", m->name->name);

	if (m->params != 0) {
	  fprintf(f,"  %sConsts the%sConsts;\n",
	             m->name->name, m->name->name);
    fprintf(f,"  cout << \"Module %s requires parameters. Please enter them:\" << endl;\n",
               m->name->name);
    fprintf(f,"  Synchronous *getArg;\n");
      
    plist = m->params;
    while(plist != 0) {
      nt = NameOfType(plist->val->theType);
      fprintf(f,"  getArg = new InputModule<%s>(\"%s (%s):\",\"\",false,\"\",false);\n",
                 nt, plist->val->name, nt);
      fprintf(f, "  getArg->react();\n");
      fprintf(f, "  the%sConsts.%s = ((InputModule<%s>*)getArg)->input().value().value();\n",
                    m->name->name, plist->val->name, nt);
      free(nt);
      fprintf(f, "  delete getArg;\n");
      plist = plist->next;
    }
    fprintf(f,"\n");
	}
  if (m->argsToConst != 0) {
    if (strcmp(m->argsToConst->name, "#param") == 0) {
      fprintf(f,"  %s the%s(the%sConsts, \"%s\");\n\n",
                   fullName, m->name->name,
                   m->name->name, m->name->name);
    } else  {
      fprintf(f,"  %s the%s(/* %s */, \"%s\");\n\n",
                   fullName, m->name->name,
                   NameOfType(m->argsToConst),
                   m->name->name);
      if (m->params != 0)
        fprintf(f,"  the%s.setParams(the%sConsts);\n\n",
                  m->name->name, m->name->name);
      mdfWarn("You will have to edit the simulation program to set a parameter of the module.");
    }
  } else {
      fprintf(f,"  %s the%s(\"%s\");\n",
                fullName, m->name->name, m->name->name);
    if (m->params != 0)
      fprintf(f,"  the%s.setParams(the%sConsts);\n\n",
                m->name->name, m->name->name);
    else
      fprintf(f,"\n");
  }

	for(i = 0; i < m->nbInputs; i++) {
		fprintf(f,"  plug(x, the%s.%s());\n",m->name->name,m->inputs[i]->name);
	}
	for(i = 0; i < m->nbOutputs; i++) {
		fprintf(f,"  plug(x, the%s.%s());\n",m->name->name,m->outputs[i]->name);
	}
	fprintf(f,"\n  XInterface::runsimul();\n\n  return 0;\n}\n");
	free(fullName);
}

void buildP(pMod* m, FILE *f) {
	int i;
	typdefList *tdef;
	char *fullName;
	char *ot, *nt;
	atype *instdef;
	paramList *plist;
	
	if (m->name->params != 0) {	/* If template */
		if (instTypesStr == 0) {
			fatalError("Module is a template. Specify its type with -type.");
		}
		instdef = parseTypeOpt(m->name->name,instTypesStr);
		if(compatTypes(m->name,instdef) != 0) {
			sprintf(tmpStr,"Wrong type list for template %s.",
					NameOfType(m->name));
			fatalError(tmpStr);
		}
	} else {
		instdef = m->name;
	}
	
	fullName = NameOfType(instdef);
	fprintf(f,"//\n// Simulation protocol program for %s\n", fullName);
	fprintf(f,"// Created by mdlc v%s on %s\n//\n\n",
				mdlVers, dateAndTime());

	if ((m->cpp != 0) && (m->cpp[0] != 0))
		fprintf(f,"#include \"%s\"\n\n", m->cpp);
	else {
		if (optIncludeEsterel == 0)
			fprintf(f, "# include %s\n\n", includeEsterel);
		else
			fprintf(f, "# include %s\n\n", optIncludeEsterel);
	}

	tdef = theTypdefs;
	while(tdef != 0) {
		if (tdef->theDef->mustDef) {
		  ot = NameOfType(tdef->theDef->oldType);
		  nt = NameOfType(tdef->theDef->newType);
		  fprintf(f,"typedef %s %s;\n\n", ot, nt);
		  free(ot); free(nt);
		}
		tdef = tdef->next;
	}
	GNUDeclarePlug(m,f,instdef, "_GNUG_DECLARE_PROTOPLUG_");
	CXXInstanciate(m,f,1,instdef, "_PROTOSIM_INSTANCIATE_");
	fprintf(f,"int main() {\n");
	fprintf(f,"  ProtoSim x(\"%s\");\n", m->name->name);

	if (m->params != 0) {
    mdfWarn("You will have to edit the simulation program to set the parameters of the module.");
	  fprintf(f,"  %sConsts the%sConsts;\n",
	             m->name->name, m->name->name);
      
    plist = m->params;
    while(plist != 0) {
      nt = NameOfType(plist->val->theType);
      fprintf(f, "  the%sConsts.%s = /* Put a \'%s\' here */;\n",
                    m->name->name, plist->val->name, nt);
      free(nt);
      plist = plist->next;
    }
    fprintf(f,"\n");
	}
  if (m->argsToConst != 0) {
    if (strcmp(m->argsToConst->name, "#param") == 0) {
      fprintf(f,"  %s the%s(the%sConsts, \"%s\");\n\n",
                   fullName, m->name->name,
                   m->name->name, m->name->name);
    } else  {
      fprintf(f,"  %s the%s(/* %s */, \"%s\");\n\n",
                   fullName, m->name->name,
                   NameOfType(m->argsToConst),
                   m->name->name);
      if (m->params != 0)
        fprintf(f,"  the%s.setParams(the%sConsts);\n\n",
                  m->name->name, m->name->name);
      mdfWarn("You will have to edit the simulation program to set a parameter of the constructor.");
    }
  } else {
      fprintf(f,"  %s the%s(\"%s\");\n",
                fullName, m->name->name, m->name->name);
    if (m->params != 0)
      fprintf(f,"  the%s.setParams(the%sConsts);\n\n",
                m->name->name, m->name->name);
    else
      fprintf(f,"\n");
  }

	for(i = 0; i < m->nbInputs; i++) {
		if (m->inputs[i]->kind == sPure)
			fprintf(f,"  plug(x, the%s.%s());\n",m->name->name,m->inputs[i]->name);
		else {
			fprintf(f,"  plug(x, the%s.%s(), \"%s\");\n",m->name->name,
							m->inputs[i]->name,
							resolveTemplate(m->inputs[i]->theType,instdef,m->name));
		}
	}
	for(i = 0; i < m->nbOutputs; i++) {
		if (m->outputs[i]->kind == sPure)
			fprintf(f,"  plug(x, the%s.%s());\n",m->name->name,m->outputs[i]->name);
		else {
			fprintf(f,"  plug(x, the%s.%s(), \"%s\");\n",m->name->name,
							m->outputs[i]->name,
							resolveTemplate(m->outputs[i]->theType,instdef,m->name));
		}
	}
	fprintf(f,"\n  ProtoSim::runsimul();\n\n  return 0;\n}\n");
	free(fullName);
}

void GNUDeclarePlug(pMod *m, FILE *f, atype *instdef, char *macro) {
	int i,j,done,first = 1;
	atype *curType;
	
	for(i = 0; i < m->nbInputs; i++) {
		if (m->inputs[i]->kind == sPure)
			continue;
		curType = m->inputs[i]->theType;
		done = 0;
		for(j = 0; j < i; j++) {
			if (m->inputs[j]->kind == sPure)
				continue;
			if (compareTypes(curType,m->inputs[j]->theType) == 0) {
				done = 1;
				break;
			}
		}
		if (!done) {
			if (first) {
				fprintf(f,"#ifdef __GNUG__\n");
				first = 0;
			}
			fprintf(f,"%s(%s)\n", macro,
				resolveTemplate(curType, instdef, m->name));
		}
	}
	for(i = 0; i < m->nbOutputs; i++) {
		if (m->outputs[i]->kind == sPure)
			continue;
		curType = m->outputs[i]->theType;
		done = 0;
		for(j = 0; j < m->nbInputs; j++) {
			if (m->inputs[j]->kind == sPure)
				continue;
			if (compareTypes(curType,m->inputs[j]->theType) == 0) {
				done = 1;
				break;
			}
		}
		if (!done) {
			for(j = 0; j < i; j++) {
				if (m->outputs[j]->kind == sPure)
					continue;
				if (compareTypes(curType,m->outputs[j]->theType) == 0) {
					done = 1;
					break;
				}
			}
		}
		if (!done) {
			if (first) {
				fprintf(f,"#ifdef __GNUG__\n");
				first = 0;
			}
			fprintf(f,"%s(%s)\n", macro, 
				resolveTemplate(curType, instdef, m->name));
		}
	}
	if (!first)
		fprintf(f,"#endif\n\n");
}

void CXXInstanciate(pMod *m, FILE *f, int Graph, atype *instdef, char *sym) {
	int i,j,done,first = 1;
	atype *curType;
	
	for(i = 0; i < m->nbInputs; i++) {
		if (m->inputs[i]->kind == sPure)
			continue;
		curType = m->inputs[i]->theType;
		done = 0;
		for(j = 0; j < i; j++) {
			if (m->inputs[j]->kind == sPure)
				continue;
			if (compareTypes(curType,m->inputs[j]->theType) == 0) {
				done = 1;
				break;
			}
		}
		if(!done)
			done = isPrimType(curType,instdef,m->name);
		if (!done) {
			if(first) {
				fprintf(f,"#ifdef __DECCXX\n");
				if (Graph)
					fprintf(f,"#define %s\n",sym);
				first = 0;
			}
			fprintf(f,"#define _INSTANCIATE_TYPE_ %s\n",
				resolveTemplate(curType, instdef, m->name));
			if (optCxxInstanciate == 0)
				fprintf(f,"#include %s\n",cxxInstanciate);
			else
				fprintf(f,"#include %s\n",optCxxInstanciate);
		}
	}
	for(i = 0; i < m->nbOutputs; i++) {
		if (m->outputs[i]->kind == sPure)
			continue;
		curType = m->outputs[i]->theType;
		done = 0;
		for(j = 0; j < m->nbInputs; j++) {
			if (m->inputs[j]->kind == sPure)
				continue;
			if (compareTypes(curType,m->inputs[j]->theType) == 0) {
				done = 1;
				break;
			}
		}
		if (!done) {
			for(j = 0; j < i; j++) {
				if (m->outputs[j]->kind == sPure)
					continue;
				if (compareTypes(curType,m->outputs[j]->theType) == 0) {
					done = 1;
					break;
				}
			}
		}
		if(!done)
			done = isPrimType(curType,instdef,m->name);
		if (!done) {
			if(first) {
				fprintf(f,"#ifdef __DECCXX\n");
				if (Graph)
					fprintf(f,"#define _XINTERFACE_INSTANCIATE_\n");
				first = 0;
			}
			fprintf(f,"#define _INSTANCIATE_TYPE_ %s\n",
				resolveTemplate(curType,instdef,m->name));
			if (optCxxInstanciate == 0)
				fprintf(f,"#include %s\n",cxxInstanciate);
			else
				fprintf(f,"#include %s\n",optCxxInstanciate);
		}
	}
	if (!first)
		fprintf(f,"#endif\n\n");
}

int isPrimType(atype *t, atype *instdef, atype *tmpl) {
	char *typName;
	
	typName = resolveTemplate(t, instdef, tmpl);

	if (strcmp(typName,"int") == 0)
		return 1;

	if (strcmp(typName,"integer") == 0)
		return 1;

	if (strcmp(typName,"SBoolean") == 0)
		return 1;

	if (strcmp(typName,"boolean") == 0)
		return 1;

	if (strcmp(typName,"bool") == 0)
		return 1;

	if (strcmp(typName,"float") == 0)
		return 1;

	if (strcmp(typName,"double") == 0)
		return 1;

	if (strcmp(typName,"SString") == 0)
		return 1;

	if (strcmp(typName,"string") == 0)
		return 1;

	return 0;
}
