/*
//  File mdlTables.c
//
//  September 3 1992 - December 1994
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1994.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

#include "mdlParseTypes.h"
#include "mdlUtils.h"
#include "mdl.h"
#include "mdlErrors.h"
#include <stdio.h>
#include <string.h>

parsedList *parsedModules = 0;
typdefList *theTypdefs = 0;
nameList *HInsertData = 0;
nameList *CInsertData = 0;
nameList *useList = 0;

void addTypdef(atype *newType, atype *oldType, int isTypdef);

void declareModule(module *newModule){
  parsedList *newParsed, *curParsed;
  if ((newModule->source == 0)||(strlen(newModule->source) == 0))
    newModule->source = sourceFile;
  newParsed = (parsedList*)checkAlloc(1, sizeof(parsedList));
  newParsed->next = 0;
  newParsed->parsedMod = newModule;
  if (parsedModules == 0)
    parsedModules = newParsed;
  else {
    curParsed = parsedModules;
    while(curParsed->next != 0)
      curParsed = curParsed->next;
    curParsed->next = newParsed;
  }
}

module *newModule(modType theType, atype* name, atype *argsToConst, atype* super, declList* body){
  module* newMod;
  newMod = (module*)checkAlloc(1, sizeof(module));
  newMod->kind = theType;
  newMod->name = name;
  newMod->argsToConst = argsToConst;
  newMod->super = super;
  newMod->body = body;
  return newMod;
}

declList *addDecl(declList* list, decl *newNode){
  declList *newLink, *curLink;
  newLink = (declList*)checkAlloc(1, sizeof(declList));
  newLink->next = 0;
  newLink->val = newNode;
  if (list == 0)
    return newLink;
  else {
    curLink = list;
    while (curLink->next != 0)
      curLink = curLink->next;
    curLink->next = newLink;
    return list;
  }
}

decl *newInpDecl(sigList* body){
  decl *newD;
  newD = (decl*)checkAlloc(1, sizeof(decl));
  newD->kind = dInput;
  newD->val.sig = body;
  return newD;
}

decl *newOutpDecl(sigList* body){
  decl *newD;
  newD = (decl*)checkAlloc(1, sizeof(decl));
  newD->kind = dOutput;
  newD->val.sig = body;
  return newD;
}

decl *newSubModDecl(subMod* body){
  decl *newD;
  newD = (decl*)checkAlloc(1, sizeof(decl));
  newD->kind = dModule;
  newD->val.sub = body;
  return newD;
}

decl *newConnDecl(connect* body){
  decl *newD;
  newD = (decl*)checkAlloc(1, sizeof(decl));
  newD->kind = dConnect;
  newD->val.con = body;
  return newD;
}

decl *newEquivDecl(connect* body) {
  decl *newD;
  newD = (decl*)checkAlloc(1, sizeof(decl));
  newD->kind = dEquiv;
  newD->val.con = body;
  return newD;
}

decl *newNoDepDecl(nameList *noDepSigs) {
  decl *newD;
  newD = (decl*)checkAlloc(1, sizeof(decl));
  newD->kind = dNoDep;
  newD->val.name = noDepSigs;
  return newD;
}

sigList *addSig(sigList *list, sig *newSig){
  sigList *newLink, *curLink;
  newLink = (sigList*)checkAlloc(1, sizeof(sigList));
  newLink->next = 0;
  newLink->val = newSig;
  if (list == 0)
    return newLink;
  else {
    curLink = list;
    while (curLink->next != 0)
      curLink = curLink->next;
    curLink->next = newLink;
    return list;
  }
}

sig *newSignal(sigType theKind, atype *theType, char* owner, char *name){
  sig *newSig;
  newSig = (sig*)checkAlloc(1, sizeof(sig));
  newSig->kind = theKind;
  newSig->theType = theType;
  newSig->owner = owner;
  newSig->name = name;
  return newSig;
}

paramList *addParam(paramList *list, param *newParam) {
  paramList *newLink, *curLink;
  newLink = (paramList*)checkAlloc(1, sizeof(paramList));
  newLink->next = 0;
  newLink->val = newParam;
  if (list == 0)
    return newLink;
  else {
    curLink = list;
    while (curLink->next != 0)
      curLink = curLink->next;
    curLink->next = newLink;
    return list;
  }
}

param *newParam(atype *theType, char *name) {
  param *newParam;
  newParam = (param*)checkAlloc(1, sizeof(param));
  newParam->theType = theType;
  newParam->name = name;
  return newParam;
}

modList *addModule(modList *list, subModInst* newModule){
  modList *newLink, *curLink;
  newLink = (modList*)checkAlloc(1, sizeof(modList));
  newLink->next = 0;
  newLink->modInst = newModule;
  if (list == 0)
    return newLink;
  else {
    curLink = list;
    while (curLink->next != 0)
      curLink = curLink->next;
    curLink->next = newLink;
    return list;
  }
}

subModInst* newSubModInst(char *name, char *litArgs) {
  subModInst *newSub;
  newSub = (subModInst*)checkAlloc(1, sizeof(subModInst));
  newSub->modName = name;
  newSub->litArgs = litArgs;
  return newSub;
}

connect *newConnect(sig *from, sig *to){
  connect *newCon;
  newCon = (connect*)checkAlloc(1, sizeof(connect));
  newCon->from = from;
  newCon->to = to;
  return newCon;
}

subMod *newSubModules(atype *theType, modList *theMods){
  subMod *newSub;
  newSub = (subMod*)checkAlloc(1, sizeof(subMod));
  newSub->theType = theType;
  newSub->theMods = theMods;
  return newSub;
}

nameList *addName(nameList *list, char *name) {
  nameList *newLink, *curLink;
  newLink = (nameList*)checkAlloc(1, sizeof(nameList));
  newLink->next = 0;
  newLink->name = name;
  if (list == 0)
    return newLink;
  else {
    curLink = list;
    while (curLink->next != 0)
      curLink = curLink->next;
    curLink->next = newLink;
    return list;
  }
}

atype *newType(char *name, typeList *parms) {
  atype *newType;
  
  newType = (atype*)checkAlloc(1, sizeof(atype));
  newType->name = name;
  newType->params = parms;
  return newType;
}

atype *buildTypeFromLiteral(char *literal) {
  int i;
  int l;
  int go = 0;
  int j = 0;
  char *typname;
  
  l = strlen(literal);
  for(i = 0; i < l; i++) {
    if (go) {
      tmpStr[j++] = literal[i];
    } else {
      if (literal[i] != ' ') {
        go = 1;
        tmpStr[j++] = literal[i];
      }
    }
  }
  tmpStr[j] = 0;
  
  for(i = j-1; i >= 0; i--) {
    if (tmpStr[i] == ' ') {
      tmpStr[i] = 0;
    } else {
      i = -1;
    }
  }
  
  typname = (char*)checkAlloc(strlen(tmpStr)+1,1);
  strcpy(typname,tmpStr);
  
  return newType(typname,0);
}

typeList *addType(typeList* list, atype *typ) {
  typeList *newLink, *curLink;
  newLink = (typeList*)checkAlloc(1, sizeof(typeList));
  newLink->next = 0;
  newLink->theType = typ;
  if (list == 0)
    return newLink;
  else {
    curLink = list;
    while (curLink->next != 0)
      curLink = curLink->next;
    curLink->next = newLink;
    return list;
  }
}

void checkLoopInTypdef(atype *ntyp, atype *otyp) {
  atype *tdtyp, *lstyp;
  char *ot, *nt;
  
  lstyp = otyp;
  while (((tdtyp = findTypdef(lstyp)) != 0)
         && (strcmp(tdtyp->name,ntyp->name) != 0))
    lstyp = tdtyp;
  if (tdtyp != 0) {
    nt = NameOfType(ntyp);
	ot = NameOfType(otyp);
    sprintf(tmpStr,"Defining type %s as %s introduces a typedef loop:",
          nt, ot);
    mdfLog(tmpStr);
    sprintf(tmpStr,"  %s is %s",nt, ot);
    mdfLog(tmpStr);
	free(nt); free(ot);
    lstyp = otyp;
    while (((tdtyp = findTypdef(lstyp)) != 0)
           && (strcmp(tdtyp->name, ntyp->name) != 0)) {
	  ot = NameOfType(tdtyp);
      sprintf(tmpStr,"  which is %s", ot);
      mdfLog(tmpStr);
	  free(ot);
      lstyp = tdtyp;
    }
	ot = NameOfType(tdtyp);
    sprintf(tmpStr,"  which is %s", ot);
    mdfLog(tmpStr);
	free(ot);
    mdfError("Infinite recursive type definition.");
  }
}

void declareTypdef(atype *newType, atype *oldType) {
  addTypdef(newType, oldType, 1);
}

void declareTypequ(atype *newType, atype *oldType) {
  addTypdef(newType, oldType, 0);
}

void addTypdef(atype *newType, atype *oldType, int isTypdef) {
  typdef *newDef;
  atype *oldDef;
  typdefList *newLink, *curLink;
  char *ot, *nt, *dt;
  
  if ((oldDef = findTypdef(newType)) != 0) {
    if (cmpRawTypes(oldDef, oldType) != 0) {
	  nt = NameOfType(newType);
	  ot = NameOfType(oldType);
	  dt = NameOfType(oldDef);
      sprintf(tmpStr,"Re-defining %s as %s. Already defined as %s",
              nt, ot, dt);
	  free(nt); free(ot); free(dt);
      mdfError(tmpStr);
    } else {
	  nt = NameOfType(newType);
	  ot = NameOfType(oldType);
      sprintf(tmpStr,"Re-defining %s as %s. Ignored.", nt, ot);
	  free(nt); free(ot);
      mdfWarn(tmpStr);
    }
  } else {
    checkLoopInTypdef(newType, oldType);
    newDef = (typdef*)checkAlloc(1, sizeof(typdef));
    newDef->newType = newType;
    newDef->oldType = oldType;
    newDef->mustDef = isTypdef;
    newLink = (typdefList*)checkAlloc(1, sizeof(typdefList));
    newLink->next = 0;
    newLink->theDef = newDef;
  
    if (theTypdefs == 0)
      theTypdefs = newLink;
    else {
      curLink = theTypdefs;
      while(curLink->next != 0)
        curLink = curLink->next;
      curLink->next = newLink;
    }
  }
}

void HInsert(nameList* h) {
  HInsertData = h;
}

void CInsert(nameList* h) {
  CInsertData = h;
}

void declareUse(char *usename) {
  useList = addName(useList,usename);
}

char* emptyNullStr(char* str) {
  if (str == 0)
    return "";
  else
    return str;
}

void printModule(module *newModule){
  declList *curDecl;
  List curArgs;
  sig *from, *to;
  char *t;
  
  switch (newModule->kind) {
    case mDeclared:
      printf("Module declaration\n");
      break;
    case mComposite:
      printf("Composite module definition\n");
      break;
    case mDerived:
      printf("Derived module definition\n");
      break;
    default:
      printf("Unknown module kind\n");
      break;
  }
  printf("  Source file: %s\n",emptyNullStr(newModule->source));
  printf("  oc source file: %s\n",emptyNullStr(newModule->oc));
  printf("  C++ include file: %s\n",emptyNullStr(newModule->cpp));
  t = NameOfType(newModule->name);
  printf("  Name: %s\n",emptyNullStr(t));
  free(t);
  if (newModule->super != 0) {
    t = NameOfType(newModule->super);
    printf("  Derived from: %s\n",emptyNullStr(t));
	free(t);
  }
  curDecl = newModule->body;
  while (curDecl != 0) {
    curArgs = curDecl->val->val;
    switch (curDecl->val->kind) {
      case dInput:
        printf("  Input signals:\n");
        while (curArgs.sig != 0) {
          switch (curArgs.sig->val->kind) {
            case sTyped:
			  t = NameOfType(curArgs.sig->val->theType);
              printf("    Signal %s of type %s\n",
                      curArgs.sig->val->name, t);
			  free(t);
              break;
            case sPure:
              printf("    Pure signal %s \n",
                      curArgs.sig->val->name);
              break;
            case sMain:
              printf("    Signal %s of super\n",
                      curArgs.sig->val->name);
              break;
            case sSub:
              printf("    Signal %s of module %s\n",
                      curArgs.sig->val->name,
                      curArgs.sig->val->owner);
              break;
            default:
              printf("    Unknown kind of signal\n");
              break;
          }
          curArgs.sig = curArgs.sig->next;
        }
        break;
      case dOutput:
        printf("  Output signals:\n");
        while (curArgs.sig != 0) {
          switch (curArgs.sig->val->kind) {
            case sTyped:
			  t = NameOfType(curArgs.sig->val->theType);
              printf("    Signal %s of type %s\n",
                      curArgs.sig->val->name, t);
			  free(t);
              break;
            case sPure:
              printf("    Pure signal %s \n",
                      curArgs.sig->val->name);
              break;
            case sMain:
              printf("    Signal %s of super\n",
                      curArgs.sig->val->name);
              break;
            case sSub:
              printf("    Signal %s of module %s\n",
                      curArgs.sig->val->name,
                      curArgs.sig->val->owner);
              break;
            default:
              printf("    Unknown kind of signal\n");
              break;
          }
          curArgs.sig = curArgs.sig->next;
        }
        break;
      case dModule:
	    t = NameOfType(curArgs.sub->theType);
        printf ("  %s :\n", t);
		free(t);
        curArgs.mod = curArgs.sub->theMods;
        while (curArgs.mod != 0) {
          if (curArgs.mod->modInst->litArgs == 0)
            printf("    %s\n", curArgs.mod->modInst->modName);
          else
            printf("    %s(%s)\n",
                   curArgs.mod->modInst->modName,
                   curArgs.mod->modInst->litArgs);
          curArgs.mod = curArgs.mod->next;
        }
        break;
      case dConnect:
      case dEquiv:
        from = curArgs.con->from;
        to = curArgs.con->to;
        switch (from->kind) {
          case sTyped:
          case sPure:
            printf("  Unexpected sig type for connect! %d ", from->kind);
            break;
          case sMain:
            printf("  %s ",from->name);
            break;
          case sSub:
            printf("  %s.%s ", from->owner, from->name);
            break;
          default:
            printf("  Unknown sig type %d ", from->kind);
        }
        if (curDecl->val->kind == dConnect)
          printf("<< ");
        if (curDecl->val->kind == dEquiv)
          printf("= ");
        switch (to->kind) {
          case sTyped:
          case sPure:
            printf("Unexpected sig type for connect! %d\n", to->kind);
            break;
          case sMain:
            printf("%s\n",to->name);
            break;
          case sSub:
            printf("%s.%s\n", to->owner, to->name);
            break;
          default:
            printf("Unknown sig type %d\n", to->kind);
        }
        break;
      case dNoDep:
        while(curArgs.name != 0) {
          printf("  Signal %s does not [depend on | contribute to] module reaction.\n",
                 curArgs.name->name);
          curArgs.name = curArgs.name->next;
        }
        break;
      default:
        printf(" Unknown kind of declaration\n");
        break;
    }
    curDecl = curDecl->next;
  }
}

void printLastModule() {
  parsedList *curParsed;
  
  curParsed = parsedModules;
  while(curParsed->next != 0)
    curParsed = curParsed->next;
  printModule(curParsed->parsedMod);
}
